// $Header: /cvsroot/esrg/sfesrg/esrgpcpj/shared/c_datd/gmp_ints.c,v 1.22 2002/01/27 15:18:44 dtashley Exp $

//--------------------------------------------------------------------------------
//Copyright 2001 David T. Ashley
//-------------------------------------------------------------------------------------------------
//This source code and any program in which it is compiled/used is provided under the GNU GENERAL
//PUBLIC LICENSE, Version 3, full license text below.
//-------------------------------------------------------------------------------------------------
//                    GNU GENERAL PUBLIC LICENSE
//                       Version 3, 29 June 2007
//
// Copyright (C) 2007 Free Software Foundation, Inc. <http://fsf.org/>
// Everyone is permitted to copy and distribute verbatim copies
// of this license document, but changing it is not allowed.
//
//                            Preamble
//
//  The GNU General Public License is a free, copyleft license for
//software and other kinds of works.
//
//  The licenses for most software and other practical works are designed
//to take away your freedom to share and change the works.  By contrast,
//the GNU General Public License is intended to guarantee your freedom to
//share and change all versions of a program--to make sure it remains free
//software for all its users.  We, the Free Software Foundation, use the
//GNU General Public License for most of our software; it applies also to
//any other work released this way by its authors.  You can apply it to
//your programs, too.
//
//  When we speak of free software, we are referring to freedom, not
//price.  Our General Public Licenses are designed to make sure that you
//have the freedom to distribute copies of free software (and charge for
//them if you wish), that you receive source code or can get it if you
//want it, that you can change the software or use pieces of it in new
//free programs, and that you know you can do these things.
//
//  To protect your rights, we need to prevent others from denying you
//these rights or asking you to surrender the rights.  Therefore, you have
//certain responsibilities if you distribute copies of the software, or if
//you modify it: responsibilities to respect the freedom of others.
//
//  For example, if you distribute copies of such a program, whether
//gratis or for a fee, you must pass on to the recipients the same
//freedoms that you received.  You must make sure that they, too, receive
//or can get the source code.  And you must show them these terms so they
//know their rights.
//
//  Developers that use the GNU GPL protect your rights with two steps:
//(1) assert copyright on the software, and (2) offer you this License
//giving you legal permission to copy, distribute and/or modify it.
//
//  For the developers' and authors' protection, the GPL clearly explains
//that there is no warranty for this free software.  For both users' and
//authors' sake, the GPL requires that modified versions be marked as
//changed, so that their problems will not be attributed erroneously to
//authors of previous versions.
//
//  Some devices are designed to deny users access to install or run
//modified versions of the software inside them, although the manufacturer
//can do so.  This is fundamentally incompatible with the aim of
//protecting users' freedom to change the software.  The systematic
//pattern of such abuse occurs in the area of products for individuals to
//use, which is precisely where it is most unacceptable.  Therefore, we
//have designed this version of the GPL to prohibit the practice for those
//products.  If such problems arise substantially in other domains, we
//stand ready to extend this provision to those domains in future versions
//of the GPL, as needed to protect the freedom of users.
//
//  Finally, every program is threatened constantly by software patents.
//States should not allow patents to restrict development and use of
//software on general-purpose computers, but in those that do, we wish to
//avoid the special danger that patents applied to a free program could
//make it effectively proprietary.  To prevent this, the GPL assures that
//patents cannot be used to render the program non-free.
//
//  The precise terms and conditions for copying, distribution and
//modification follow.
//
//                       TERMS AND CONDITIONS
//
//  0. Definitions.
//
//  "This License" refers to version 3 of the GNU General Public License.
//
//  "Copyright" also means copyright-like laws that apply to other kinds of
//works, such as semiconductor masks.
//
//  "The Program" refers to any copyrightable work licensed under this
//License.  Each licensee is addressed as "you".  "Licensees" and
//"recipients" may be individuals or organizations.
//
//  To "modify" a work means to copy from or adapt all or part of the work
//in a fashion requiring copyright permission, other than the making of an
//exact copy.  The resulting work is called a "modified version" of the
//earlier work or a work "based on" the earlier work.
//
//  A "covered work" means either the unmodified Program or a work based
//on the Program.
//
//  To "propagate" a work means to do anything with it that, without
//permission, would make you directly or secondarily liable for
//infringement under applicable copyright law, except executing it on a
//computer or modifying a private copy.  Propagation includes copying,
//distribution (with or without modification), making available to the
//public, and in some countries other activities as well.
//
//  To "convey" a work means any kind of propagation that enables other
//parties to make or receive copies.  Mere interaction with a user through
//a computer network, with no transfer of a copy, is not conveying.
//
//  An interactive user interface displays "Appropriate Legal Notices"
//to the extent that it includes a convenient and prominently visible
//feature that (1) displays an appropriate copyright notice, and (2)
//tells the user that there is no warranty for the work (except to the
//extent that warranties are provided), that licensees may convey the
//work under this License, and how to view a copy of this License.  If
//the interface presents a list of user commands or options, such as a
//menu, a prominent item in the list meets this criterion.
//
//  1. Source Code.
//
//  The "source code" for a work means the preferred form of the work
//for making modifications to it.  "Object code" means any non-source
//form of a work.
//
//  A "Standard Interface" means an interface that either is an official
//standard defined by a recognized standards body, or, in the case of
//interfaces specified for a particular programming language, one that
//is widely used among developers working in that language.
//
//  The "System Libraries" of an executable work include anything, other
//than the work as a whole, that (a) is included in the normal form of
//packaging a Major Component, but which is not part of that Major
//Component, and (b) serves only to enable use of the work with that
//Major Component, or to implement a Standard Interface for which an
//implementation is available to the public in source code form.  A
//"Major Component", in this context, means a major essential component
//(kernel, window system, and so on) of the specific operating system
//(if any) on which the executable work runs, or a compiler used to
//produce the work, or an object code interpreter used to run it.
//
//  The "Corresponding Source" for a work in object code form means all
//the source code needed to generate, install, and (for an executable
//work) run the object code and to modify the work, including scripts to
//control those activities.  However, it does not include the work's
//System Libraries, or general-purpose tools or generally available free
//programs which are used unmodified in performing those activities but
//which are not part of the work.  For example, Corresponding Source
//includes interface definition files associated with source files for
//the work, and the source code for shared libraries and dynamically
//linked subprograms that the work is specifically designed to require,
//such as by intimate data communication or control flow between those
//subprograms and other parts of the work.
//
//  The Corresponding Source need not include anything that users
//can regenerate automatically from other parts of the Corresponding
//Source.
//
//  The Corresponding Source for a work in source code form is that
//same work.
//
//  2. Basic Permissions.
//
//  All rights granted under this License are granted for the term of
//copyright on the Program, and are irrevocable provided the stated
//conditions are met.  This License explicitly affirms your unlimited
//permission to run the unmodified Program.  The output from running a
//covered work is covered by this License only if the output, given its
//content, constitutes a covered work.  This License acknowledges your
//rights of fair use or other equivalent, as provided by copyright law.
//
//  You may make, run and propagate covered works that you do not
//convey, without conditions so long as your license otherwise remains
//in force.  You may convey covered works to others for the sole purpose
//of having them make modifications exclusively for you, or provide you
//with facilities for running those works, provided that you comply with
//the terms of this License in conveying all material for which you do
//not control copyright.  Those thus making or running the covered works
//for you must do so exclusively on your behalf, under your direction
//and control, on terms that prohibit them from making any copies of
//your copyrighted material outside their relationship with you.
//
//  Conveying under any other circumstances is permitted solely under
//the conditions stated below.  Sublicensing is not allowed; section 10
//makes it unnecessary.
//
//  3. Protecting Users' Legal Rights From Anti-Circumvention Law.
//
//  No covered work shall be deemed part of an effective technological
//measure under any applicable law fulfilling obligations under article
//11 of the WIPO copyright treaty adopted on 20 December 1996, or
//similar laws prohibiting or restricting circumvention of such
//measures.
//
//  When you convey a covered work, you waive any legal power to forbid
//circumvention of technological measures to the extent such circumvention
//is effected by exercising rights under this License with respect to
//the covered work, and you disclaim any intention to limit operation or
//modification of the work as a means of enforcing, against the work's
//users, your or third parties' legal rights to forbid circumvention of
//technological measures.
//
//  4. Conveying Verbatim Copies.
//
//  You may convey verbatim copies of the Program's source code as you
//receive it, in any medium, provided that you conspicuously and
//appropriately publish on each copy an appropriate copyright notice;
//keep intact all notices stating that this License and any
//non-permissive terms added in accord with section 7 apply to the code;
//keep intact all notices of the absence of any warranty; and give all
//recipients a copy of this License along with the Program.
//
//  You may charge any price or no price for each copy that you convey,
//and you may offer support or warranty protection for a fee.
//
//  5. Conveying Modified Source Versions.
//
//  You may convey a work based on the Program, or the modifications to
//produce it from the Program, in the form of source code under the
//terms of section 4, provided that you also meet all of these conditions:
//
//    a) The work must carry prominent notices stating that you modified
//    it, and giving a relevant date.
//
//    b) The work must carry prominent notices stating that it is
//    released under this License and any conditions added under section
//    7.  This requirement modifies the requirement in section 4 to
//    "keep intact all notices".
//
//    c) You must license the entire work, as a whole, under this
//    License to anyone who comes into possession of a copy.  This
//    License will therefore apply, along with any applicable section 7
//    additional terms, to the whole of the work, and all its parts,
//    regardless of how they are packaged.  This License gives no
//    permission to license the work in any other way, but it does not
//    invalidate such permission if you have separately received it.
//
//    d) If the work has interactive user interfaces, each must display
//    Appropriate Legal Notices; however, if the Program has interactive
//    interfaces that do not display Appropriate Legal Notices, your
//    work need not make them do so.
//
//  A compilation of a covered work with other separate and independent
//works, which are not by their nature extensions of the covered work,
//and which are not combined with it such as to form a larger program,
//in or on a volume of a storage or distribution medium, is called an
//"aggregate" if the compilation and its resulting copyright are not
//used to limit the access or legal rights of the compilation's users
//beyond what the individual works permit.  Inclusion of a covered work
//in an aggregate does not cause this License to apply to the other
//parts of the aggregate.
//
//  6. Conveying Non-Source Forms.
//
//  You may convey a covered work in object code form under the terms
//of sections 4 and 5, provided that you also convey the
//machine-readable Corresponding Source under the terms of this License,
//in one of these ways:
//
//    a) Convey the object code in, or embodied in, a physical product
//    (including a physical distribution medium), accompanied by the
//    Corresponding Source fixed on a durable physical medium
//    customarily used for software interchange.
//
//    b) Convey the object code in, or embodied in, a physical product
//    (including a physical distribution medium), accompanied by a
//    written offer, valid for at least three years and valid for as
//    long as you offer spare parts or customer support for that product
//    model, to give anyone who possesses the object code either (1) a
//    copy of the Corresponding Source for all the software in the
//    product that is covered by this License, on a durable physical
//    medium customarily used for software interchange, for a price no
//    more than your reasonable cost of physically performing this
//    conveying of source, or (2) access to copy the
//    Corresponding Source from a network server at no charge.
//
//    c) Convey individual copies of the object code with a copy of the
//    written offer to provide the Corresponding Source.  This
//    alternative is allowed only occasionally and noncommercially, and
//    only if you received the object code with such an offer, in accord
//    with subsection 6b.
//
//    d) Convey the object code by offering access from a designated
//    place (gratis or for a charge), and offer equivalent access to the
//    Corresponding Source in the same way through the same place at no
//    further charge.  You need not require recipients to copy the
//    Corresponding Source along with the object code.  If the place to
//    copy the object code is a network server, the Corresponding Source
//    may be on a different server (operated by you or a third party)
//    that supports equivalent copying facilities, provided you maintain
//    clear directions next to the object code saying where to find the
//    Corresponding Source.  Regardless of what server hosts the
//    Corresponding Source, you remain obligated to ensure that it is
//    available for as long as needed to satisfy these requirements.
//
//    e) Convey the object code using peer-to-peer transmission, provided
//    you inform other peers where the object code and Corresponding
//    Source of the work are being offered to the general public at no
//    charge under subsection 6d.
//
//  A separable portion of the object code, whose source code is excluded
//from the Corresponding Source as a System Library, need not be
//included in conveying the object code work.
//
//  A "User Product" is either (1) a "consumer product", which means any
//tangible personal property which is normally used for personal, family,
//or household purposes, or (2) anything designed or sold for incorporation
//into a dwelling.  In determining whether a product is a consumer product,
//doubtful cases shall be resolved in favor of coverage.  For a particular
//product received by a particular user, "normally used" refers to a
//typical or common use of that class of product, regardless of the status
//of the particular user or of the way in which the particular user
//actually uses, or expects or is expected to use, the product.  A product
//is a consumer product regardless of whether the product has substantial
//commercial, industrial or non-consumer uses, unless such uses represent
//the only significant mode of use of the product.
//
//  "Installation Information" for a User Product means any methods,
//procedures, authorization keys, or other information required to install
//and execute modified versions of a covered work in that User Product from
//a modified version of its Corresponding Source.  The information must
//suffice to ensure that the continued functioning of the modified object
//code is in no case prevented or interfered with solely because
//modification has been made.
//
//  If you convey an object code work under this section in, or with, or
//specifically for use in, a User Product, and the conveying occurs as
//part of a transaction in which the right of possession and use of the
//User Product is transferred to the recipient in perpetuity or for a
//fixed term (regardless of how the transaction is characterized), the
//Corresponding Source conveyed under this section must be accompanied
//by the Installation Information.  But this requirement does not apply
//if neither you nor any third party retains the ability to install
//modified object code on the User Product (for example, the work has
//been installed in ROM).
//
//  The requirement to provide Installation Information does not include a
//requirement to continue to provide support service, warranty, or updates
//for a work that has been modified or installed by the recipient, or for
//the User Product in which it has been modified or installed.  Access to a
//network may be denied when the modification itself materially and
//adversely affects the operation of the network or violates the rules and
//protocols for communication across the network.
//
//  Corresponding Source conveyed, and Installation Information provided,
//in accord with this section must be in a format that is publicly
//documented (and with an implementation available to the public in
//source code form), and must require no special password or key for
//unpacking, reading or copying.
//
//  7. Additional Terms.
//
//  "Additional permissions" are terms that supplement the terms of this
//License by making exceptions from one or more of its conditions.
//Additional permissions that are applicable to the entire Program shall
//be treated as though they were included in this License, to the extent
//that they are valid under applicable law.  If additional permissions
//apply only to part of the Program, that part may be used separately
//under those permissions, but the entire Program remains governed by
//this License without regard to the additional permissions.
//
//  When you convey a copy of a covered work, you may at your option
//remove any additional permissions from that copy, or from any part of
//it.  (Additional permissions may be written to require their own
//removal in certain cases when you modify the work.)  You may place
//additional permissions on material, added by you to a covered work,
//for which you have or can give appropriate copyright permission.
//
//  Notwithstanding any other provision of this License, for material you
//add to a covered work, you may (if authorized by the copyright holders of
//that material) supplement the terms of this License with terms:
//
//    a) Disclaiming warranty or limiting liability differently from the
//    terms of sections 15 and 16 of this License; or
//
//    b) Requiring preservation of specified reasonable legal notices or
//    author attributions in that material or in the Appropriate Legal
//    Notices displayed by works containing it; or
//
//    c) Prohibiting misrepresentation of the origin of that material, or
//    requiring that modified versions of such material be marked in
//    reasonable ways as different from the original version; or
//
//    d) Limiting the use for publicity purposes of names of licensors or
//    authors of the material; or
//
//    e) Declining to grant rights under trademark law for use of some
//    trade names, trademarks, or service marks; or
//
//    f) Requiring indemnification of licensors and authors of that
//    material by anyone who conveys the material (or modified versions of
//    it) with contractual assumptions of liability to the recipient, for
//    any liability that these contractual assumptions directly impose on
//    those licensors and authors.
//
//  All other non-permissive additional terms are considered "further
//restrictions" within the meaning of section 10.  If the Program as you
//received it, or any part of it, contains a notice stating that it is
//governed by this License along with a term that is a further
//restriction, you may remove that term.  If a license document contains
//a further restriction but permits relicensing or conveying under this
//License, you may add to a covered work material governed by the terms
//of that license document, provided that the further restriction does
//not survive such relicensing or conveying.
//
//  If you add terms to a covered work in accord with this section, you
//must place, in the relevant source files, a statement of the
//additional terms that apply to those files, or a notice indicating
//where to find the applicable terms.
//
//  Additional terms, permissive or non-permissive, may be stated in the
//form of a separately written license, or stated as exceptions;
//the above requirements apply either way.
//
//  8. Termination.
//
//  You may not propagate or modify a covered work except as expressly
//provided under this License.  Any attempt otherwise to propagate or
//modify it is void, and will automatically terminate your rights under
//this License (including any patent licenses granted under the third
//paragraph of section 11).
//
//  However, if you cease all violation of this License, then your
//license from a particular copyright holder is reinstated (a)
//provisionally, unless and until the copyright holder explicitly and
//finally terminates your license, and (b) permanently, if the copyright
//holder fails to notify you of the violation by some reasonable means
//prior to 60 days after the cessation.
//
//  Moreover, your license from a particular copyright holder is
//reinstated permanently if the copyright holder notifies you of the
//violation by some reasonable means, this is the first time you have
//received notice of violation of this License (for any work) from that
//copyright holder, and you cure the violation prior to 30 days after
//your receipt of the notice.
//
//  Termination of your rights under this section does not terminate the
//licenses of parties who have received copies or rights from you under
//this License.  If your rights have been terminated and not permanently
//reinstated, you do not qualify to receive new licenses for the same
//material under section 10.
//
//  9. Acceptance Not Required for Having Copies.
//
//  You are not required to accept this License in order to receive or
//run a copy of the Program.  Ancillary propagation of a covered work
//occurring solely as a consequence of using peer-to-peer transmission
//to receive a copy likewise does not require acceptance.  However,
//nothing other than this License grants you permission to propagate or
//modify any covered work.  These actions infringe copyright if you do
//not accept this License.  Therefore, by modifying or propagating a
//covered work, you indicate your acceptance of this License to do so.
//
//  10. Automatic Licensing of Downstream Recipients.
//
//  Each time you convey a covered work, the recipient automatically
//receives a license from the original licensors, to run, modify and
//propagate that work, subject to this License.  You are not responsible
//for enforcing compliance by third parties with this License.
//
//  An "entity transaction" is a transaction transferring control of an
//organization, or substantially all assets of one, or subdividing an
//organization, or merging organizations.  If propagation of a covered
//work results from an entity transaction, each party to that
//transaction who receives a copy of the work also receives whatever
//licenses to the work the party's predecessor in interest had or could
//give under the previous paragraph, plus a right to possession of the
//Corresponding Source of the work from the predecessor in interest, if
//the predecessor has it or can get it with reasonable efforts.
//
//  You may not impose any further restrictions on the exercise of the
//rights granted or affirmed under this License.  For example, you may
//not impose a license fee, royalty, or other charge for exercise of
//rights granted under this License, and you may not initiate litigation
//(including a cross-claim or counterclaim in a lawsuit) alleging that
//any patent claim is infringed by making, using, selling, offering for
//sale, or importing the Program or any portion of it.
//
//  11. Patents.
//
//  A "contributor" is a copyright holder who authorizes use under this
//License of the Program or a work on which the Program is based.  The
//work thus licensed is called the contributor's "contributor version".
//
//  A contributor's "essential patent claims" are all patent claims
//owned or controlled by the contributor, whether already acquired or
//hereafter acquired, that would be infringed by some manner, permitted
//by this License, of making, using, or selling its contributor version,
//but do not include claims that would be infringed only as a
//consequence of further modification of the contributor version.  For
//purposes of this definition, "control" includes the right to grant
//patent sublicenses in a manner consistent with the requirements of
//this License.
//
//  Each contributor grants you a non-exclusive, worldwide, royalty-free
//patent license under the contributor's essential patent claims, to
//make, use, sell, offer for sale, import and otherwise run, modify and
//propagate the contents of its contributor version.
//
//  In the following three paragraphs, a "patent license" is any express
//agreement or commitment, however denominated, not to enforce a patent
//(such as an express permission to practice a patent or covenant not to
//sue for patent infringement).  To "grant" such a patent license to a
//party means to make such an agreement or commitment not to enforce a
//patent against the party.
//
//  If you convey a covered work, knowingly relying on a patent license,
//and the Corresponding Source of the work is not available for anyone
//to copy, free of charge and under the terms of this License, through a
//publicly available network server or other readily accessible means,
//then you must either (1) cause the Corresponding Source to be so
//available, or (2) arrange to deprive yourself of the benefit of the
//patent license for this particular work, or (3) arrange, in a manner
//consistent with the requirements of this License, to extend the patent
//license to downstream recipients.  "Knowingly relying" means you have
//actual knowledge that, but for the patent license, your conveying the
//covered work in a country, or your recipient's use of the covered work
//in a country, would infringe one or more identifiable patents in that
//country that you have reason to believe are valid.
//
//  If, pursuant to or in connection with a single transaction or
//arrangement, you convey, or propagate by procuring conveyance of, a
//covered work, and grant a patent license to some of the parties
//receiving the covered work authorizing them to use, propagate, modify
//or convey a specific copy of the covered work, then the patent license
//you grant is automatically extended to all recipients of the covered
//work and works based on it.
//
//  A patent license is "discriminatory" if it does not include within
//the scope of its coverage, prohibits the exercise of, or is
//conditioned on the non-exercise of one or more of the rights that are
//specifically granted under this License.  You may not convey a covered
//work if you are a party to an arrangement with a third party that is
//in the business of distributing software, under which you make payment
//to the third party based on the extent of your activity of conveying
//the work, and under which the third party grants, to any of the
//parties who would receive the covered work from you, a discriminatory
//patent license (a) in connection with copies of the covered work
//conveyed by you (or copies made from those copies), or (b) primarily
//for and in connection with specific products or compilations that
//contain the covered work, unless you entered into that arrangement,
//or that patent license was granted, prior to 28 March 2007.
//
//  Nothing in this License shall be construed as excluding or limiting
//any implied license or other defenses to infringement that may
//otherwise be available to you under applicable patent law.
//
//  12. No Surrender of Others' Freedom.
//
//  If conditions are imposed on you (whether by court order, agreement or
//otherwise) that contradict the conditions of this License, they do not
//excuse you from the conditions of this License.  If you cannot convey a
//covered work so as to satisfy simultaneously your obligations under this
//License and any other pertinent obligations, then as a consequence you may
//not convey it at all.  For example, if you agree to terms that obligate you
//to collect a royalty for further conveying from those to whom you convey
//the Program, the only way you could satisfy both those terms and this
//License would be to refrain entirely from conveying the Program.
//
//  13. Use with the GNU Affero General Public License.
//
//  Notwithstanding any other provision of this License, you have
//permission to link or combine any covered work with a work licensed
//under version 3 of the GNU Affero General Public License into a single
//combined work, and to convey the resulting work.  The terms of this
//License will continue to apply to the part which is the covered work,
//but the special requirements of the GNU Affero General Public License,
//section 13, concerning interaction through a network will apply to the
//combination as such.
//
//  14. Revised Versions of this License.
//
//  The Free Software Foundation may publish revised and/or new versions of
//the GNU General Public License from time to time.  Such new versions will
//be similar in spirit to the present version, but may differ in detail to
//address new problems or concerns.
//
//  Each version is given a distinguishing version number.  If the
//Program specifies that a certain numbered version of the GNU General
//Public License "or any later version" applies to it, you have the
//option of following the terms and conditions either of that numbered
//version or of any later version published by the Free Software
//Foundation.  If the Program does not specify a version number of the
//GNU General Public License, you may choose any version ever published
//by the Free Software Foundation.
//
//  If the Program specifies that a proxy can decide which future
//versions of the GNU General Public License can be used, that proxy's
//public statement of acceptance of a version permanently authorizes you
//to choose that version for the Program.
//
//  Later license versions may give you additional or different
//permissions.  However, no additional obligations are imposed on any
//author or copyright holder as a result of your choosing to follow a
//later version.
//
//  15. Disclaimer of Warranty.
//
//  THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
//APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
//HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
//OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
//THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
//PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
//IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
//ALL NECESSARY SERVICING, REPAIR OR CORRECTION.
//
//  16. Limitation of Liability.
//
//  IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
//WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
//THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
//GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
//USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
//DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
//PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
//EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
//SUCH DAMAGES.
//
//  17. Interpretation of Sections 15 and 16.
//
//  If the disclaimer of warranty and limitation of liability provided
//above cannot be given local legal effect according to their terms,
//reviewing courts shall apply local law that most closely approximates
//an absolute waiver of all civil liability in connection with the
//Program, unless a warranty or assumption of liability accompanies a
//copy of the Program in return for a fee.
//
//                     END OF TERMS AND CONDITIONS
//
//            How to Apply These Terms to Your New Programs
//
//  If you develop a new program, and you want it to be of the greatest
//possible use to the public, the best way to achieve this is to make it
//free software which everyone can redistribute and change under these terms.
//
//  To do so, attach the following notices to the program.  It is safest
//to attach them to the start of each source file to most effectively
//state the exclusion of warranty; and each file should have at least
//the "copyright" line and a pointer to where the full notice is found.
//
//    <one line to give the program's name and a brief idea of what it does.>
//    Copyright (C) <year>  <name of author>
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//Also add information on how to contact you by electronic and paper mail.
//
//  If the program does terminal interaction, make it output a short
//notice like this when it starts in an interactive mode:
//
//    <program>  Copyright (C) <year>  <name of author>
//    This program comes with ABSOLUTELY NO WARRANTY; for details type `show w'.
//    This is free software, and you are welcome to redistribute it
//    under certain conditions; type `show c' for details.
//
//The hypothetical commands `show w' and `show c' should show the appropriate
//parts of the General Public License.  Of course, your program's commands
//might be different; for a GUI interface, you would use an "about box".
//
//  You should also get your employer (if you work as a programmer) or school,
//if any, to sign a "copyright disclaimer" for the program, if necessary.
//For more information on this, and how to apply and follow the GNU GPL, see
//<http://www.gnu.org/licenses/>.
//
//  The GNU General Public License does not permit incorporating your program
//into proprietary programs.  If your program is a subroutine library, you
//may consider it more useful to permit linking proprietary applications with
//the library.  If this is what you want to do, use the GNU Lesser General
//Public License instead of this License.  But first, please read
//<http://www.gnu.org/philosophy/why-not-lgpl.html>.
//-------------------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------
#define MODULE_GMP_INTS

#include <assert.h>
#include <ctype.h>
#include <process.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>
#include <time.h>


/* Only included the guarded allocation header if we are compiling
** a DOS console type application.  Other types of applications have
** other ways of protecting for out of memory.  Including the
** header would do no harm in these cases, but do no good, either.
*/
#if defined(APP_TYPE_SIMPLE_DOS_CONSOLE)
   #include "ccmalloc.h"
#elif defined(APP_TYPE_IJUSCRIPTER_IJUCONSOLE)
   #include "tclalloc.h"
#else
   /* Do nothing. */
#endif

#include "bstrfunc.h"
#include "charfunc.h"
#include "fcmiof.h"
#include "gmp_ints.h"
#include "intfunc.h"


/******************************************************************/
/***  CUSTOM ALLOCATION FUNCTIONS   *******************************/
/******************************************************************/
/* We need to decide here on how memory not on the stack will be
** allocated (i.e. what will become of the standard functions
** like malloc, free, etc.).  This is dependent on the type of
** application we're making.  The possible types are so far are:
**    APP_TYPE_SIMPLE_DOS_CONSOLE : 
**       Simple DOS console application.
**    APP_TYPE_IJUSCRIPTER_IJUCONSOLE:
**       The Tcl tool build.
** 
** The custom allocation functions here are a "portal" or "wrapper"
** for how the integer and rational number functions should
** get memory.
**
** The functions below are standard, except that the GNU MP team
** built more generality into what allocation schemes could be
** used by including size information in some calls that don't
** normally get it.  That is why there are some extra calls below
** where the information is discarded.  Other than that, these are
** standard allocation calls.
*/
//07/15/01:  Visual inspection only.  Function deemed too 
//simple for unit testing.
void *GMP_INTS_malloc( size_t size )
   {
   #if defined(APP_TYPE_SIMPLE_DOS_CONSOLE)
      return(CCMALLOC_malloc(size));
   #elif defined(APP_TYPE_IJUSCRIPTER_IJUCONSOLE)
      return(TclpAlloc(size));
   #else
      return(malloc(size));
   #endif 
   }


//07/15/01:  Visual inspection only.  Function deemed too 
//simple for unit testing.
void *GMP_INTS_calloc( size_t num, size_t size )
   {
   #if defined(APP_TYPE_SIMPLE_DOS_CONSOLE)
      return(CCMALLOC_calloc(num, size));
   #elif defined(APP_TYPE_IJUSCRIPTER_IJUCONSOLE)
      return(TclpCalloc(num, size));
   #else
      return(calloc(num, size));
   #endif 
   }


//07/15/01:  Visual inspection only.  Function deemed too 
//simple for unit testing.
void *GMP_INTS_realloc( void *memblock, size_t size )
   {
   #if defined(APP_TYPE_SIMPLE_DOS_CONSOLE)
      return(CCMALLOC_realloc(memblock, size));
   #elif defined(APP_TYPE_IJUSCRIPTER_IJUCONSOLE)
      return(TclpRealloc(memblock, size));
   #else
      return(realloc(memblock, size));
   #endif 
   }


//07/15/01:  Visual inspection only.  Function deemed too 
//simple for unit testing.
void *GMP_INTS_realloc_w_size( void *memblock, 
                               size_t old_size,
                               size_t size )
   {
   #if defined(APP_TYPE_SIMPLE_DOS_CONSOLE)
      return(CCMALLOC_realloc(memblock, size));
   #elif defined(APP_TYPE_IJUSCRIPTER_IJUCONSOLE)
      return(TclpRealloc(memblock, size));
   #else
      return(realloc(memblock, size));
   #endif 
   }


//07/15/01:  Visual inspection only.  Function deemed too 
//simple for unit testing.
void GMP_INTS_free( void *memblock )
   {
   #if defined(APP_TYPE_SIMPLE_DOS_CONSOLE)
      CCMALLOC_free(memblock);
   #elif defined(APP_TYPE_IJUSCRIPTER_IJUCONSOLE)
      TclpFree(memblock);
   #else
      free(memblock);
   #endif 
   }


//07/15/01:  Visual inspection only.  Function deemed too 
//simple for unit testing.
void GMP_INTS_free_w_size( void *memblock, size_t size )
   {
   #if defined(APP_TYPE_SIMPLE_DOS_CONSOLE)
      CCMALLOC_free(memblock);
   #elif defined(APP_TYPE_IJUSCRIPTER_IJUCONSOLE)
      TclpFree(memblock);
   #else
      free(memblock);
   #endif 
   }


/******************************************************************/
/***  PORTABILITY CHECK FUNCTIONS   *******************************/
/******************************************************************/
//Because there is the risk that Microsoft Visual C++ might 
//change in the future, the following function can be called
//to see if the assumptions about data sizes are valid.  This
//function returns TRUE if there is a problem, or FALSE
//otherwise.

//07/15/01:  Unit testing complete.
int GMP_INTS_data_sizes_are_wrong(void)
   {
   int i;
   GMP_INTS_limb_t tv;
   _int64 tv64;

   //Check the number of bit rolls required to get the limb
   //to go to zero again.  This had better be 32.
   tv = 1;
   i = 0;
   while (tv)
      {
      tv <<= 1;
      i++;
      }
   if (i != 32)
      return(1);

   //Check that an _int64 is really and truly 64 bits.
   tv64 = 1;
   i = 0;
   while (tv64)
      {
      tv64 <<= 1;
      i++;
      }
   if (i != 64)
      return(1);

   //Room for additional tests here if needed later.

   return(0);
   }


/******************************************************************/
/***  ERROR STRING IDENTIFICATION AND PROCESSING FUNCTIONS  *******/
/******************************************************************/

int GMP_INTS_identify_nan_string(const char *s)
   {
   assert(s != NULL);

   if (!strcmp(s, GMP_INTS_EF_INTOVF_POS_STRING))
      return(0);
   else if (!strcmp(s, GMP_INTS_EF_INTOVF_NEG_STRING))
      return(1);
   else if (!strcmp(s, GMP_INTS_EF_INTOVF_TAINT_POS_STRING))
      return(2);
   else if (!strcmp(s, GMP_INTS_EF_INTOVF_TAINT_NEG_STRING))
      return(3);
   else
      return(-1);
   }


const char *GMP_INTS_supply_nan_string(int idx)
   {
   assert((idx >= 0) && (idx <= 3));

   if (idx==0)
      return(GMP_INTS_EF_INTOVF_POS_STRING);
   else if (idx==1)
      return(GMP_INTS_EF_INTOVF_NEG_STRING);
   else if (idx==2)
      return(GMP_INTS_EF_INTOVF_TAINT_POS_STRING);
   else
      return(GMP_INTS_EF_INTOVF_TAINT_NEG_STRING);
   }


/******************************************************************/
/***  DEBUG PRINTING FUNCTIONS   **********************************/
/******************************************************************/
//These functions are for printing out integers and limbs
//and groups of limbs for unit testing and debugging.                        

//07/15/01:  Exempt from testing because debug/development
//function.
void GMP_INTS_print_limb_group(FILE *stream,
                               GMP_INTS_limb_srcptr lg,
                               GMP_INTS_size_t n,
                               char *desc)
   {
   int i;

   assert(stream != NULL);
   assert(n >= 0);
   assert(desc != NULL);

   if (!lg)
      {
      fprintf(stream, "   %s: NULL\n", desc);
      }
   else
      {
      for (i=n-1; i>=0; i--)
         {
         fprintf(stream, "   %s[%2d]: 0x%8X\n", desc, i, lg[i]);
         }
      }
   }


void GMP_INTS_mpz_print_int(FILE *stream,
                            const GMP_INTS_mpz_struct *arg,
                            char *desc)
   {
   int i;

   assert(stream != NULL);
   assert(arg != NULL);
   assert(desc != NULL);

   fprintf(stream, "Printing integer:\n   %s\n", desc);

   fprintf(stream, "               flags: %d\n", arg->flags);
   fprintf(stream, "   ptr value to body: %p\n", arg);
   fprintf(stream, "            n_allocd: %d\n", arg->n_allocd);
   fprintf(stream, "                size: %d\n", arg->size);
   fprintf(stream, "     limbs (ptr val): %p\n", arg->limbs);

   for (i=arg->n_allocd-1; i>=0; i--)
      {
      fprintf(stream, "   limb[%3d]: %8X\n", i, arg->limbs[i]);
      }
   }


/******************************************************************/
/***  LOW-LEVEL MACRO REPLACEMENTS   ******************************/
/******************************************************************/
//The functions in this category are replacements for macros.
//Clarity was gained at the expense of speed.

int GMP_INTS_mpz_get_flags (const GMP_INTS_mpz_struct *arg)
   {
   assert(arg != NULL);
   assert(arg->n_allocd > 0);

   return(arg->flags);
   } 


//07/15/01:  Visual inspection only.  Function deemed too 
//simple for unit testing.
GMP_INTS_size_t GMP_INTS_abs_of_size_t(GMP_INTS_size_t arg)
   {
   //Be sure that the bit pattern does not represent the maximum
   //negative argument.  Negating this would give the result of
   //zero, which is not what is intended.
   assert(arg != 0x80000000);

   if (arg < 0)
      return(-arg);
   else
      return(arg);
   }


//07/15/01:  Visual inspection only.  Function deemed too 
//simple for unit testing.
int GMP_INTS_mpz_sgn(const GMP_INTS_mpz_struct *arg)
   {
   assert(arg != NULL);
   assert(arg->n_allocd > 0);

   if (arg->size > 0)
      return(1);
   else if (arg->size == 0)
      return(0);
   else
      return(-1);
   }


//07/15/01:  Visual inspection only.  Function deemed too 
//simple for unit testing.
int GMP_INTS_mpz_is_neg(const GMP_INTS_mpz_struct *arg)
   {
   assert(arg != NULL);
   assert(arg->n_allocd > 0);

   if (GMP_INTS_mpz_sgn(arg) == -1)
      return(1);
   else
      return(0);
   }


//07/15/01:  Visual inspection only.  Function deemed too 
//simple for unit testing.
int GMP_INTS_mpz_is_zero(const GMP_INTS_mpz_struct *arg)
   {
   assert(arg != NULL);
   assert(arg->n_allocd > 0);

   if (GMP_INTS_mpz_sgn(arg) == 0)
      return(1);
   else
      return(0);
   }


//07/15/01:  Visual inspection only.  Function deemed too 
//simple for unit testing.
int GMP_INTS_mpz_is_pos(const GMP_INTS_mpz_struct *arg)
   {
   assert(arg != NULL);
   assert(arg->n_allocd > 0);

   if (GMP_INTS_mpz_sgn(arg) == 1)
      return(1);
   else
      return(0);
   }


//07/15/01:  Visual inspection only.  Function deemed too 
//simple for unit testing.
int GMP_INTS_mpz_is_odd(const GMP_INTS_mpz_struct *arg)
   {
   assert(arg != NULL);
   assert(arg->n_allocd > 0);

   if (arg->size == 0)
      return 0;
   else if ((arg->limbs[0] & 0x1) != 0)
      return 1;
   else
      return 0;
   }


//07/15/01:  Visual inspection only.  Function deemed too 
//simple for unit testing.
int GMP_INTS_mpz_is_even(const GMP_INTS_mpz_struct *arg)
   {
   assert(arg != NULL);
   assert(arg->n_allocd > 0);

   if (GMP_INTS_mpz_is_odd(arg))
      return 0;
   else
      return 1;
   }


void GMP_INTS_mpz_negate(GMP_INTS_mpz_struct *arg)
   {
   //Eyeball the input parameters.
   assert(arg != NULL);
   assert(arg->n_allocd > 0);
   assert(arg->limbs != NULL);

   arg->size = -(arg->size);
   }


//07/15/01:  Visual inspection only.  Function deemed too 
//simple for unit testing.
void GMP_INTS_mpn_normalize(GMP_INTS_limb_ptr  limb_array,
                            GMP_INTS_size_t   *idx)
   {
   assert(limb_array != NULL);
   assert(idx != NULL);
   assert(idx >= 0);

   while (*idx > 0)
      {
      if (limb_array[*idx - 1] != 0)
         break;
      (*idx)--;
      }
   }


//07/15/01:  Visual inspection only.  Function deemed too 
//simple for unit testing.
void GMP_INTS_mpn_copy_limbs(GMP_INTS_limb_ptr    dest,
                             GMP_INTS_limb_srcptr src,
                             GMP_INTS_size_t      n)
  {
  GMP_INTS_size_t i;

  assert(dest != NULL);
  assert(src != NULL);
  assert(n >= 0);

  for (i=0; i<n; i++)
    dest[i] = src[i];
  }


/******************************************************************/
/***  LOW-LEVEL ARITHMETIC FUNCTIONS   ****************************/
/******************************************************************/

int GMP_INTS_two_op_flags_map(int flags1, int flags2)
   {
   int cf;

   if (!flags1 && !flags2)
      {
      return(0);
      }
   else
      {
      cf = flags1 | flags2;

      if (cf & (GMP_INTS_EF_INTOVF_POS | GMP_INTS_EF_INTOVF_TAINT_POS))
         {
         //In either case here, the result will be tainted 
         //positive.
         return(GMP_INTS_EF_INTOVF_TAINT_POS);
         }
      else if (cf & (GMP_INTS_EF_INTOVF_NEG | GMP_INTS_EF_INTOVF_TAINT_NEG))
         {
         //In either case here, the result will be tainted
         //negative.
         return(GMP_INTS_EF_INTOVF_TAINT_NEG);
         }
      else
         {
         //This case is where the flags ints are non-zero, but
         //no known bits are set.  This is surely some kind of
         //an internal software error.  In debug mode, want to
         //alert to error.  In actual operation, want to just
         //pretend an ordinary positive taint.
         assert(0);
         return(GMP_INTS_EF_INTOVF_TAINT_POS);
         }
      }
   }


GMP_INTS_limb_t GMP_INTS_mpn_add_1 (GMP_INTS_limb_ptr    res_ptr,
                                    GMP_INTS_limb_srcptr s1_ptr,
                                    GMP_INTS_size_t      s1_size,
                                    GMP_INTS_limb_t      s2_limb)
   {
   GMP_INTS_limb_t x;

   assert(res_ptr != NULL);
   assert(s1_ptr != NULL);
   assert(s1_size > 0);

   x = *s1_ptr++;
   s2_limb = x + s2_limb;
   *res_ptr++ = s2_limb;
   //Since limbs are unsigned, the test below tests if there
   //was a carry, i.e. a positive rollover.
   if (s2_limb < x)
      {
      while (--s1_size != 0)
         {
         x = *s1_ptr++ + 1;
         *res_ptr++ = x;
         if (x != 0)
            goto fin;
         }
      
      return 1;
      }
   
   fin:
   if (res_ptr != s1_ptr)
      {
      GMP_INTS_size_t i;
      for (i = 0; i < s1_size - 1; i++)
         {
         res_ptr[i] = s1_ptr[i];
         }
      }

   return 0;
   }


GMP_INTS_limb_t GMP_INTS_mpn_sub_1(GMP_INTS_limb_ptr    res_ptr,
                                   GMP_INTS_limb_srcptr s1_ptr,
                                   GMP_INTS_size_t      s1_size,
                                   GMP_INTS_limb_t      s2_limb)
   {
   GMP_INTS_limb_t x;

   assert(res_ptr != NULL);
   assert(s1_ptr  != NULL);
   assert(s1_size > 0);

   x = *s1_ptr++;
   s2_limb = x - s2_limb;
   *res_ptr++ = s2_limb;
   //The test below detects a borrow.
   if (s2_limb > x)
      {
      while (--s1_size != 0)
         {
         x = *s1_ptr++;
         *res_ptr++ = x - 1;
         if (x != 0)
            goto fin;
         }
      
      return 1;
      }
   
   fin:
   if (res_ptr != s1_ptr)
      {
      GMP_INTS_size_t i;
      for (i = 0; i < s1_size - 1; i++)
         {
            res_ptr[i] = s1_ptr[i];
         }
      }
      return 0;
   }


//07/15/01:  Am willing to skip unit-testing on this.
//Understand the logic (i.e. passes visual inspection), 
//and comes from GNU-MP.  Hope any defects here will be 
//caught in testing of GMP_INTS_mpz_mul() and other 
//higher-level functions.
GMP_INTS_limb_t GMP_INTS_mpn_mul_1 (GMP_INTS_limb_ptr res_ptr, 
						                  GMP_INTS_limb_srcptr s1_ptr, 
						                  GMP_INTS_size_t s1_size, 
						                  GMP_INTS_limb_t s2_limb)
   {
   GMP_INTS_limb_t cy_limb;
   GMP_INTS_size_t j;
   GMP_INTS_limb_t prod_high, prod_low;
   unsigned _int64 temp;
   
   assert(res_ptr != NULL);
   assert(s1_ptr != NULL);
   assert(s1_size > 0);

   /* The loop counter and index J goes from -S1_SIZE to -1.  This way
   the loop becomes faster.  */
   j = -s1_size;
   
   /* Offset the base pointers to compensate for the negative indices.  */
   s1_ptr -= j;
   res_ptr -= j;
   
   cy_limb = 0;
   do
      {
      //The original code here was the following macro:
      //umul_ppmm (prod_high, prod_low, s1_ptr[j], s2_limb);
      //Will use the 64-bit data type of MSVC++ to achieve
      //the same effect.
      //
      //NOTE AS OF 07/13/01:  I have looked at the assembly-
      //language, and the lines below are a real sore spot.
      //The multiply is fairly direct (although there is a
      //function call), but the shift does not behave as
      //expected--there is a function call and a loop to
      //go through the 32 iterations.  After logical testing,
      //may want to clean this out--this would surely
      //result in a speed increase.  This is a sore spot.
      temp = ((unsigned _int64)s1_ptr[j]) * ((unsigned _int64)s2_limb);
      prod_low  = (GMP_INTS_limb_t)temp;
      prod_high = (GMP_INTS_limb_t)(temp >> 32);

      prod_low += cy_limb;
      cy_limb = (prod_low < cy_limb) + prod_high;
      
      res_ptr[j] = prod_low;
      }
   while (++j != 0);
      
   return cy_limb;
   }


//07/15/01:  Am willing to skip unit-testing on this.
//Understand the logic (i.e. passes visual inspection), 
//and comes from GNU-MP.  Hope any defects here will be 
//caught in testing of GMP_INTS_mpz_add() and other 
//higher-level functions.
GMP_INTS_limb_t GMP_INTS_mpn_add_n(GMP_INTS_limb_ptr    res_ptr,  
                                   GMP_INTS_limb_srcptr s1_ptr, 
                                   GMP_INTS_limb_srcptr s2_ptr, 
                                   GMP_INTS_size_t      size)
   {
   GMP_INTS_limb_t x, y, cy;
   GMP_INTS_size_t j;

   assert(res_ptr != NULL);
   assert(s1_ptr  != NULL);
   assert(s2_ptr  != NULL);

   /* The loop counter and index J goes from -SIZE to -1.  This way
      the loop becomes faster.  */
   j = -size;

   /* Offset the base pointers to compensate for the negative indices.  */
   s1_ptr -= j;
   s2_ptr -= j;
   res_ptr -= j;

   cy = 0;
   do
      {
      y = s2_ptr[j];
      x = s1_ptr[j];
      y += cy;   /* add previous carry to one addend */
      cy = (y < cy);  /* get out carry from that addition */
      y = x + y;  /* add other addend */
      cy = (y < x) + cy; /* get out carry from that add, combine */
      res_ptr[j] = y;
      }
   while (++j != 0);

   return cy;
   }


//07/15/01:  Am willing to skip unit-testing on this.
//Understand the logic (i.e. passes visual inspection), 
//and comes from GNU-MP.  Hope any defects here will be 
//caught in testing of GMP_INTS_mpz_mul() and other 
//higher-level functions.
GMP_INTS_limb_t GMP_INTS_mpn_addmul_1 (GMP_INTS_limb_ptr    res_ptr, 
                                       GMP_INTS_limb_srcptr s1_ptr, 
                                       GMP_INTS_size_t      s1_size, 
                                       GMP_INTS_limb_t      s2_limb)
   {
   GMP_INTS_limb_t cy_limb;
   GMP_INTS_size_t j;
   GMP_INTS_limb_t prod_high, prod_low;
   GMP_INTS_limb_t x;
   unsigned _int64 temp;

   //Eyeball the inputs carefully.
   assert(res_ptr != NULL);
   assert(s1_ptr != NULL);
   assert(s1_size > 0);

   /* The loop counter and index J goes from -SIZE to -1.  This way
      the loop becomes faster.  */
   j = -s1_size;

   /* Offset the base pointers to compensate for the negative indices.  */
   res_ptr -= j;
   s1_ptr -= j;

   cy_limb = 0;
   do
      {
      //The original code here was the following macro:
      //umul_ppmm (prod_high, prod_low, s1_ptr[j], s2_limb);
      //Will use the 64-bit data type of MSVC++ to achieve
      //the same effect.
      //
      //NOTE AS OF 07/14/01:  I have not looked at the assembly-
      //language, but the assembly-language generated by what
      //is below is suspected to have performance problems.
      //May want to come back to this.
      temp = ((unsigned _int64)s1_ptr[j]) * ((unsigned _int64)s2_limb);
      prod_low  = (GMP_INTS_limb_t)temp;
      prod_high = (GMP_INTS_limb_t)(temp >> 32);

      prod_low += cy_limb;
      cy_limb = (prod_low < cy_limb) + prod_high;

      x = res_ptr[j];
      prod_low = x + prod_low;
      cy_limb += (prod_low < x);
      res_ptr[j] = prod_low;
      }
   while (++j != 0);

   return cy_limb;
   }


//07/15/01:  Am willing to skip unit-testing on this.
//Understand the logic (i.e. passes visual inspection), 
//and comes from GNU-MP.
GMP_INTS_limb_t GMP_INTS_mpn_add (GMP_INTS_limb_ptr    res_ptr,
                                  GMP_INTS_limb_srcptr s1_ptr,
                                  GMP_INTS_size_t      s1_size,
                                  GMP_INTS_limb_srcptr s2_ptr,
                                  GMP_INTS_size_t      s2_size)
   {
   GMP_INTS_limb_t cy_limb = 0;

   assert(res_ptr != NULL);
   assert(s1_ptr  != NULL);
   assert(s2_ptr  != NULL);

   //Numbers apparently must be arranged with sizes so that 
   //LIMBS(s1) >= LIMBS(s2).
   //Add the parts up to the most significant limb of S2.
   if (s2_size != 0)
      cy_limb = GMP_INTS_mpn_add_n (res_ptr, 
                                    s1_ptr, 
                                    s2_ptr, 
                                    s2_size);

   //Process the carry result, and propagate the carries up through
   //the parts of S1 that don't exist in S2, i.e. propagate the
   //carries upward in S1.
   if (s1_size - s2_size != 0)
      cy_limb = GMP_INTS_mpn_add_1 (res_ptr + s2_size,
                                    s1_ptr + s2_size,
                                    s1_size - s2_size,
                                    cy_limb);
   return cy_limb;
   }


//07/15/01:  Am willing to skip unit-testing on this.
//Understand the logic (i.e. passes visual inspection), 
//and comes from GNU-MP.
GMP_INTS_limb_t GMP_INTS_mpn_sub_n(GMP_INTS_limb_ptr    res_ptr, 
                                   GMP_INTS_limb_srcptr s1_ptr, 
                                   GMP_INTS_limb_srcptr s2_ptr, 
                                   GMP_INTS_size_t      size)
   {
   GMP_INTS_limb_t x, y, cy;
   GMP_INTS_size_t j;

   assert(res_ptr != NULL);
   assert(s1_ptr  != NULL);
   assert(s2_ptr  != NULL);

   /* The loop counter and index J goes from -SIZE to -1.  This way
      the loop becomes faster.  */
   j = -size;

   /* Offset the base pointers to compensate for the negative indices.  */
   s1_ptr -= j;
   s2_ptr -= j;
   res_ptr -= j;

   cy = 0;
   do
      {
      y = s2_ptr[j];
      x = s1_ptr[j];
      y += cy;   /* add previous carry to subtrahend */
      cy = (y < cy);  /* get out carry from that addition */
      y = x - y;  /* main subtract */
      cy = (y > x) + cy; /* get out carry from the subtract, combine */
      res_ptr[j] = y;
      }
   while (++j != 0);

   return cy;
   }


//07/17/01:  Am willing to skip unit-testing on this.
//Understand the logic (i.e. passes visual inspection), 
//and comes from GNU-MP.
GMP_INTS_limb_t GMP_INTS_mpn_sub (GMP_INTS_limb_ptr    res_ptr,
                                  GMP_INTS_limb_srcptr s1_ptr,
                                  GMP_INTS_size_t      s1_size,
                                  GMP_INTS_limb_srcptr s2_ptr,
                                  GMP_INTS_size_t      s2_size)
   {
   GMP_INTS_limb_t cy_limb = 0;

   assert(res_ptr != NULL);
   assert(s1_ptr  != NULL);
   assert(s2_ptr  != NULL);

   if (s2_size != 0)
      cy_limb = GMP_INTS_mpn_sub_n(res_ptr, 
                                   s1_ptr, 
                                   s2_ptr, 
                                   s2_size);

   if (s1_size - s2_size != 0)
      cy_limb = GMP_INTS_mpn_sub_1(res_ptr + s2_size,
                                   s1_ptr + s2_size,
                                   s1_size - s2_size,
                                   cy_limb);
   return cy_limb;
   }


//07/17/01:  Am willing to skip unit-testing on this.
//Understand the logic (i.e. passes visual inspection), 
//and comes from GNU-MP.
GMP_INTS_limb_t GMP_INTS_mpn_lshift(GMP_INTS_limb_ptr    wp,
                                    GMP_INTS_limb_srcptr up, 
                                    GMP_INTS_size_t      usize,
                                    unsigned int         cnt)
   {
   GMP_INTS_limb_t high_limb, low_limb;
   unsigned sh_1, sh_2;
   GMP_INTS_size_t i;
   GMP_INTS_limb_t retval;

   assert(wp != NULL);
   assert(up != NULL);
   assert(usize > 0);
   assert(cnt > 0);

   sh_1 = cnt;

   wp += 1;
   sh_2 = GMP_INTS_BITS_PER_LIMB - sh_1;
     //This automatically implies that can't call this function to shift more
     //than 31 places.
   i = usize - 1;
   low_limb = up[i];
   retval = low_limb >> sh_2;  //Return value is the amount shifted
                               //off the top.
   high_limb = low_limb;
   while (--i >= 0)
      {
      low_limb = up[i];
      wp[i] = (high_limb << sh_1) | (low_limb >> sh_2);
      high_limb = low_limb;
      }
   wp[i] = high_limb << sh_1;

   return retval;
   }


//07/17/01:  Am willing to skip unit-testing on this.
//Understand the logic more or less (i.e. passes visual inspection), 
//and comes from GNU-MP.
/* Shift U (pointed to by UP and USIZE limbs long) CNT bits to the right
   and store the USIZE least significant limbs of the result at WP.
   The bits shifted out to the right are returned.

   Argument constraints:
   1. 0 < CNT < BITS_PER_MP_LIMB
   2. If the result is to be written over the input, WP must be <= UP.
*/
GMP_INTS_limb_t GMP_INTS_mpn_rshift (GMP_INTS_limb_ptr    wp,
                                     GMP_INTS_limb_srcptr up, 
                                     GMP_INTS_size_t      usize,
                                     unsigned int         cnt)
   {
   GMP_INTS_limb_t high_limb, low_limb;
   unsigned sh_1, sh_2;
   GMP_INTS_size_t i;
   GMP_INTS_limb_t retval;

   assert(wp != NULL);
   assert(up != NULL);
   assert(usize > 0);
   assert(cnt > 0);

   sh_1 = cnt;

   wp -= 1;
   sh_2 = GMP_INTS_BITS_PER_LIMB - sh_1;
   high_limb = up[0];
   retval = high_limb << sh_2;
   low_limb = high_limb;

   for (i = 1; i < usize; i++)
      {
      high_limb = up[i];
      wp[i] = (low_limb >> sh_1) | (high_limb << sh_2);
      low_limb = high_limb;
      }
   wp[i] = low_limb >> sh_1;

   return retval;
   }


//07/17/01:  Am willing to skip unit-testing on this.
//Understand the logic (i.e. passes visual inspection), 
//and comes from GNU-MP.
int GMP_INTS_mpn_cmp (GMP_INTS_limb_srcptr op1_ptr, 
                      GMP_INTS_limb_srcptr op2_ptr, 
                      GMP_INTS_size_t      size)
   {
   GMP_INTS_size_t i;
   GMP_INTS_limb_t op1_word, op2_word;

   assert(op1_ptr != NULL);
   assert(op2_ptr != NULL);

   for (i = size - 1; i >= 0; i--)
      {
      op1_word = op1_ptr[i];
      op2_word = op2_ptr[i];
      if (op1_word != op2_word)
         goto diff;
      }
   return 0;

   diff:
   //This can *not* be simplified to
   //   op2_word - op2_word
   //since that expression might give signed overflow.
   return (op1_word > op2_word) ? 1 : -1;
   }


//07/15/01:  Am willing to skip unit-testing on this.
//Understand the logic (i.e. passes visual inspection), 
//and comes from GNU-MP.  Hope any defects here will be 
//caught in testing of GMP_INTS_mpz_mul() and other 
//higher-level functions.
void GMP_INTS_mpn_mul_basecase (GMP_INTS_limb_ptr prodp,
                                GMP_INTS_limb_srcptr up, 
                                GMP_INTS_size_t usize,
                                GMP_INTS_limb_srcptr vp, 
                                GMP_INTS_size_t vsize)
   {
   assert(prodp != NULL);
   assert(up != NULL);
   assert(usize > 0);
   assert(vp != NULL);
   assert(vsize > 0);

   /* We first multiply by the low order one or two limbs, as the result can
      be stored, not added, to PROD.  We also avoid a loop for zeroing this
      way.  */
   prodp[usize] = GMP_INTS_mpn_mul_1 (prodp, up, usize, vp[0]);
   prodp++;
   vp++;
   vsize--;

   /* For each iteration in the loop, multiply U with one limb from V, and
      add the result to PROD.  */
   while (vsize != 0)
      {
      prodp[usize] = GMP_INTS_mpn_addmul_1 (prodp, up, usize, vp[0]);
      prodp++, 
      vp++, 
      vsize--;
      }
   }


//07/15/01:  No unit testing possible--this is a passthrough.
//In the original GNU MP code, there were several multiplication
//algorithms, and this function would select one based on the
//size of the operands and other considerations.  The code has been
//pared so that only simple multiplication is used, which is why
//this function contains only a single pass-thru function call.
void GMP_INTS_mpn_mul_n (GMP_INTS_limb_ptr    p, 
                         GMP_INTS_limb_srcptr a, 
                         GMP_INTS_limb_srcptr b, 
                         GMP_INTS_size_t      n)
   {
   GMP_INTS_mpn_mul_basecase (p, a, n, b, n);
   }


//07/16/01:  Visual inspection OK.  Will not perform unit testing.
GMP_INTS_limb_t GMP_INTS_mpn_mul(GMP_INTS_limb_ptr    prodp,
                                 GMP_INTS_limb_srcptr up, 
                                 GMP_INTS_size_t      un,
                                 GMP_INTS_limb_srcptr vp, 
                                 GMP_INTS_size_t      vn)
   {
   //This is a gutted version of the GNU MP function.  The GNU
   //MP function considered the case of a square, and also 
   //better algorithms that pay off with large operands.
   //This gutted version uses only basic multiplication
   //(O(N**2)).

   //Eyeball the input parameters.
   assert(prodp != NULL);
   assert(up != NULL);
   assert(un >= 0);
   assert(vp != NULL);
   assert(vn >= 0);

   /* Basic long multiplication. */
   GMP_INTS_mpn_mul_basecase (prodp, up, un, vp, vn);

   //Return the most significant limb (which might be zero).  
   //This is different than
   //most other functions, which return the spillover.
   return prodp[un + vn - 1];
   }


/******************************************************************/
/***  LIMB SPACE REALLOCATION FUNCTIONS   *************************/
/******************************************************************/

void *GMP_INTS_mpz_realloc (GMP_INTS_mpz_struct *m, 
                            GMP_INTS_size_t      new_size)
   {
   /* Never allocate zero space. */
   if (new_size <= 0)
      new_size = 1;

   m->limbs = (GMP_INTS_limb_ptr) 
               GMP_INTS_realloc_w_size (m->limbs,
                                        m->n_allocd * sizeof(GMP_INTS_limb_t),
                                        new_size * sizeof(GMP_INTS_limb_t));
   m->n_allocd = new_size;

   return (void *) m->limbs;
   }


/******************************************************************/
/***  PUBLIC INITIALIZATION AND MEMORY MANAGEMENT FUNCTIONS   *****/
/******************************************************************/

void GMP_INTS_mpz_init (GMP_INTS_mpz_struct *x)
   {
   assert(x != NULL);

   //The structure (the header block) exists in the
   //caller's area.  Most likely it is a local variable.
   //This is OK, because it doesn't take up much space.

   //Start off with no errors.
   x->flags = 0;
   
   //Allocate space for one limb, which is the most
   //basic amount.  This will grow, almost certainly.
   x->limbs = GMP_INTS_malloc(sizeof(GMP_INTS_limb_t));

   //Indicate that one limb was allocated.
   x->n_allocd = 1;

   //Set the size to 0.  This signals a value of zero.
   x->size = 0;
   }


void GMP_INTS_mpz_clear (GMP_INTS_mpz_struct *x)
   {
   //Be sure the passed pointer is not NULL.
   assert(x != NULL);

   //Be sure that the amount allocated is also above zero.
   //Anything else represents a logical error.
   assert(x->n_allocd > 0);

   //Be sure that the pointer to the allocated limbs
   //is not NULL.  Anything else would be a logical
   //error.
   assert(x->limbs != NULL);

   //Deallocate the space for the limbs.  The pointer is
   //set NULL and the allocated amount set to zero 
   // so in case clear is called again it will be
   //a detectable error.
   GMP_INTS_free_w_size(x->limbs, 
                        x->n_allocd * sizeof(GMP_INTS_limb_t));
   x->limbs = NULL;
   x->n_allocd = 0;
   }


/******************************************************************/
/***  PUBLIC ASSIGNMENT FUNCTIONS   *******************************/
/******************************************************************/

void GMP_INTS_mpz_copy(      GMP_INTS_mpz_struct *dst,
                       const GMP_INTS_mpz_struct *src)
   {
   GMP_INTS_size_t i, n;

   //Eyeball the input parameters.
   assert(dst != NULL);
   assert(dst->n_allocd > 0);
   assert(dst->limbs != NULL);
   assert(src != NULL);
   assert(src->n_allocd > 0);
   assert(src->limbs != NULL);

   //Source and destination may not be the same.
   assert(src != dst);
   
   //Figure out the real size of the source.  We need to take the absolute
   //value.
   n = GMP_INTS_abs_of_size_t(src->size);

   //Reallocate the destination to be bigger if necessary.
   if (dst->n_allocd < n)
      {
      GMP_INTS_mpz_realloc (dst, n);
      }

   //Copy the non-dynamic fields in the header.
   dst->flags     = src->flags;
   dst->size      = src->size;

   //Copy the limbs.
   for (i=0; i<n; i++)
      dst->limbs[i] = src->limbs[i];
   }


void GMP_INTS_mpz_set_ui (GMP_INTS_mpz_struct *dest, 
                          unsigned long int val)
   { 
   assert(dest != NULL);
   
   /* We don't check if the allocation is enough, since the rest of the
      package ensures it's at least 1, which is what we need here.  */

   dest->flags = 0;
      //A set operation resets any errors.

   if (val > 0)
      {
      dest->limbs[0] = val;
      dest->size = 1;
      }
   else
      {
      dest->size = 0;
      }
   }


void GMP_INTS_mpz_set_si (GMP_INTS_mpz_struct *dest, 
                          signed long int val)
   {
   assert(dest != NULL);

   /* We don't check if the allocation is enough, since the rest of the
      package ensures it's at least 1, which is what we need here.  */

   dest->flags = 0;
     //A set operation resets any errors.

   if (val > 0)
      {
      dest->limbs[0] = val;
      dest->size = 1;
      }
   else if (val < 0)
      {
      dest->limbs[0] = (unsigned long) -val;
      dest->size = -1;
      }
   else
      {
      dest->size = 0;
      }
   }


void GMP_INTS_mpz_set_simple_char_str(GMP_INTS_mpz_struct *z, 
                                      const char *s)
   {
   int sign=1;
   int digval;
   GMP_INTS_mpz_struct digvalz, k10;

   //Eyeball the arguments.
   assert(z != NULL);
   assert(z->n_allocd > 0);
   assert(z->limbs != NULL);
   assert(s != NULL);

   //Set the arbitrary integer to zero.  This will also kill
   //any error flags.
   GMP_INTS_mpz_set_ui(z, 0);

   //Allocate an integer for our private use to hold each digit
   //value.
   GMP_INTS_mpz_init(&digvalz);

   //Allocate the constant 10, which we will use often.
   GMP_INTS_mpz_init(&k10);
   GMP_INTS_mpz_set_ui(&k10, 10);

   //As long as there are are digits and no flags set, keep
   //multiplying and adding the value of the digit.  Non-
   //digits are simply ignored.
   while (!(z->flags) && (*s))
      {
      if (*s == '-')
         {
         sign = -sign;
         }
      else
         {
         digval = CHARFUNC_digit_to_val(*s);
         if (digval >= 0)
            {
            GMP_INTS_mpz_set_ui(&digvalz, digval);
            GMP_INTS_mpz_mul(z, z, &k10);
            GMP_INTS_mpz_add(z, z, &digvalz);
            }
         }
      s++;
      }

   //Adjust the final sign of the result.
   if (sign < 0)
      z->size = -(z->size);

   //Deallocate our temporary integers.
   GMP_INTS_mpz_clear(&digvalz);
   GMP_INTS_mpz_clear(&k10);
   }


void GMP_INTS_mpz_set_sci_not_num(GMP_INTS_mpz_struct *z,
                                  int *failure, 
                                  const char *s)
   {
   int parse_failure;
      //Return code from the floating point parsing
      //function.
   char mant_sign;
      //Sign character, if any, from the mantissa,
      //or N otherwise.
   size_t mant_bdp;
      //The index to the start of the mantissa before
      //the decimal point.
   size_t mant_bdp_len;
      //The length of the mantissa before the decimal
      //point.  Zero means not defined, i.e. that
      //no characters were parsed and interpreted as
      //that part of a floating point number.
   size_t mant_adp;
   size_t mant_adp_len;
      //Similar fields for after the decimal point.
   char exp_sign;
      //Sign of the exponent, if any, or N otherwise.
   size_t exp;
   size_t exp_len;
      //Similar fields as to the mantissa, but for the
      //exponent.
   size_t si;
      //Iteration variable.
   int exponent_val;
      //The value of the exponent.  We can't accept
      //an exponent outside the range of a 24-bit
      //signed integer.  The 24-bit limit is arbitrary.
      //For one thing, it gives room to detect overflow
      //as are adding and multiplying by 10.

   //Eyeball the input parameters.
   assert(z != NULL);
   assert(z->n_allocd > 0);
   assert(z->limbs != NULL);
   assert(failure != NULL);
   assert(s != NULL);

   //Start off believing no failure.
   *failure = 0;

   //Set the output to zero.  This is the default case for some
   //steps below.
   GMP_INTS_mpz_set_ui(z, 0);

   //Attempt to parse the number as a general number
   //in scientific notation.
   BSTRFUNC_parse_gen_sci_not_num(s,
                                  &parse_failure,
                                  &mant_sign,
                                  &mant_bdp,
                                  &mant_bdp_len,
                                  &mant_adp,
                                  &mant_adp_len,
                                  &exp_sign,
                                  &exp,
                                  &exp_len);

   //If it wouldn't parse as a general number, can't go further.
   if (parse_failure)
      {
      *failure = 1;
      return;
      }
   else if (!exp_len && !mant_adp_len)
      {
      //There was no exponent, and no portion after
      //the decimal point.  Can just parse as an integer.
      char *temp_buf;

      //Allocate the temporary buffer to be one character longer
      //than the length specified for the parsed mantissa.
      temp_buf = GMP_INTS_malloc(sizeof(char) * (mant_bdp_len + 1));

      //Copy from the parsed area into the temporary buffer.
      for (si=mant_bdp; si<(mant_bdp+mant_bdp_len); si++)
         temp_buf[si-mant_bdp] = s[si];
      temp_buf[mant_bdp_len] = 0;

      //Set the arbitrary integer to the value of the character
      //string.
      GMP_INTS_mpz_set_simple_char_str(z, temp_buf);

      //If the number parsed as negative, invert.
      if (mant_sign == '-')
         z->size = -z->size;

      //Deallocate the temporary buffer.
      GMP_INTS_free(temp_buf);
      }
   else if (!exp_len && mant_adp_len)
      {
      char *temp_buf;

      //In this case, there are digits after the decimal point,
      //but no exponent specified.  The only way this makes
      //sense is if all of the digits are zero--otherwise it
      //cannot be an integer.
      for (si=mant_adp; si<(mant_adp+mant_adp_len); si++)
         {
         if (s[si] != '0')
            {
            *failure = 1;
            return;
            }
         }

      //We're clean.  They are only zeros.  Execute as per
      //integer code.

      //Allocate the temporary buffer to be one character longer
      //than the length specified for the parsed mantissa.
      temp_buf = GMP_INTS_malloc(sizeof(char) * (mant_bdp_len + 1));

      //Copy from the parsed area into the temporary buffer.
      for (si=mant_bdp; si<(mant_bdp+mant_bdp_len); si++)
         temp_buf[si-mant_bdp] = s[si];
      temp_buf[mant_bdp_len] = 0;

      //Set the arbitrary integer to the value of the character
      //string.
      GMP_INTS_mpz_set_simple_char_str(z, temp_buf);

      //If the number parsed as negative, invert.
      if (mant_sign == '-')
         z->size = -z->size;

      //Deallocate the temporary buffer.
      GMP_INTS_free(temp_buf);
      }
   else if (exp_len)
      {
      //This is the most difficult case, where an exponent
      //is specified.  There are several complex subcases,
      //such as:
      //  a)If the exponent is too positive or too negative,
      //    we can't use it.  In general, we won't tackle
      //    an exponent that won't fit in a signed 24-bit
      //    integer.  This provides a range of from
      //    -8,388,608 to +8,388,607.  This dwarfs the
      //    100,000 or so digit preprocessor limit,
      //    and should be adequate for any practical
      //    application.
      //  b)If the exponent is zero, we ignore it.
      //  c)If the exponent is positive, it has to
      //    be large enough to overcome any 
      //    digits past the decimal point, otherwise
      //    we don't end up with an integer.
      //  d)If the exponent is negative, there have to
      //    be enough digits so that an integer remains
      //    after the exponent is applied.  This
      //    generally requires trailing zeros on the
      //    part before the decimal point.

      //First, tackle the exponent.  Process the
      //exponent into a signed integer.  We have to
      //balk at anything outside of 24 bits.  The
      //procedure used automatically handles
      //leading zeros correctly.
      exponent_val = 0;
      for (si=exp; si<(exp+exp_len); si++)
         {
         int val;

         val = CHARFUNC_digit_to_val(s[si]);

         assert(val >= 0 && val <= 9);

         exponent_val *= 10;
         exponent_val += val;

         if (((exp_sign=='-') && (exponent_val>8388608)) 
            || 
            ((exp_sign != '-') && (exponent_val>8388607)))
            {
            *failure = 1;
            return;
            }
         }

      //If we're here, the exponent has been computed and
      //is within 24 bits.  However, we need to adjust for
      //the sign.
      if (exp_sign == '-')
         exponent_val = -exponent_val;

      //We need to make accurate assertions about the
      //portion of the number, if any, after the decimal point.
      //This means that we need to effectively discard
      //trailing zeros.  To do this, we do not need to
      //relocate the string, we can just back off the index
      //to bypass any trailing zeros.
      while ((mant_adp_len > 0) && (s[mant_adp + mant_adp_len - 1]=='0'))
         mant_adp_len--;

      //We also need to make accurate assertions about the
      //portion of the number, if any, before the decimal
      //point.  It is known that the parsing function
      //isn't tolerant of multiple zeros, but zero is a 
      //special case.  Let's advance the pointer to the
      //part before the decimal point so that zero will
      //have zero length.
      while ((mant_bdp_len > 0) && (s[mant_bdp]=='0'))
         {
         mant_bdp++;
         mant_bdp_len--;
         }

      //If we are dealing with zero, who cares about the
      //exponent?  Just return the value of zero.
      if (!mant_bdp_len && !mant_adp_len)
         {
         *failure = 0;
         GMP_INTS_mpz_set_ui(z, 0);
         return;
         }
      
      //Beyond this point, we have something non-zero.
      //If the exponent is positive, it must be at least
      //as large as the number of digits beyond the
      //decimal point in order to form an integer.  If the
      //exponent is zero, there must be no digits after the
      //decimal point.  If the exponent is negative, there 
      //must be no digits after the decimal point, and the
      //trailing zeros on the part before the decimal point
      //must be adequate to handle the right decimal shift.
      if (exponent_val == 0)
         {
         if (mant_adp_len)
            {
            *failure = 1;
            return;
            }
         }
      else if (exponent_val > 0)
         {
         if ((int)mant_adp_len > exponent_val)
            {
            *failure = 1;
            return;
            }
         }
      else //exponent_val < 0
         {
         if (mant_adp_len)
            {
            *failure = 1;
            return;
            }
         else
            {
            //Count the number of trailing zeros on the part
            //before the decimal point.
            size_t trailing_zero_count;
            int idx;

            trailing_zero_count = 0;
        
            for(idx = mant_bdp + mant_bdp_len - 1; 
                (mant_bdp_len != 0) && (idx >= (int)mant_bdp);
                idx--)
                {
                if (s[idx] == '0')
                   trailing_zero_count++;
                else
                   break;
                }
             
            //Check on the assertion about trailing zeros.
            if ((int)trailing_zero_count < -exponent_val)
               {
               *failure = 1;
               return;
               }
            }
         }
     
         {
         //Create a string long enough to hold the digits
         //before the decimal point plus the ones after and
         //convert that to an arbitrary integer.
         //Form a power of 10 which is 10 exponentiated to
         //the absolute value of the exponent.  If the 
         //exponent was positive, multiply by it.  If the
         //exponent was negative, divide by it.
         char *conv_str;
         size_t sidx;
         GMP_INTS_mpz_struct power_of_ten, k10, trash;

         GMP_INTS_mpz_init(&power_of_ten);
         GMP_INTS_mpz_init(&k10);
         GMP_INTS_mpz_init(&trash);

         conv_str = GMP_INTS_malloc(sizeof(char) * (mant_bdp_len + mant_adp_len + 1));

         sidx=0;

         for (si=mant_bdp; si<(mant_bdp+mant_bdp_len); si++)
            {
            conv_str[sidx] = s[si];
            sidx++;
            }
         for (si=mant_adp; si<(mant_adp+mant_adp_len); si++)
            {
            conv_str[sidx] = s[si];
            sidx++;
            }
         conv_str[sidx] = 0;

         assert(sidx == (mant_bdp_len + mant_adp_len));
         
         GMP_INTS_mpz_set_simple_char_str(z, conv_str); 

         GMP_INTS_mpz_set_ui(&k10, 10);

         if (exponent_val > 0)
            GMP_INTS_mpz_pow_ui(&power_of_ten, &k10, exponent_val-mant_adp_len);
         else
            GMP_INTS_mpz_pow_ui(&power_of_ten, &k10, -exponent_val);

         if (exponent_val >= 0)
            {
            GMP_INTS_mpz_mul(z, z, &power_of_ten);
            }
         else
            {
            GMP_INTS_mpz_tdiv_qr (&k10, 
                                  &trash, 
                                  z, 
                                  &power_of_ten);
            GMP_INTS_mpz_copy(z, &k10);
            }

         //If the argument had a minus sign, invert.
         if (mant_sign == '-')
            z->size = -z->size;

         GMP_INTS_free(conv_str);

         GMP_INTS_mpz_clear(&trash);
         GMP_INTS_mpz_clear(&k10);
         GMP_INTS_mpz_clear(&power_of_ten);

         //Finally, if the arbitrary integer has overflowed, this is
         //a parse failure.  Must declare as such.
         if (z->flags)
            *failure = 1;
         }
      }
   else
      {
      *failure = 1;
      return;
      }
   }


void GMP_INTS_mpz_set_general_int(GMP_INTS_mpz_struct *z,
                                  int *failure, 
                                  const char *s)
   {
   //Eyeball the input parameters.
   assert(z != NULL);
   assert(z->n_allocd > 0);
   assert(z->limbs != NULL);
   assert(failure != NULL);
   assert(s != NULL);

   //Try to parse it as a simple integer.
   if (BSTRFUNC_is_sint_wo_commas(s))
      {
      GMP_INTS_mpz_set_simple_char_str(z, s);
      *failure = 0;
      return;
      }
   //If that didn't work, try to parse it as a simple
   //integer with commas.
   else if (BSTRFUNC_is_sint_w_commas(s))
      {
      GMP_INTS_mpz_set_simple_char_str(z, s);
      *failure = 0;
      return;
      }
   //If neither of those worked, try to parse it as
   //something containing scientific notation.
   else
      {
      GMP_INTS_mpz_set_sci_not_num(z, failure, s);

      if (!*failure)
         {
         //We were able to parse it that way.
         //Everything is set up, just return.
         return;
         }
      else
         {
         //We're out of options.  All parsing failed.
         GMP_INTS_mpz_set_ui(z, 0);
         *failure = 1;
         return;
         }
      }
   }


void GMP_INTS_mpz_parse_into_uint32(unsigned *result,
                                    int *failure,
                                    char *s)
   {
   GMP_INTS_mpz_struct arb_int;

   //Eyeball the input parameters.
   assert(result != NULL);
   assert(failure != NULL);
   assert(s != NULL);

   //Allocate space for the one arbitrary integer we need.
   GMP_INTS_mpz_init(&arb_int);

   //Try to parse the string into an arbitrary length integer
   //using all methods known to man.
   GMP_INTS_mpz_set_general_int(&arb_int, failure, s);

   //If the parse failed, we must declare failure and return
   //0.
   if (*failure)
      {
      *result = 0;
      *failure = 1;
      }
   else
      {
      //We might have success, but it might be negative or
      //too big.
      if (arb_int.size == 1)
         {
         *result = arb_int.limbs[0];
         *failure = 0;
         }
      else if (arb_int.size == 0)
         {
         *result =  0;
         *failure = 0;
         }
      else
         {
         *result = 0;
         *failure = 1;
         }
      }

   //Deallocate the arbitrary integer.
   GMP_INTS_mpz_clear(&arb_int);
   }


void GMP_INTS_mpz_swap(GMP_INTS_mpz_struct *a, 
                       GMP_INTS_mpz_struct *b)
   {
   GMP_INTS_mpz_struct temp;

   //Eyeball the input parameters.
   assert(a           != NULL);
   assert(a->n_allocd >  0);
   assert(a->limbs    != NULL);
   assert(b           != NULL);
   assert(b->n_allocd >  0);
   assert(b->limbs    != NULL);

   //Make the swap via memory copy.
   memcpy(&temp, a,     sizeof(GMP_INTS_mpz_struct));
   memcpy(a,     b,     sizeof(GMP_INTS_mpz_struct));
   memcpy(b,     &temp, sizeof(GMP_INTS_mpz_struct));
   }


/******************************************************************/
/***  PUBLIC ARITHMETIC FUNCTIONS   *******************************/
/******************************************************************/

//07/15/01:  Unit test and visual inspection passed.
void GMP_INTS_mpz_add (      GMP_INTS_mpz_struct *w, 
                       const GMP_INTS_mpz_struct *u, 
                       const GMP_INTS_mpz_struct *v)
   {
   GMP_INTS_limb_srcptr up, vp;
   GMP_INTS_limb_ptr wp;
   GMP_INTS_size_t usize, vsize, wsize;
   GMP_INTS_size_t abs_usize;
   GMP_INTS_size_t abs_vsize;

   //Look at the input parameters carefully.
   assert(w != NULL);
   assert(u != NULL);
   assert(v != NULL);
   assert(w->n_allocd > 0);
   assert(u->n_allocd > 0);
   assert(v->n_allocd > 0);
   assert(w->limbs != NULL);
   assert(u->limbs != NULL);
   assert(v->limbs != NULL);

   //Handle the case of a tainted result.  If either of the
   //two inputs are either direct overflows or tainted by
   //an overflow, mark the result tainted and do not perform
   //any arithmetic operation.
      {
      int taint;

      taint = GMP_INTS_two_op_flags_map(u->flags, v->flags);

      w->flags = 0; 
         //"w" starts off with a clean slate.  Must do this
         //after taint calculation in case locations of u or v 
         //are the same as w.
      if (taint)
         {
         w->flags = taint;
         return;
         }
      }

   usize = u->size;
   vsize = v->size;
   abs_usize = GMP_INTS_abs_of_size_t(usize);
   abs_vsize = GMP_INTS_abs_of_size_t(vsize);

   //Arrange things so that U has at least as many
   //limbs as V, i.e. limbs(U) >= limbs(V);
   if (abs_usize < abs_vsize)
      {
      const GMP_INTS_mpz_struct *tmp_ptr;
      GMP_INTS_size_t      tmp_size;

      //Swap U and V.  This does no harm, because we are
      //manipulating only local variables.  This does not
      //affect the caller.
      tmp_ptr = u;
      u = v;
      v = tmp_ptr;
      tmp_size = usize;
      usize = vsize;
      vsize = tmp_size;
      tmp_size = abs_usize;
      abs_usize = abs_vsize;
      abs_vsize = tmp_size;
      }

   /* True: ABS_USIZE >= ABS_VSIZE.  */

   /* If not space for w (and possible carry), increase space.  */
   wsize = abs_usize + 1;
   if (w->n_allocd < wsize)
      GMP_INTS_mpz_realloc(w, wsize);

   //These pointers must be obtained after realloc.  At this point,
   //u or v may be the same as w.
   up = u->limbs;
   vp = v->limbs;
   wp = w->limbs;

   if ((usize ^ vsize) < 0)
      {
      //U and V have different sign.  Need to compare them to determine
      //which operand to subtract from which.

      //This test is right since ABS_USIZE >= ABS_VSIZE.
      //If the equality case is ruled out, then U has more limbs
      //than V, which means that it is bigger in magnitude.
      if (abs_usize != abs_vsize)
         {
         GMP_INTS_mpn_sub (wp, up, abs_usize, vp, abs_vsize);
         wsize = abs_usize;

         //Normalize the result.  This was formerly a macro.
         //To normalize in this context means to trim the size
         //down to eliminate any leading zero limbs that came
         //about because the size of the result of an operation
         //was overestimated.
         GMP_INTS_mpn_normalize(wp, &wsize);

         if (usize < 0)
            wsize = -wsize;
         }
      else if (GMP_INTS_mpn_cmp (up, vp, abs_usize) < 0)
         {
         GMP_INTS_mpn_sub_n (wp, vp, up, abs_usize);
         wsize = abs_usize;
         GMP_INTS_mpn_normalize(wp, &wsize);
         if (usize >= 0)
            wsize = -wsize;
         }
      else
         {
         GMP_INTS_mpn_sub_n (wp, up, vp, abs_usize);
         wsize = abs_usize;
         GMP_INTS_mpn_normalize(wp, &wsize);
         if (usize < 0)
            wsize = -wsize;
         }
      }
   else
      {
      /* U and V have same sign.  Add them.  */
      GMP_INTS_limb_t cy_limb 
         = GMP_INTS_mpn_add (wp, up, abs_usize, vp, abs_vsize);
      wp[abs_usize] = cy_limb;
      wsize = abs_usize + cy_limb;
      if (usize < 0)
         wsize = -wsize;
      }

   w->size = wsize;

   //Handle the case of an overflowed result.  If the result
   //of the addition is too big or too small, mark it as
   //overflowed.
   if (w->size > GMP_INTS_MAXIMUM_LIMBS_PER_INT)
      {
      w->flags = GMP_INTS_EF_INTOVF_POS;
      }
   else if (w->size < -GMP_INTS_MAXIMUM_LIMBS_PER_INT)
      {
      w->flags = GMP_INTS_EF_INTOVF_NEG;
      }
   }


//07/15/01:  Unit testing skipped because of recursive
//nature.  Visual inspection OK.
void GMP_INTS_mpz_add_ui (      GMP_INTS_mpz_struct *w, 
                          const GMP_INTS_mpz_struct *u, 
                          unsigned long int          v)
   {
   //The GNU MP version of this is quite efficient, and this
   //makes sense since it is a common operation.  However,
   //for simplicity just define this recursively in terms
   //of the ADD function.  This can always be made quicker
   //later (by changing back to the GNU MP version).
   GMP_INTS_mpz_struct temp;

   //Eyeball the inputs carefully.
   assert(w != NULL);
   assert(w->n_allocd > 0);
   assert(w->limbs != NULL);
   assert(u != NULL);
   assert(u->n_allocd > 0);
   assert(u->limbs != NULL);

   //Create a temporary integer.
   GMP_INTS_mpz_init(&temp);

   //Set the temporary integer to the value of the input
   //argument.
   GMP_INTS_mpz_set_ui(&temp, v);

   //Do the actual addition.  This recursive definition
   //is inherently wasteful, but I'm after clarity, not
   //warp speed.
   GMP_INTS_mpz_add(w, u, &temp);

   //Destroy the temporary integer (this will reclaim the
   //memory).
   GMP_INTS_mpz_clear(&temp);
   }


//07/15/01:  Visual inspection passed.  Not unit tested
//because of symmetry with GMP_INTS_mpz_add().
void GMP_INTS_mpz_sub (      GMP_INTS_mpz_struct *w, 
                       const GMP_INTS_mpz_struct *u, 
                       const GMP_INTS_mpz_struct *v)
   {
   GMP_INTS_limb_srcptr up, vp;
   GMP_INTS_limb_ptr wp;
   GMP_INTS_size_t usize, vsize, wsize;
   GMP_INTS_size_t abs_usize;
   GMP_INTS_size_t abs_vsize;

   //Look at the input parameters carefully.
   assert(w != NULL);
   assert(u != NULL);
   assert(v != NULL);
   assert(w->n_allocd > 0);
   assert(u->n_allocd > 0);
   assert(v->n_allocd > 0);
   assert(w->limbs != NULL);
   assert(u->limbs != NULL);
   assert(v->limbs != NULL);

   //Handle the case of a tainted result.  If either of the
   //two inputs are either direct overflows or tainted by
   //an overflow, mark the result tainted and do not perform
   //any arithmetic operation.
      {
      int taint;

      taint = GMP_INTS_two_op_flags_map(u->flags, v->flags);

      w->flags = 0; 
         //"w" starts off with a clean slate.  Must do this
         //after taint calculation in case locations of u or v 
         //are the same as w.
      if (taint)
         {
         w->flags = taint;
         return;
         }
      }

   usize = u->size;
   vsize = -(v->size);  /* The "-" makes the difference from mpz_add */
   abs_usize = GMP_INTS_abs_of_size_t(usize);
   abs_vsize = GMP_INTS_abs_of_size_t(vsize);

   if (abs_usize < abs_vsize)
      {
      const GMP_INTS_mpz_struct *tmp_ptr;
      GMP_INTS_size_t      tmp_size;

      //Swap U and V.  This does no harm, because we are
      //manipulating only local variables.  This does not
      //affect the caller.
      tmp_ptr = u;
      u = v;
      v = tmp_ptr;
      tmp_size = usize;
      usize = vsize;
      vsize = tmp_size;
      tmp_size = abs_usize;
      abs_usize = abs_vsize;
      abs_vsize = tmp_size;
      }

   /* True: ABS_USIZE >= ABS_VSIZE.  */

   /* If not space for w (and possible carry), increase space.  */
   wsize = abs_usize + 1;
   if (w->n_allocd < wsize)
      GMP_INTS_mpz_realloc (w, wsize);

   /* These must be after realloc (u or v may be the same as w).  */
   up = u->limbs;
   vp = v->limbs;
   wp = w->limbs;

   if ((usize ^ vsize) < 0)
      {
      //U and V have different sign.  Need to compare them to determine
      //which operand to subtract from which.

      //This test is right since ABS_USIZE >= ABS_VSIZE.
      if (abs_usize != abs_vsize)
         {
         GMP_INTS_mpn_sub (wp, up, abs_usize, vp, abs_vsize);
         wsize = abs_usize;
         GMP_INTS_mpn_normalize(wp, &wsize);
         if (usize < 0)
            wsize = -wsize;
         }
      else if (GMP_INTS_mpn_cmp (up, vp, abs_usize) < 0)
         {
         GMP_INTS_mpn_sub_n (wp, vp, up, abs_usize);
         wsize = abs_usize;
         GMP_INTS_mpn_normalize(wp, &wsize);
         if (usize >= 0)
            wsize = -wsize;
         }
      else
         {
         GMP_INTS_mpn_sub_n (wp, up, vp, abs_usize);
         wsize = abs_usize;
         GMP_INTS_mpn_normalize (wp, &wsize);
         if (usize < 0)
            wsize = -wsize;
         }
      }
   else
      {
      /* U and V have same sign.  Add them.  */
      GMP_INTS_limb_t cy_limb 
         = GMP_INTS_mpn_add (wp, up, abs_usize, vp, abs_vsize);
      wp[abs_usize] = cy_limb;
      wsize = abs_usize + cy_limb;
      if (usize < 0)
         wsize = -wsize;
      }

   w->size = wsize;

   //Handle the case of an overflowed result.  If the result
   //of the addition is too big or too small, mark it as
   //overflowed.
   if (w->size > GMP_INTS_MAXIMUM_LIMBS_PER_INT)
      {
      w->flags = GMP_INTS_EF_INTOVF_POS;
      }
   else if (w->size < -GMP_INTS_MAXIMUM_LIMBS_PER_INT)
      {
      w->flags = GMP_INTS_EF_INTOVF_NEG;
      }
   }


//07/15/01:  Unit testing skipped because of recursive
//nature.  Visual inspection OK.
void GMP_INTS_mpz_sub_ui (      GMP_INTS_mpz_struct *w, 
                          const GMP_INTS_mpz_struct *u, 
                          unsigned long int          v)
   {
   //The GNU MP version of this is quite efficient, and this
   //makes sense since it is a common operation.  However,
   //for simplicity just define this recursively in terms
   //of the SUB function.  This can always be made quicker
   //later (by changing back to the GNU MP version).
   GMP_INTS_mpz_struct temp;

   //Eyeball the inputs carefully.
   assert(w != NULL);
   assert(w->n_allocd > 0);
   assert(w->limbs != NULL);
   assert(u != NULL);
   assert(u->n_allocd > 0);
   assert(u->limbs != NULL);

   //Create a temporary integer.
   GMP_INTS_mpz_init(&temp);

   //Set the temporary integer to the value of the input
   //argument.
   GMP_INTS_mpz_set_ui(&temp, v);

   //Do the actual subtraction.  This recursive definition
   //is inherently wasteful, but I'm after clarity, not
   //warp speed.
   GMP_INTS_mpz_sub(w, u, &temp);

   //Destroy the temporary integer (this will reclaim the
   //memory).
   GMP_INTS_mpz_clear(&temp);
   }


void GMP_INTS_mpz_mul (      GMP_INTS_mpz_struct *w, 
                       const GMP_INTS_mpz_struct *u, 
                       const GMP_INTS_mpz_struct *v)
   {
   GMP_INTS_size_t    usize = u->size;
   GMP_INTS_size_t    vsize = v->size;
   GMP_INTS_size_t    wsize;
   GMP_INTS_size_t    sign_product;
   GMP_INTS_limb_ptr  up, vp;
   GMP_INTS_limb_ptr  wp;
   GMP_INTS_limb_ptr  free_me = NULL;
   GMP_INTS_size_t    free_me_size;
   GMP_INTS_limb_t    cy_limb;

   //Eyeball the inputs.
   assert(w != NULL);
   assert(w->n_allocd > 0);
   assert(w->limbs != NULL);
   assert(u != NULL);
   assert(u->n_allocd > 0);
   assert(u->limbs != NULL);
   assert(v != NULL);
   assert(v->n_allocd > 0);
   assert(v->limbs != NULL);

   //Handle the case of a tainted result.  If either of the
   //two inputs are either direct overflows or tainted by
   //an overflow, mark the result tainted and do not perform
   //any arithmetic operation.
      {
      int taint;

      taint = GMP_INTS_two_op_flags_map(u->flags, v->flags);

      w->flags = 0; 
         //"w" starts off with a clean slate.  Must do this
         //after taint calculation in case locations of u or v 
         //are the same as w.
      if (taint)
         {
         w->flags = taint;
         return;
         }
      }

   sign_product = usize ^ vsize;
   usize = GMP_INTS_abs_of_size_t(usize);
   vsize = GMP_INTS_abs_of_size_t(vsize);

   //Handle the case of a certain result overflow (why do the math when
   //the result is certain?).  In general, when multiplying two inputs
   //whose sizes are M limbs and N limbs, the size of the result will be
   //either M+N or M+N-1 limbs.  If M+N-1 > MAX_ALLOWED, then can declare
   //an early overflow.
   if ((usize + vsize - 1) > GMP_INTS_MAXIMUM_LIMBS_PER_INT)
      {
      if (sign_product < 0)
         w->flags = GMP_INTS_EF_INTOVF_NEG;
      else 
         w->flags = GMP_INTS_EF_INTOVF_POS;

      return;
      }


   if (usize < vsize)
      {
      //Temporary variables just for the swap.
      const GMP_INTS_mpz_struct *tmp_ptr;
      GMP_INTS_size_t            tmp_size;

      //Swap U and V.
      tmp_ptr = u; 
      u = v; 
      v = tmp_ptr;
      tmp_size = usize; 
      usize = vsize; 
      vsize = tmp_size;
      }

   //Grab pointers to the arrays of limbs.
   up = u->limbs;
   vp = v->limbs;
   wp = w->limbs;

   /* Ensure W has space enough to store the result.  */
   wsize = usize + vsize;
   if (w->n_allocd < wsize)
      {
      if (wp == up || wp == vp)
         {
         free_me = wp;
         free_me_size = w->n_allocd;
         }
      else
         {
         GMP_INTS_free_w_size (wp, w->n_allocd * sizeof(GMP_INTS_limb_t));
         }

      w->n_allocd = wsize;
      wp = (GMP_INTS_limb_ptr) 
           GMP_INTS_malloc (wsize * sizeof(GMP_INTS_limb_t));
      w->limbs = wp;
      }
   else
      {
      /* Make U and V not overlap with W.  */
      if (wp == up)
         {
         /* W and U are identical.  Allocate temporary space for U.  */
         up = (GMP_INTS_limb_ptr) 
              _alloca(usize * sizeof(GMP_INTS_limb_t));
         /* Is V identical too?  Keep it identical with U.  */
         if (wp == vp)
            vp = up;
         /* Copy to the temporary space.  */
         GMP_INTS_mpn_copy_limbs(up, wp, usize);
         }
      else if (wp == vp)
         {
         /* W and V are identical.  Allocate temporary space for V.  */
         vp = (GMP_INTS_limb_ptr) 
              _alloca(vsize * sizeof(GMP_INTS_limb_t));
         /* Copy to the temporary space.  */
         GMP_INTS_mpn_copy_limbs(vp, wp, vsize);
         }
      }

   if (vsize == 0)
      {
      wsize = 0;
      }
   else
      {
      cy_limb = GMP_INTS_mpn_mul (wp, up, usize, vp, vsize);
      wsize = usize + vsize;
      wsize -= cy_limb == 0;
      }

   w->size = sign_product < 0 ? -wsize : wsize;

   if (free_me != NULL)
      GMP_INTS_free_w_size (free_me, free_me_size * sizeof(GMP_INTS_limb_t));

   //Final check for overflow.
   if (w->size > GMP_INTS_MAXIMUM_LIMBS_PER_INT)
      w->flags = GMP_INTS_EF_INTOVF_POS;
   else if (w->size < -GMP_INTS_MAXIMUM_LIMBS_PER_INT)
      w->flags = GMP_INTS_EF_INTOVF_NEG;
   }


//07/15/01:  Unit testing skipped because of recursive
//nature.  Visual inspection OK.
void GMP_INTS_mpz_mul_si (      GMP_INTS_mpz_struct *w, 
                          const GMP_INTS_mpz_struct *u, 
                                long int             v)
   {
   GMP_INTS_mpz_struct temp;

   //Eyeball the inputs carefully.
   assert(w != NULL);
   assert(w->n_allocd > 0);
   assert(w->limbs != NULL);
   assert(u != NULL);
   assert(u->n_allocd > 0);
   assert(u->limbs != NULL);

   //Create a temporary integer.
   GMP_INTS_mpz_init(&temp);

   //Set the temporary integer to the value of the input
   //argument.
   GMP_INTS_mpz_set_si(&temp, v);

   //Do the actual multiplication.  This recursive definition
   //is inherently wasteful, but I'm after clarity, not
   //warp speed.
   GMP_INTS_mpz_mul(w, u, &temp);

   //Destroy the temporary integer (this will reclaim the
   //memory).
   GMP_INTS_mpz_clear(&temp);
   }


//07/15/01:  Unit testing skipped because of recursive
//nature.  Visual inspection OK.
void GMP_INTS_mpz_mul_ui (      GMP_INTS_mpz_struct *w, 
                          const GMP_INTS_mpz_struct *u, 
                          unsigned long int          v)
   {
   GMP_INTS_mpz_struct temp;

   //Eyeball the inputs carefully.
   assert(w != NULL);
   assert(w->size >= 0);
   assert(w->limbs != NULL);
   assert(u != NULL);
   assert(u->size >= 0);
   assert(u->limbs != NULL);

   //Create a temporary integer.
   GMP_INTS_mpz_init(&temp);

   //Set the temporary integer to the value of the input
   //argument.
   GMP_INTS_mpz_set_ui(&temp, v);

   //Do the actual multiplication.  This recursive definition
   //is inherently wasteful, but I'm after clarity, not
   //warp speed.
   GMP_INTS_mpz_mul(w, u, &temp);

   //Destroy the temporary integer (this will reclaim the
   //memory).
   GMP_INTS_mpz_clear(&temp);
   }


void GMP_INTS_mpz_tdiv_qr (      GMP_INTS_mpz_struct *quot, 
                                 GMP_INTS_mpz_struct *rem, 
                           const GMP_INTS_mpz_struct *num, 
                           const GMP_INTS_mpz_struct *den)
   {
   GMP_INTS_size_t   abs_num_size, 
                     abs_den_size, 
                     quotient_sign, 
                     remainder_sign,
                     numerator_bitsize, 
                     denominator_bitsize,
                     division_loop_count,
                     division_loop_count_mod_32,
                     division_loop_count_div_32,
                     division_counter,
                     i;
   GMP_INTS_limb_t   temp_limb;
   GMP_INTS_limb_ptr trial_divisor;

   //Eyeball the input parameters.
   assert(quot != NULL);
   assert(quot->n_allocd > 0);
   assert(quot->limbs != NULL);
   assert(rem != NULL);
   assert(rem->n_allocd > 0);
   assert(rem->limbs != NULL);
   assert(num != NULL);
   assert(num->n_allocd > 0);
   assert(num->limbs != NULL);
   assert(den != NULL);
   assert(den->n_allocd > 0);
   assert(den->limbs != NULL);

   //We require for this function that the numerator, denominator, quotient, and
   //remainder all be distinct.
   assert(quot != rem);
   assert(quot != num);
   assert(quot != den);
   assert(rem  != num);
   assert(rem  != den);
   assert(num  != den);

   //The GNU code was probably very efficient, but exceeded
   //my abilities to analyze.  This is the classic
   //division algorithm.

   //First, start off with the quotient and remainder having
   //no error flags set.  These will be set if appropriate.
   quot->flags = 0; 
   rem->flags  = 0;

   //First, handle tainted inputs.  If the numerator or denominator
   //are bad or tainted, the quotient and remainder get tainted
   //automatically.
      {
      int taint;

      taint = GMP_INTS_two_op_flags_map(num->flags, den->flags);

      if (taint)
         {
         quot->flags = taint;
         rem->flags  = taint;
         return;
         }
      }

   //The second possible cause for taint is if the divisor is
   //zero.  This will get both the value of positive overflow.
   if (den->size == 0)
      {
      quot->flags = GMP_INTS_EF_INTOVF_POS;
      rem->flags  = GMP_INTS_EF_INTOVF_POS;
      return;
      }

   //Handle the special case of a numerator of zero.  If the numerator
   //is zero, the quotient and remainder are zero automatically.
   if (num->size == 0)
      {
      GMP_INTS_mpz_set_ui(quot, 0);
      GMP_INTS_mpz_set_ui(rem,  0);
      return;
      }

   //Generally, nothing else can go wrong as far as taint.  The
   //value of the quotient is confined to be no larger than the
   //numerator, and the value of the remainder is confined to
   //be no larger than denominator-1.  So, generally, if the
   //inputs are in size bounds, the outputs will be also.

   //Figure out how large in limbs the numerator and denominator actually
   //are.
   abs_num_size = GMP_INTS_abs_of_size_t(num->size);
   abs_den_size = GMP_INTS_abs_of_size_t(den->size);

   //Figure out the sign of things.  We want the following relationship
   //to be true:
   //   num/den = quot + rem/den.
   //The way to achieve this is to assign the sign of the quotient in the traditional
   //way, then to assign the remainder to have the same sign as the numerator.
   quotient_sign  = num->size ^ den->size;
   remainder_sign = num->size;

   //The remainder starts off with the absolute value of the numerator, and then
   //we subtract from it as part of the division loop.
   GMP_INTS_mpz_copy(rem, num);
      //We know after the copy that the amount of space allocated in the remainder
      //MUST be at least as large as the absolute value of the numerator.  So from
      //this point forward we use the space.
   assert(rem->n_allocd >= abs_num_size);

   //Figure out the number of significant bits in the numerator and denominator.
   //This determines the loop count over which we do the shift division loop.
   numerator_bitsize = GMP_INTS_BITS_PER_LIMB * abs_num_size;
   
   i = abs_num_size - 1;

   //We need to be extra careful here.  One failure mode is that an integer
   //data structure is corrupted and the "size" field reflects limbs
   //that are zero.  Need to watch that this kind of failure doesn't
   //cause memory access errors.
   assert(num->limbs[i] != 0);
   if (num->limbs[i] == 0)
      {
      quot->flags = GMP_INTS_EF_INTOVF_POS;
      rem->flags  = GMP_INTS_EF_INTOVF_POS;
      return;
      }

   temp_limb = 0x80000000;

   while (((num->limbs[i] & temp_limb) == 0) && (temp_limb != 0))
      {
      numerator_bitsize--;
      temp_limb >>= 1;
      }

   denominator_bitsize = GMP_INTS_BITS_PER_LIMB * abs_den_size;

   i = abs_den_size - 1;

   //We need to be extra careful here.  One failure mode is that an integer
   //data structure is corrupted and the "size" field reflects limbs
   //that are zero.  Need to watch that this kind of failure doesn't
   //cause memory access errors.
   assert(den->limbs[i] != 0);
   if (den->limbs[i] == 0)
      {
      quot->flags = GMP_INTS_EF_INTOVF_POS;
      rem->flags  = GMP_INTS_EF_INTOVF_POS;
      return;
      }

   temp_limb = 0x80000000;

   while (((den->limbs[i] & temp_limb) == 0) && (temp_limb != 0))
      {
      denominator_bitsize--;
      temp_limb >>= 1;
      }

   //The quotient starts off with the value of zero, but we consistently may 
   //mask 1 into it and shift left.  We need to be sure that we have as much
   //shift space there as is in the numerator.  For this purpose we need to 
   //prepare a block of clear memory as large as the numerator's.
   if (quot->n_allocd < abs_num_size)
      {
      GMP_INTS_mpz_realloc(quot, abs_num_size);  //Make it big enough.
      }
   //Now, zero the memory.
   for (i=0; i<abs_num_size; i++)
      quot->limbs[i] = 0;

   //Determine the division loop count.  This is the difference
   //in bit sizes between the numerator and denominator.  It is 
   //possible for this number to be negative, which means that the
   //main division loop will be executed zero times.  This gives the
   //right results.
   division_loop_count = numerator_bitsize - denominator_bitsize;

   //We need to calculate some important numbers from the division loop
   //count.  We need to know this number MOD 32 (which tells how far to
   //shift the divisor bitwise to line up with the numerator), and we 
   //also need this number DIV 32 for the limb-wise shift.
   division_loop_count_mod_32 = division_loop_count % 32;
   division_loop_count_div_32 = division_loop_count / 32;
   
   //We now need a shift register in which we shift the denominator up
   //for repeated comparisons.  We should dynamically allocate this to
   //be the same size as the numerator.  Using _alloca() is OK, as one
   //of the unit tests is to be sure that _alloca() will handle integer
   //of the maximum allowed size.
   trial_divisor = _alloca(abs_num_size * sizeof(GMP_INTS_limb_t));

   //Our trial divisor needs to start off with the divisor shifted up
   //so that the most significant bit is aligned with the numerator.
   for (i = 0; i < abs_num_size; i++)
      {
      if ((division_loop_count < 0) || (i < division_loop_count_div_32))
         {
         trial_divisor[i] = 0;
         }
      else
         {
         if ((i-division_loop_count_div_32) < abs_den_size)
            trial_divisor[i] = den->limbs[i - division_loop_count_div_32];
         else
            trial_divisor[i] = 0;
         }
      }

   //The code above planted the limbs in the right place.  Now need to shift bits
   //upward by the remaining number.
   if ((division_loop_count > 0) && (division_loop_count_mod_32 > 0))
      {
      //There is an existing function we can call to do the left shift.
      GMP_INTS_mpn_lshift(trial_divisor, 
                          trial_divisor, 
                          abs_num_size, 
                          division_loop_count_mod_32);
      }


   //Everything is ready to go.  Now begin the division loop itself.  It is possible
   //for the loop to execute zero times, which will happen if the denominator is longer
   //in bits than the numerator.  In such cases, we can't execute this loop even once
   //because the math assumes that the numerator is at least as long as the denominator.
   for (division_counter = 0; division_counter < division_loop_count+1; division_counter++)
      {
      //Shift the quotient left one bit.
      GMP_INTS_mpn_lshift(quot->limbs, 
                          quot->limbs, 
                          abs_num_size, 
                          1);

      //If the remainder is at least as large as the trial divisor, subtract the trial
      //divisor from the remainder and mask in the quotient.
      if (GMP_INTS_mpn_cmp(rem->limbs, 
                           trial_divisor,
                           abs_num_size) >= 0)
         {
         GMP_INTS_mpn_sub(rem->limbs,
                          rem->limbs,
                          abs_num_size,
                          trial_divisor,
                          abs_num_size);
         quot->limbs[0] |= 1;
         }

      //Shift the trial divisor right one bit.
      GMP_INTS_mpn_rshift(trial_divisor, 
                          trial_divisor, 
                          abs_num_size, 
                          1);
      }  //End for each iteration of the division loop.

   //Normalize the quotient and the remainder.  The normalization
   //process is to bring the sizes down if we have leading
   //zeros.
   quot->size = abs_num_size;
   GMP_INTS_mpn_normalize(quot->limbs, &(quot->size));
   rem->size  = abs_num_size;
   GMP_INTS_mpn_normalize(rem->limbs, &(rem->size));

   //Adjust the signs as required.
   if (quotient_sign < 0)
      quot->size = -(quot->size);
   if (remainder_sign < 0)
      rem->size = -(rem->size);
   }


void GMP_INTS_mpz_fac_ui(GMP_INTS_mpz_struct *result, 
                         unsigned long int n)
   {
   //Just multiply the numbers in ascending order.  The original
   //GNU library contained a much more elegant algorithm, but
   //this is more direct.

   unsigned long int k;

   GMP_INTS_mpz_set_ui (result, 1L);

   for (k = 2; (k <= n) && !(result->flags); k++)
      GMP_INTS_mpz_mul_ui (result, result, k);
   }


/******************************************************************/
/***  PUBLIC CONVERSION AND OUTPUT FUNCTIONS   ********************/
/******************************************************************/
//07/18/01:  Visual inspection OK.  Function returns
//reasonable values even out to 100,000 digits--seems OK.
int GMP_INTS_mpz_size_in_base_10(const GMP_INTS_mpz_struct *arg)
   {
   _int64 n;

   //Eyeball the input parameter.
   assert(arg != NULL);
   assert(arg->n_allocd > 0);
   assert(arg->limbs != NULL);

   //Get the number of limbs occupied by the integer.
   //Because even the digit zero takes some space,
   //don't accept zero for an answer.
   n = GMP_INTS_abs_of_size_t(arg->size);
   if (n==0)
      n = 1;
   
   //Convert this to the number of bits.  Generously 
   //ignore any unused leading bits.
   n *= 32;

   //Used a slightly high best rational approximation in F_{65535}
   //to go from the number of bits to the number of 
   //digits.  The division discards, so bump the result
   //up by 1 to compensate for possible truncation.  The number
   //we are aproximating is ln(2)/ln(10).
   n *= 12655;
   n /= 42039;
   n++;

   //Compensate for possible commas in the result.  Again,
   //consider truncation.
   n *= 4;
   n /= 3;
   n++;

   //Compensate for the minus sign, the trailing zero,
   //cosmic rays striking the computer from the martian
   //listening post camoflaged on the moon, and the
   //possibility that we might need to put text in the
   //string if any flag is set.
   n += 100;

   //And that should be a good return value.
   return((int) n);
   }


//07/19/01:  Visual inspection and unit test is OK.
void GMP_INTS_mpz_to_string(char *out,
                            const GMP_INTS_mpz_struct *in)
   {
   //Eyeball the input parameters.
   assert(out != NULL);
   assert(in != NULL);
   assert(in->n_allocd > 0);
   assert(in->limbs != NULL);

   //If any of the flags are set, stuff in the text.
   if (in->flags)
      {
      if (in->flags & GMP_INTS_EF_INTOVF_POS)
         {
         strcpy(out, GMP_INTS_EF_INTOVF_POS_STRING);
         }
      else if (in->flags & GMP_INTS_EF_INTOVF_NEG)
         {
         strcpy(out, GMP_INTS_EF_INTOVF_NEG_STRING);
         }
      else if (in->flags & GMP_INTS_EF_INTOVF_TAINT_POS)
         {
         strcpy(out, GMP_INTS_EF_INTOVF_TAINT_POS_STRING);
         }
      else if (in->flags & GMP_INTS_EF_INTOVF_TAINT_NEG)
         {
         strcpy(out, GMP_INTS_EF_INTOVF_TAINT_NEG_STRING);
         }
      else
         {
         strcpy(out, "INTERNAL_ERROR");
         }
      }  
   else
      {
      //Ordinary integer conversion.
      GMP_INTS_mpz_struct num, den, quot, rem, k10;

      //Allocate space for the temporary integers.
      GMP_INTS_mpz_init(&num);
      GMP_INTS_mpz_init(&den);
      GMP_INTS_mpz_init(&quot);
      GMP_INTS_mpz_init(&rem);
      GMP_INTS_mpz_init(&k10);

      //Assign the constant 10.
      GMP_INTS_mpz_set_ui(&k10, 10);

      //If the integer is zero, assign that.
      if (in->size == 0)
         {
         strcpy(out, "0");
         }
      else
         {
         //We have to do a full conversion.  The algorithm
         //is division by 10, each time obtaining the least
         //significant digit, until finally the quotient is 
         //zero.
         char *ptr;

         ptr = out;

         GMP_INTS_mpz_copy(&num, in);
         GMP_INTS_mpz_copy(&den, &k10);
         do
            {
#if 0
            printf("Values before division:\n");
            FCMIOF_hline();
            GMP_INTS_mpz_print_int(stdout, &num, "Numerator");
            FCMIOF_hline();
            GMP_INTS_mpz_print_int(stdout, &den, "Denominator");
            FCMIOF_hline();
            GMP_INTS_mpz_print_int(stdout, &quot, "Quotient");
            FCMIOF_hline();
            GMP_INTS_mpz_print_int(stdout, &rem, "Remainder");
            FCMIOF_hline();

if (num.size > 1)
            FCMIOF_hline();
#endif

            GMP_INTS_mpz_tdiv_qr(&quot, &rem, &num, &den);
#if 0
            printf("Values after division:\n");
            FCMIOF_hline();
            GMP_INTS_mpz_print_int(stdout, &num, "Numerator");
            FCMIOF_hline();
            GMP_INTS_mpz_print_int(stdout, &den, "Denominator");
            FCMIOF_hline();
            GMP_INTS_mpz_print_int(stdout, &quot, "Quotient");
            FCMIOF_hline();
            GMP_INTS_mpz_print_int(stdout, &rem, "Remainder");
            FCMIOF_hline();
#endif

            if (rem.size != 0)
               {
               *ptr = '0' + (char)(rem.limbs[0]);
               }
            else
               {
               *ptr = '0';
               }
            ptr++;
            GMP_INTS_mpz_copy(&num, &quot);
            //printf("digit\n");
            }
         while (!GMP_INTS_mpz_is_zero(&quot));

         //Finally, if the input was negative, tack on the 
         //minus sign.
         if (GMP_INTS_mpz_is_neg(in))
            {
            *ptr = '-';
            ptr++;
            }
        
         //Finally, tack on the trailing zero terminator.
         *ptr = 0;
         ptr++;

         //Reverse the string.
         BSTRFUNC_str_reverse(out);
         }
 
      //Deallocate the integers.
      GMP_INTS_mpz_clear(&num);
      GMP_INTS_mpz_clear(&den);
      GMP_INTS_mpz_clear(&quot);
      GMP_INTS_mpz_clear(&rem);
      GMP_INTS_mpz_clear(&k10);
      } 
   }


void GMP_INTS_mpz_long_int_format_to_stream(FILE *s,
                                            const GMP_INTS_mpz_struct *i,
                                            const char *desc)
   {
   int line_len;
   int digits_per_line;
   char *digits;
   int num_digits;
   int nlines;
   int cur_line;
   int number_desc_width;

   //Eyeball the inputs, make sure the caller isn't doing
   //something stupid.
   assert(s != NULL);
   assert(i != NULL);
   assert(i->n_allocd > 0);
   assert(i->limbs != NULL);
   assert(desc != NULL);

   //Obtain the line length assumed for formatted output.
   line_len = FCMIOF_get_line_len();

   //The description width allowed is 20.
   number_desc_width = 20;

   /* The number of digits per line that we assume must be a multiple of
   ** three.  The formula below was not examined very carefully, but it
   ** works fine for a line length of 78.  If line length is changed,
   ** this formula may need to be examined very carefully and rewritten.
   */
   digits_per_line = INTFUNC_max(3, ((((line_len-42)*3)/4)/3)*3);
   assert(digits_per_line >= 3);

   /* We now need to get a digit string corresponding to this 
   ** number.  First, need to figure out how much and
   ** allocate the space.
   */
   digits = GMP_INTS_malloc(GMP_INTS_mpz_size_in_base_10(i) * sizeof(char));
   GMP_INTS_mpz_to_string(digits, i);

   //If the number is negative, delete the leading minus sign.
   //The rest of the display algorithm needs an unsigned
   //series of digits.
   if (*digits == '-')
      {
      int i = 0;

      do
         {
         digits[i] = digits[i+1];
         i++;
         }
      while(digits[i-1]);
      }

   //Figure out how many digits in the string representation.
   num_digits = strlen(digits);

   /* As the first order of business, figure out how many lines the beast
   ** will require.
   */
   if (i->flags)
      {
      nlines = 1;  /* Only one line required for NAN verbeage. */
      }
   else if (GMP_INTS_mpz_is_zero(i))
      {
      nlines = 1;  /* The zero value requires one line. */
      }
   else
      {
      /* In any other case, have a formula.
      */
      nlines = 1 + (num_digits - 1) / digits_per_line;
      }

   /* Iterate through each line, spitting out whatever is appropriate. */
   for (cur_line = 0; cur_line < nlines; cur_line++)
      {
      int cur_digit_on_line;

      /* If this is the first line, spit out the description, right-aligned.
      ** Otherwise, spit spaces.
      */
      if (!cur_line)
         {
         /* First line. */
         int len;

         len = strlen(desc);

         if (len <= number_desc_width)
            {
            /* Description is shorter or equal, pad on left. */
            FCMIOF_stream_repchar(s, ' ', number_desc_width - len);
            fprintf(s, "%s", desc);
            }
         else
            {
            /* Description is too long, truncate. */
            int i;

            for (i=0; i<number_desc_width; i++)
               fprintf(s, "%c", desc[i]);
            }

         fprintf(s, ": ");

         /* If the number is negative, throw in a minus sign. */
         if (GMP_INTS_mpz_is_neg(i) && !(i->flags))
            {
            fprintf(s, "- ");
            }
         else
            {
            fprintf(s, "  ");
            }
         }
      else
         {
         /* Every line but first line. */
         FCMIOF_stream_repchar(s, ' ', number_desc_width+4);
         }

      for(cur_digit_on_line=0; cur_digit_on_line < digits_per_line; cur_digit_on_line++)
         {
         int idx_into_string;
            /* Index into the string which is our digit of interest.
            */

         /* Compute the index.  The equation is based on the ordering
         ** of presentation, for example,
         **
         **            7 6
         **          5 4 3
         **          2 1 0.
         **
         ** With a little thought, the equation should make sense.
         ** The index won't always be used to index into the string.
         */
         idx_into_string = 
           ((((nlines-1) - cur_line) * digits_per_line)
           +
           (digits_per_line - 1 - cur_digit_on_line));

         /* Print the appropriate digit or a space.  The NAN case and the 
         ** zero case need to be treated specially.
         */
         if (i->flags)
            {
            /* Not a number.  Everything is blank, except spell out
            ** description of condition at the end of the string of 
            ** digits.
            */
            int index_from_right;
            int virtual_index;

            index_from_right = digits_per_line - 1 - cur_digit_on_line;
               //The index calculated above is calculated so that the
               //final position on the line has index [0].
            assert(index_from_right >= 0 && index_from_right < digits_per_line);

            //Now, calculate the "virtual index".  The virtual index
            //is the actual number of characters from the right, taking
            //into account commas.
            virtual_index = index_from_right + index_from_right/3;

            if (((index_from_right % 3) == 2) && cur_digit_on_line)
               {
               //We are one position past a comma.  This means
               //that we might need a "fill" character to go
               //where the comma should have gone.

               if (virtual_index + 1 < num_digits)
                  {
                  //The character we should print exists.
                  fprintf(s, "%c", digits[num_digits - 2 - virtual_index]);
                  }
               else
                  {
                  //The character doesn't exist, because the error
                  //string is apparently too short.  Must print a
                  //space, instead.
                  fprintf(s, " ");
                  }
               }

            //We've done the fill character, if the position we're in
            //is one past a comma.  Now, do the ordinary character
            //corresponding to a digit position. 
            if (virtual_index < num_digits)
               {
               //The character we should print exists.
               fprintf(s, "%c", digits[num_digits - 1 - virtual_index]);
               }
            else
               {
               //The character doesn't exist, because the error
               //string is apparently too short.  Must print a
               //space, instead.
               fprintf(s, " ");
               }
            }
         else if (GMP_INTS_mpz_is_zero(i))
            {
            /* This is the zero case.  For zero, there is only one line,
            ** and every character except the last one is a blank.
            */
            if (cur_digit_on_line == (digits_per_line - 1))
               {
               fprintf(s, "0");
               }
            else
               {
               fprintf(s, " ");
               }
            }
         else
            {
            /* This is a valid number which is not zero.  Need to print
            ** the digits.
            */

            if (idx_into_string < num_digits)
               {
               int actual_index;

               actual_index = num_digits - 1 - idx_into_string;
                  //This is a string reversal mapping.  The original
                  //code stored strings least significant digit first,
                  //but this code uses most significant digit first.
               assert((actual_index >= 0) && (actual_index < num_digits));
               fprintf(s, "%c", digits[actual_index]);
               }
            else
               {
               fprintf(s, " ");
               }
            } /* End of digit case.

         /* Now handle the commas.  The rules for commas are straightforward.
         **    a)NAN never has a comma.
         **    b)Zeros never have a comma.
         **    c)The final line, last digit never has a comma.
         **    d)Everything else in multiples of three ...
         */
         if (!(idx_into_string % 3) && (idx_into_string))
            {
            if (i->flags)
               {
               //fprintf(s, " ");
               }
            else if (!num_digits)
               {
               fprintf(s, " ");
               }
            else
               {
               if (idx_into_string < num_digits)
                  {
                  fprintf(s, ",");
                  }
               else
                  {
                  fprintf(s, " ");
                  }
               }
            }
         } /* End for each digit on the current line. */
      
      /* For the first line, print out an informative message
      ** advising of the number of digits.  For all other lines
      ** print nothing.
      */
      if (!cur_line && !(i->flags))
         {
         if (nlines == 1)
            fprintf(s, " ");

         if (num_digits <= 1)
            {
            fprintf(s, "  (      1 digit )\n");
            }
         else if (num_digits < 1000)
            {
            fprintf(s, "  (%7d digits)\n", num_digits);
            }
         else
            {
            fprintf(s, "  (%3d,%03d digits)\n", num_digits / 1000, num_digits % 1000);
            }
         }
      else
         {
         fprintf(s, "\n");
         }
      } /* End for each line. */

   //Deallocate the string space.
   GMP_INTS_free(digits);
   }


void GMP_INTS_mpz_arb_int_raw_to_stream(FILE *s,
                                        const GMP_INTS_mpz_struct *i)
   {
   int size_reqd;
   char *digits;

   //Eyeball the input parameters.
   assert(s != NULL);
   assert(i != NULL);
   assert(i->n_allocd > 0);
   assert(i->limbs != NULL);

   size_reqd = GMP_INTS_mpz_size_in_base_10(i);
   digits = GMP_INTS_malloc(size_reqd * sizeof(char));
   GMP_INTS_mpz_to_string(digits, i);
   fprintf(s, "%s", digits);
   GMP_INTS_free(digits);
   }


//07/24/01:  Passed visual inspection and unit tests.
void GMP_INTS_mpz_pow_ui(         GMP_INTS_mpz_struct *result,
                         const    GMP_INTS_mpz_struct *base,
                         unsigned                      exponent)
   {
   GMP_INTS_mpz_struct temp;
      //Temporary location to hold the base raised to
      //a binary power (repeated squaring).

   //Eyeball the input parameters.
   assert(result != NULL);
   assert(result->n_allocd > 0);
   assert(result->limbs != NULL);
   assert(base != NULL);
   assert(base->n_allocd > 0);
   assert(base->limbs != NULL);

   //For this function, the base and the result may not
   //be the same object.
   assert(result != base);

   //If the base is tained, the output is tainted by association.
      {
      int taint;

      taint = GMP_INTS_two_op_flags_map(base->flags, 0);

      if (taint)
         {
         result->flags = taint;
         return;
         }
      }

   //Allocate our temporary variable and set it to the base.
   GMP_INTS_mpz_init(&temp);
   GMP_INTS_mpz_copy(&temp, base);

   //The result begins with the value of 1.
   GMP_INTS_mpz_set_ui(result, 1);

   //Loop through, processing each bit of the exponent.  This is a fairly effective
   //algorithm, but not the optimal one (Knuth points this out).
   while (exponent && !result->flags)
      {
      if (exponent & 0x1)
         {
         GMP_INTS_mpz_mul(result, result, &temp);
         }

      //Square the temporary variable.  Because squaring of arb integer
      //may be very expensive, the test against 1 (i.e. last iteration)
      //certainly pays for itself.
      if (exponent != 1)
         GMP_INTS_mpz_mul(&temp, &temp, &temp);
     
      exponent >>= 1;
      }

   //Deallocate our temporary variable.
   GMP_INTS_mpz_clear(&temp);
   }


void GMP_INTS_mpz_abs(GMP_INTS_mpz_struct *arg)
   {
   //Eyeball the input parameter.
   assert(arg != NULL);
   assert(arg->n_allocd > 0);
   assert(arg->limbs != NULL);

   //Take the absolute value.
   if (arg->size < 0)
      arg->size = -arg->size;
   }


//07/29/01:  Visual inspection passed.  Seems to work fine--not explicitly unit-tested
//directly, but was tested from Tcl.
void GMP_INTS_mpz_gcd(GMP_INTS_mpz_struct       *result,
                      const GMP_INTS_mpz_struct *arg1,
                      const GMP_INTS_mpz_struct *arg2)
   {
   GMP_INTS_mpz_struct u, v, q, r;
   int loop_count;

   //Eyeball the inputs carefully.
   assert(result != NULL);
   assert(result->n_allocd > 0);
   assert(result->limbs != NULL);
   assert(arg1 != NULL);
   assert(arg1->n_allocd > 0);
   assert(arg1->limbs != NULL);
   assert(arg2 != NULL);
   assert(arg2->n_allocd > 0);
   assert(arg2->limbs != NULL);

   //Args are not allowed to be same object.
   assert(arg1 != arg2);

   //If either input is error or taint, taint the output.
      {
      int taint;

      taint = GMP_INTS_two_op_flags_map(arg1->flags, arg2->flags);

      result->flags = 0; 
         //"result" starts off with a clean slate.  Must do this
         //after taint calculation in case locations of arg1 or arg2 
         //are the same as result.
      if (taint)
         {
         result->flags = taint;
         return;
         }
      }

   //If either input is zero, the result is 1.
   if (GMP_INTS_mpz_is_zero(arg1) || GMP_INTS_mpz_is_zero(arg2))
      {
      GMP_INTS_mpz_set_ui(result, 1);
      return;
      }

   //Allocate space for locals.
   GMP_INTS_mpz_init(&u);
   GMP_INTS_mpz_init(&v);
   GMP_INTS_mpz_init(&q);
   GMP_INTS_mpz_init(&r);

   //We are following Knuth Vol 2, p. 337, the modern Euclidian algorithm.
   //Note:  There are faster algorithms for GCD, but because I hacked up the
   //GMP multiple-precision library so badly, those aren't included.  This one
   //is logically correct but sub-optimal.  Perhaps at a later time faster 
   //algorithms will be re-included.
   //Copy inputs to u and v.
   GMP_INTS_mpz_copy(&u, arg1);
   GMP_INTS_mpz_copy(&v, arg2);
   
   //Take the absolute value of each argument.  We know that neither is zero, 
   //but one or both might be negative.
   GMP_INTS_mpz_abs(&u);
   GMP_INTS_mpz_abs(&v);

   //Begin Euclid's algorithm.  There are really three possibilities:
   //  a)We terminate normally.
   //  b)Somehow we generate a math error and terminate based on flags.
   //  c)Due to some unknown error in the math functions, we go on forever,
   //    and the program locks up.
   GMP_INTS_mpz_set_ui(&r, 1);
   loop_count = 0;
   while (!GMP_INTS_mpz_is_zero(&r) && !q.flags && !r.flags && (loop_count < 100000))
      {
      loop_count++;

      GMP_INTS_mpz_tdiv_qr(&q, &r, &u, &v);
      GMP_INTS_mpz_copy(&u, &v);
      GMP_INTS_mpz_copy(&v, &r);
      }

   //Let's hope we didn't get out of the loop based on loop count.
   assert(loop_count != 100000);

   //u now contains the answer.
   GMP_INTS_mpz_copy(result, &u);

   //Deallocate space for locals.
   GMP_INTS_mpz_clear(&u);
   GMP_INTS_mpz_clear(&v);
   GMP_INTS_mpz_clear(&q);
   GMP_INTS_mpz_clear(&r);
   }


/******************************************************************/
/***  COMPARISON AND SIZING FUNCTIONS   ***************************/
/******************************************************************/
//07/24/01:  Visual inspection only, due to simplicity.
int GMP_INTS_mpz_fits_uint_p (const GMP_INTS_mpz_struct *src)
   {
   GMP_INTS_size_t size;
   GMP_INTS_limb_t mpl;

   //Eyeball the input parameter.
   assert(src != NULL);
   assert(src->n_allocd > 0);
   assert(src->limbs != NULL);

   mpl =  src->limbs[0];
   size = src->size;
   if (size < 0 || size > 1)
     return(0);

   //The following line came from the original GNU code.
   //It isn't necessary in our case since limbs and ints are
   //both 32 bits, but it will do no harm.
   return (mpl <= (~(unsigned int) 0));
   }


unsigned GMP_INTS_mpz_get_limb_zero(const GMP_INTS_mpz_struct *src)
   {
   //Eyeball the inputs.
   assert(src != NULL);
   assert(src->n_allocd > 0);
   assert(src->limbs != NULL);

   if (!src->size)
      return(0);
   else
      return(src->limbs[0]);
   }


//07/24/01:  Visual inspection only.  Understood the comparisons
//and seems like they should work, but ... a little beyond my
//comfort zone without testing.  Trusting GNU on this one ...
int GMP_INTS_mpz_fits_sint_p (const GMP_INTS_mpz_struct *src)
   {
   GMP_INTS_size_t size;
   GMP_INTS_limb_t mpl;

   //Eyeball the input parameter.
   assert(src != NULL);
   assert(src->n_allocd > 0);
   assert(src->limbs != NULL);

   mpl  = src->limbs[0];
   size = src->size;
   if (size > 0)
      {
      if (size > 1)
	      return 0;
      return mpl < ~((~(unsigned int) 0) >> 1);
      }
   else
      {
      if (size < -1)
	     return 0;
      return mpl <= ~((~(unsigned int) 0) >> 1);
      }
   }


//07/24/01:  Visual inspection only.  One issue that caught
//my eye is that in one place function returned neg value if
//< and pos value if >.  Was within spec, but corrected because
//it concerned me as I often test against -1 and 1.  Seems
//to invite accidents.
int GMP_INTS_mpz_cmp (const GMP_INTS_mpz_struct *u, 
                      const GMP_INTS_mpz_struct *v)
   {
   GMP_INTS_size_t usize = u->size;
   GMP_INTS_size_t vsize = v->size;
   GMP_INTS_size_t size;
   GMP_INTS_limb_srcptr up, vp;
   int cmp;

   //Eyeball the input parameters.
   assert(u != NULL);
   assert(u->n_allocd > 0);
   assert(u->limbs != NULL);
   assert(v != NULL);
   assert(v->n_allocd > 0);
   assert(v->limbs != NULL);

   if (usize < vsize)
      return(-1);
   else if (usize > vsize)
      return(1);

   if (usize == 0)
      return(0);

   size = GMP_INTS_abs_of_size_t(usize);

   up = u->limbs;
   vp = v->limbs;

   cmp = GMP_INTS_mpn_cmp (up, vp, size);

   if (cmp == 0)
      return(0);

   if ((cmp < 0) == (usize < 0))
      return(1);
   else
      return(-1);
   }


//07/24/01:  Not visually inspected.  Relying on
//GNU ...
int GMP_INTS_mpz_cmp_ui (const GMP_INTS_mpz_struct *u, 
                         unsigned long int v_digit)
   {
   GMP_INTS_size_t usize = u->size;

   //Eyeball the input parameter.
   assert(u != NULL);
   assert(u->n_allocd > 0);
   assert(u->limbs != NULL);

   if (usize == 0)
      return -(v_digit != 0);

   if (usize == 1)
      {
      GMP_INTS_limb_t u_digit;

      u_digit = u->limbs[0];
      if (u_digit > v_digit)
	      return 1;
      if (u_digit < v_digit)
	      return -1;
      return 0;
      }

   return (usize > 0) ? 1 : -1;
   }


//07/24/01:  Not visually inspected.  Relying on GNU.
int GMP_INTS_mpz_cmp_si (const GMP_INTS_mpz_struct *u, 
                         signed long int v_digit)
   {
   GMP_INTS_size_t usize = u->size;
   GMP_INTS_size_t vsize;
   GMP_INTS_limb_t u_digit;

   //Eyeball the input parameter.
   assert(u != NULL);
   assert(u->n_allocd > 0);
   assert(u->limbs != NULL);

   vsize = 0;
   if (v_digit > 0)
      vsize = 1;
   else if (v_digit < 0)
      {
      vsize = -1;
      v_digit = -v_digit;
      }

   if (usize != vsize)
      return usize - vsize;

   if (usize == 0)
      return 0;

   u_digit = u->limbs[0];

   if (u_digit == (GMP_INTS_limb_t) (unsigned long) v_digit)
      return 0;

   if (u_digit > (GMP_INTS_limb_t) (unsigned long) v_digit)
      return usize;
   else
      return -usize;
   }


//07/24/01:  Not visually inspected.  Counting on GNU.
int GMP_INTS_mpz_cmpabs (const GMP_INTS_mpz_struct *u, 
                         const GMP_INTS_mpz_struct *v)
   {
   GMP_INTS_size_t usize = u->size;
   GMP_INTS_size_t vsize = v->size;
   GMP_INTS_limb_srcptr up, vp;
   int cmp;

   //Eyeball the input parameters.
   assert(u != NULL);
   assert(u->n_allocd > 0);
   assert(u->limbs != NULL);
   assert(v != NULL);
   assert(v->n_allocd > 0);
   assert(v->limbs != NULL);

   usize = GMP_INTS_abs_of_size_t(usize);
   vsize = GMP_INTS_abs_of_size_t(vsize);

   if (usize != vsize)
      return usize - vsize;

   if (usize == 0)
      return 0;

   up = u->limbs;
   vp = v->limbs;

   cmp = GMP_INTS_mpn_cmp (up, vp, usize);

   return cmp;
   }

//07/24/01:  Not visually inspected.  Counting on GNU.
int GMP_INTS_mpz_cmpabs_ui(const GMP_INTS_mpz_struct *u, 
                           unsigned long int v_digit)
   {
   GMP_INTS_size_t usize = u->size;

   //Eyeball the input parameter.
   assert(u != NULL);
   assert(u->n_allocd > 0);
   assert(u->limbs != NULL);

   if (usize == 0)
      return -(v_digit != 0);

   usize = GMP_INTS_abs_of_size_t(usize);

   if (usize == 1)
      {
      GMP_INTS_limb_t u_digit;

      u_digit = u->limbs[0];
      if (u_digit > v_digit)
	      return 1;
      if (u_digit < v_digit)
	      return -1;
      return 0;
      }

   return 1;
   }


/******************************************************************/
/***  VERSION CONTROL IDENTITY FUNCTIONS   ************************/
/******************************************************************/

//07/18/01:  Visual inspection only.  Function deemed too 
//simple for unit testing.
const char *GMP_INTS_cvcinfo(void)
   {
   return("$Header: /cvsroot/esrg/sfesrg/esrgpcpj/shared/c_datd/gmp_ints.c,v 1.22 2002/01/27 15:18:44 dtashley Exp $");
   }


//07/18/01:  Visual inspection only.  Function deemed too 
//simple for unit testing.
const char *GMP_INTS_hvcinfo(void)
   {
   return(GMP_INTS_H_VERSION);
   }


//**************************************************************************
// $Log: gmp_ints.c,v $
// Revision 1.22  2002/01/27 15:18:44  dtashley
// Minor typo corrected.
//
// Revision 1.21  2002/01/27 15:04:14  dtashley
// HYREACH added to batch build (which required some changes in other programs).
//
// Revision 1.20  2001/08/16 19:49:40  dtashley
// Beginning to prepare for v1.05 release.
//
// Revision 1.19  2001/08/15 06:56:05  dtashley
// Substantial progress.  Safety check-in.
//
// Revision 1.18  2001/08/07 10:42:48  dtashley
// Completion of CFRATNUM extensions and DOS command-line utility.
//
// Revision 1.17  2001/07/30 02:51:18  dtashley
// INTGCD extension and command-line utility finished up.
//
// Revision 1.16  2001/07/29 07:18:22  dtashley
// Completion of ARBINT INTFAC extension.
//
// Revision 1.15  2001/07/25 23:40:02  dtashley
// Completion of INTFAC program, many changes to handling of large
// integers.
//
// Revision 1.14  2001/07/21 01:39:01  dtashley
// Safety check-in.  Major function to output an integer as rows of digits
// has been completed.  This was the last major function that needed to be
// completed before useful command-line utilities can be constructed.
//
// Revision 1.13  2001/07/19 20:06:03  dtashley
// Division finished.  String formatting functions underway.  Safety check-in.
//
// Revision 1.12  2001/07/18 21:53:09  dtashley
// Division function finished and passes preliminary tests.  Safety check-in.
//
// Revision 1.11  2001/07/17 22:30:14  dtashley
// Safety check-in.  Division function under construction.
//
// Revision 1.10  2001/07/16 17:46:46  dtashley
// Multiplication finished, and only indirectly unit-tested.  More detailed unit
// test must follow, but expect no problems.
//
// Revision 1.9  2001/07/16 00:28:22  dtashley
// Safety check-in.  Addition and subtraction functions finished.
//
// Revision 1.8  2001/07/15 06:40:10  dtashley
// Adaptation of GNU arbitrary-size integer package integrated into IjuScripter
// and IjuConsole.
//
// Revision 1.7  2001/07/15 00:59:52  dtashley
// Safety check-in.  Commit before working on laptop.
//
// Revision 1.6  2001/07/14 07:03:37  dtashley
// Safety check-in.  Modifications and progress.
//
// Revision 1.5  2001/07/14 02:05:02  dtashley
// Safety check-in.  Almost ready for first unit-testing.
//
// Revision 1.4  2001/07/13 21:02:20  dtashley
// Version control reporting changes.
//
// Revision 1.3  2001/07/13 06:54:57  dtashley
// Safety check-in.  Substantial progress and modifications.
//
// Revision 1.2  2001/07/13 00:57:08  dtashley
// Safety check-in.  Substantial progress on port.
//
// Revision 1.1  2001/07/12 05:07:02  dtashley
// Initial checkin.
//
//**************************************************************************
// End of CCMALLOC.C.
