// $Header: /cvsroot/esrg/sfesrg/esrgpcpj/shared/c_datd/gmp_ints.h,v 1.23 2001/08/16 19:49:40 dtashley Exp $

//--------------------------------------------------------------------------------
//Copyright 2001 David T. Ashley
//-------------------------------------------------------------------------------------------------
//This source code and any program in which it is compiled/used is provided under the GNU GENERAL
//PUBLIC LICENSE, Version 3, full license text below.
//-------------------------------------------------------------------------------------------------
//                    GNU GENERAL PUBLIC LICENSE
//                       Version 3, 29 June 2007
//
// Copyright (C) 2007 Free Software Foundation, Inc. <http://fsf.org/>
// Everyone is permitted to copy and distribute verbatim copies
// of this license document, but changing it is not allowed.
//
//                            Preamble
//
//  The GNU General Public License is a free, copyleft license for
//software and other kinds of works.
//
//  The licenses for most software and other practical works are designed
//to take away your freedom to share and change the works.  By contrast,
//the GNU General Public License is intended to guarantee your freedom to
//share and change all versions of a program--to make sure it remains free
//software for all its users.  We, the Free Software Foundation, use the
//GNU General Public License for most of our software; it applies also to
//any other work released this way by its authors.  You can apply it to
//your programs, too.
//
//  When we speak of free software, we are referring to freedom, not
//price.  Our General Public Licenses are designed to make sure that you
//have the freedom to distribute copies of free software (and charge for
//them if you wish), that you receive source code or can get it if you
//want it, that you can change the software or use pieces of it in new
//free programs, and that you know you can do these things.
//
//  To protect your rights, we need to prevent others from denying you
//these rights or asking you to surrender the rights.  Therefore, you have
//certain responsibilities if you distribute copies of the software, or if
//you modify it: responsibilities to respect the freedom of others.
//
//  For example, if you distribute copies of such a program, whether
//gratis or for a fee, you must pass on to the recipients the same
//freedoms that you received.  You must make sure that they, too, receive
//or can get the source code.  And you must show them these terms so they
//know their rights.
//
//  Developers that use the GNU GPL protect your rights with two steps:
//(1) assert copyright on the software, and (2) offer you this License
//giving you legal permission to copy, distribute and/or modify it.
//
//  For the developers' and authors' protection, the GPL clearly explains
//that there is no warranty for this free software.  For both users' and
//authors' sake, the GPL requires that modified versions be marked as
//changed, so that their problems will not be attributed erroneously to
//authors of previous versions.
//
//  Some devices are designed to deny users access to install or run
//modified versions of the software inside them, although the manufacturer
//can do so.  This is fundamentally incompatible with the aim of
//protecting users' freedom to change the software.  The systematic
//pattern of such abuse occurs in the area of products for individuals to
//use, which is precisely where it is most unacceptable.  Therefore, we
//have designed this version of the GPL to prohibit the practice for those
//products.  If such problems arise substantially in other domains, we
//stand ready to extend this provision to those domains in future versions
//of the GPL, as needed to protect the freedom of users.
//
//  Finally, every program is threatened constantly by software patents.
//States should not allow patents to restrict development and use of
//software on general-purpose computers, but in those that do, we wish to
//avoid the special danger that patents applied to a free program could
//make it effectively proprietary.  To prevent this, the GPL assures that
//patents cannot be used to render the program non-free.
//
//  The precise terms and conditions for copying, distribution and
//modification follow.
//
//                       TERMS AND CONDITIONS
//
//  0. Definitions.
//
//  "This License" refers to version 3 of the GNU General Public License.
//
//  "Copyright" also means copyright-like laws that apply to other kinds of
//works, such as semiconductor masks.
//
//  "The Program" refers to any copyrightable work licensed under this
//License.  Each licensee is addressed as "you".  "Licensees" and
//"recipients" may be individuals or organizations.
//
//  To "modify" a work means to copy from or adapt all or part of the work
//in a fashion requiring copyright permission, other than the making of an
//exact copy.  The resulting work is called a "modified version" of the
//earlier work or a work "based on" the earlier work.
//
//  A "covered work" means either the unmodified Program or a work based
//on the Program.
//
//  To "propagate" a work means to do anything with it that, without
//permission, would make you directly or secondarily liable for
//infringement under applicable copyright law, except executing it on a
//computer or modifying a private copy.  Propagation includes copying,
//distribution (with or without modification), making available to the
//public, and in some countries other activities as well.
//
//  To "convey" a work means any kind of propagation that enables other
//parties to make or receive copies.  Mere interaction with a user through
//a computer network, with no transfer of a copy, is not conveying.
//
//  An interactive user interface displays "Appropriate Legal Notices"
//to the extent that it includes a convenient and prominently visible
//feature that (1) displays an appropriate copyright notice, and (2)
//tells the user that there is no warranty for the work (except to the
//extent that warranties are provided), that licensees may convey the
//work under this License, and how to view a copy of this License.  If
//the interface presents a list of user commands or options, such as a
//menu, a prominent item in the list meets this criterion.
//
//  1. Source Code.
//
//  The "source code" for a work means the preferred form of the work
//for making modifications to it.  "Object code" means any non-source
//form of a work.
//
//  A "Standard Interface" means an interface that either is an official
//standard defined by a recognized standards body, or, in the case of
//interfaces specified for a particular programming language, one that
//is widely used among developers working in that language.
//
//  The "System Libraries" of an executable work include anything, other
//than the work as a whole, that (a) is included in the normal form of
//packaging a Major Component, but which is not part of that Major
//Component, and (b) serves only to enable use of the work with that
//Major Component, or to implement a Standard Interface for which an
//implementation is available to the public in source code form.  A
//"Major Component", in this context, means a major essential component
//(kernel, window system, and so on) of the specific operating system
//(if any) on which the executable work runs, or a compiler used to
//produce the work, or an object code interpreter used to run it.
//
//  The "Corresponding Source" for a work in object code form means all
//the source code needed to generate, install, and (for an executable
//work) run the object code and to modify the work, including scripts to
//control those activities.  However, it does not include the work's
//System Libraries, or general-purpose tools or generally available free
//programs which are used unmodified in performing those activities but
//which are not part of the work.  For example, Corresponding Source
//includes interface definition files associated with source files for
//the work, and the source code for shared libraries and dynamically
//linked subprograms that the work is specifically designed to require,
//such as by intimate data communication or control flow between those
//subprograms and other parts of the work.
//
//  The Corresponding Source need not include anything that users
//can regenerate automatically from other parts of the Corresponding
//Source.
//
//  The Corresponding Source for a work in source code form is that
//same work.
//
//  2. Basic Permissions.
//
//  All rights granted under this License are granted for the term of
//copyright on the Program, and are irrevocable provided the stated
//conditions are met.  This License explicitly affirms your unlimited
//permission to run the unmodified Program.  The output from running a
//covered work is covered by this License only if the output, given its
//content, constitutes a covered work.  This License acknowledges your
//rights of fair use or other equivalent, as provided by copyright law.
//
//  You may make, run and propagate covered works that you do not
//convey, without conditions so long as your license otherwise remains
//in force.  You may convey covered works to others for the sole purpose
//of having them make modifications exclusively for you, or provide you
//with facilities for running those works, provided that you comply with
//the terms of this License in conveying all material for which you do
//not control copyright.  Those thus making or running the covered works
//for you must do so exclusively on your behalf, under your direction
//and control, on terms that prohibit them from making any copies of
//your copyrighted material outside their relationship with you.
//
//  Conveying under any other circumstances is permitted solely under
//the conditions stated below.  Sublicensing is not allowed; section 10
//makes it unnecessary.
//
//  3. Protecting Users' Legal Rights From Anti-Circumvention Law.
//
//  No covered work shall be deemed part of an effective technological
//measure under any applicable law fulfilling obligations under article
//11 of the WIPO copyright treaty adopted on 20 December 1996, or
//similar laws prohibiting or restricting circumvention of such
//measures.
//
//  When you convey a covered work, you waive any legal power to forbid
//circumvention of technological measures to the extent such circumvention
//is effected by exercising rights under this License with respect to
//the covered work, and you disclaim any intention to limit operation or
//modification of the work as a means of enforcing, against the work's
//users, your or third parties' legal rights to forbid circumvention of
//technological measures.
//
//  4. Conveying Verbatim Copies.
//
//  You may convey verbatim copies of the Program's source code as you
//receive it, in any medium, provided that you conspicuously and
//appropriately publish on each copy an appropriate copyright notice;
//keep intact all notices stating that this License and any
//non-permissive terms added in accord with section 7 apply to the code;
//keep intact all notices of the absence of any warranty; and give all
//recipients a copy of this License along with the Program.
//
//  You may charge any price or no price for each copy that you convey,
//and you may offer support or warranty protection for a fee.
//
//  5. Conveying Modified Source Versions.
//
//  You may convey a work based on the Program, or the modifications to
//produce it from the Program, in the form of source code under the
//terms of section 4, provided that you also meet all of these conditions:
//
//    a) The work must carry prominent notices stating that you modified
//    it, and giving a relevant date.
//
//    b) The work must carry prominent notices stating that it is
//    released under this License and any conditions added under section
//    7.  This requirement modifies the requirement in section 4 to
//    "keep intact all notices".
//
//    c) You must license the entire work, as a whole, under this
//    License to anyone who comes into possession of a copy.  This
//    License will therefore apply, along with any applicable section 7
//    additional terms, to the whole of the work, and all its parts,
//    regardless of how they are packaged.  This License gives no
//    permission to license the work in any other way, but it does not
//    invalidate such permission if you have separately received it.
//
//    d) If the work has interactive user interfaces, each must display
//    Appropriate Legal Notices; however, if the Program has interactive
//    interfaces that do not display Appropriate Legal Notices, your
//    work need not make them do so.
//
//  A compilation of a covered work with other separate and independent
//works, which are not by their nature extensions of the covered work,
//and which are not combined with it such as to form a larger program,
//in or on a volume of a storage or distribution medium, is called an
//"aggregate" if the compilation and its resulting copyright are not
//used to limit the access or legal rights of the compilation's users
//beyond what the individual works permit.  Inclusion of a covered work
//in an aggregate does not cause this License to apply to the other
//parts of the aggregate.
//
//  6. Conveying Non-Source Forms.
//
//  You may convey a covered work in object code form under the terms
//of sections 4 and 5, provided that you also convey the
//machine-readable Corresponding Source under the terms of this License,
//in one of these ways:
//
//    a) Convey the object code in, or embodied in, a physical product
//    (including a physical distribution medium), accompanied by the
//    Corresponding Source fixed on a durable physical medium
//    customarily used for software interchange.
//
//    b) Convey the object code in, or embodied in, a physical product
//    (including a physical distribution medium), accompanied by a
//    written offer, valid for at least three years and valid for as
//    long as you offer spare parts or customer support for that product
//    model, to give anyone who possesses the object code either (1) a
//    copy of the Corresponding Source for all the software in the
//    product that is covered by this License, on a durable physical
//    medium customarily used for software interchange, for a price no
//    more than your reasonable cost of physically performing this
//    conveying of source, or (2) access to copy the
//    Corresponding Source from a network server at no charge.
//
//    c) Convey individual copies of the object code with a copy of the
//    written offer to provide the Corresponding Source.  This
//    alternative is allowed only occasionally and noncommercially, and
//    only if you received the object code with such an offer, in accord
//    with subsection 6b.
//
//    d) Convey the object code by offering access from a designated
//    place (gratis or for a charge), and offer equivalent access to the
//    Corresponding Source in the same way through the same place at no
//    further charge.  You need not require recipients to copy the
//    Corresponding Source along with the object code.  If the place to
//    copy the object code is a network server, the Corresponding Source
//    may be on a different server (operated by you or a third party)
//    that supports equivalent copying facilities, provided you maintain
//    clear directions next to the object code saying where to find the
//    Corresponding Source.  Regardless of what server hosts the
//    Corresponding Source, you remain obligated to ensure that it is
//    available for as long as needed to satisfy these requirements.
//
//    e) Convey the object code using peer-to-peer transmission, provided
//    you inform other peers where the object code and Corresponding
//    Source of the work are being offered to the general public at no
//    charge under subsection 6d.
//
//  A separable portion of the object code, whose source code is excluded
//from the Corresponding Source as a System Library, need not be
//included in conveying the object code work.
//
//  A "User Product" is either (1) a "consumer product", which means any
//tangible personal property which is normally used for personal, family,
//or household purposes, or (2) anything designed or sold for incorporation
//into a dwelling.  In determining whether a product is a consumer product,
//doubtful cases shall be resolved in favor of coverage.  For a particular
//product received by a particular user, "normally used" refers to a
//typical or common use of that class of product, regardless of the status
//of the particular user or of the way in which the particular user
//actually uses, or expects or is expected to use, the product.  A product
//is a consumer product regardless of whether the product has substantial
//commercial, industrial or non-consumer uses, unless such uses represent
//the only significant mode of use of the product.
//
//  "Installation Information" for a User Product means any methods,
//procedures, authorization keys, or other information required to install
//and execute modified versions of a covered work in that User Product from
//a modified version of its Corresponding Source.  The information must
//suffice to ensure that the continued functioning of the modified object
//code is in no case prevented or interfered with solely because
//modification has been made.
//
//  If you convey an object code work under this section in, or with, or
//specifically for use in, a User Product, and the conveying occurs as
//part of a transaction in which the right of possession and use of the
//User Product is transferred to the recipient in perpetuity or for a
//fixed term (regardless of how the transaction is characterized), the
//Corresponding Source conveyed under this section must be accompanied
//by the Installation Information.  But this requirement does not apply
//if neither you nor any third party retains the ability to install
//modified object code on the User Product (for example, the work has
//been installed in ROM).
//
//  The requirement to provide Installation Information does not include a
//requirement to continue to provide support service, warranty, or updates
//for a work that has been modified or installed by the recipient, or for
//the User Product in which it has been modified or installed.  Access to a
//network may be denied when the modification itself materially and
//adversely affects the operation of the network or violates the rules and
//protocols for communication across the network.
//
//  Corresponding Source conveyed, and Installation Information provided,
//in accord with this section must be in a format that is publicly
//documented (and with an implementation available to the public in
//source code form), and must require no special password or key for
//unpacking, reading or copying.
//
//  7. Additional Terms.
//
//  "Additional permissions" are terms that supplement the terms of this
//License by making exceptions from one or more of its conditions.
//Additional permissions that are applicable to the entire Program shall
//be treated as though they were included in this License, to the extent
//that they are valid under applicable law.  If additional permissions
//apply only to part of the Program, that part may be used separately
//under those permissions, but the entire Program remains governed by
//this License without regard to the additional permissions.
//
//  When you convey a copy of a covered work, you may at your option
//remove any additional permissions from that copy, or from any part of
//it.  (Additional permissions may be written to require their own
//removal in certain cases when you modify the work.)  You may place
//additional permissions on material, added by you to a covered work,
//for which you have or can give appropriate copyright permission.
//
//  Notwithstanding any other provision of this License, for material you
//add to a covered work, you may (if authorized by the copyright holders of
//that material) supplement the terms of this License with terms:
//
//    a) Disclaiming warranty or limiting liability differently from the
//    terms of sections 15 and 16 of this License; or
//
//    b) Requiring preservation of specified reasonable legal notices or
//    author attributions in that material or in the Appropriate Legal
//    Notices displayed by works containing it; or
//
//    c) Prohibiting misrepresentation of the origin of that material, or
//    requiring that modified versions of such material be marked in
//    reasonable ways as different from the original version; or
//
//    d) Limiting the use for publicity purposes of names of licensors or
//    authors of the material; or
//
//    e) Declining to grant rights under trademark law for use of some
//    trade names, trademarks, or service marks; or
//
//    f) Requiring indemnification of licensors and authors of that
//    material by anyone who conveys the material (or modified versions of
//    it) with contractual assumptions of liability to the recipient, for
//    any liability that these contractual assumptions directly impose on
//    those licensors and authors.
//
//  All other non-permissive additional terms are considered "further
//restrictions" within the meaning of section 10.  If the Program as you
//received it, or any part of it, contains a notice stating that it is
//governed by this License along with a term that is a further
//restriction, you may remove that term.  If a license document contains
//a further restriction but permits relicensing or conveying under this
//License, you may add to a covered work material governed by the terms
//of that license document, provided that the further restriction does
//not survive such relicensing or conveying.
//
//  If you add terms to a covered work in accord with this section, you
//must place, in the relevant source files, a statement of the
//additional terms that apply to those files, or a notice indicating
//where to find the applicable terms.
//
//  Additional terms, permissive or non-permissive, may be stated in the
//form of a separately written license, or stated as exceptions;
//the above requirements apply either way.
//
//  8. Termination.
//
//  You may not propagate or modify a covered work except as expressly
//provided under this License.  Any attempt otherwise to propagate or
//modify it is void, and will automatically terminate your rights under
//this License (including any patent licenses granted under the third
//paragraph of section 11).
//
//  However, if you cease all violation of this License, then your
//license from a particular copyright holder is reinstated (a)
//provisionally, unless and until the copyright holder explicitly and
//finally terminates your license, and (b) permanently, if the copyright
//holder fails to notify you of the violation by some reasonable means
//prior to 60 days after the cessation.
//
//  Moreover, your license from a particular copyright holder is
//reinstated permanently if the copyright holder notifies you of the
//violation by some reasonable means, this is the first time you have
//received notice of violation of this License (for any work) from that
//copyright holder, and you cure the violation prior to 30 days after
//your receipt of the notice.
//
//  Termination of your rights under this section does not terminate the
//licenses of parties who have received copies or rights from you under
//this License.  If your rights have been terminated and not permanently
//reinstated, you do not qualify to receive new licenses for the same
//material under section 10.
//
//  9. Acceptance Not Required for Having Copies.
//
//  You are not required to accept this License in order to receive or
//run a copy of the Program.  Ancillary propagation of a covered work
//occurring solely as a consequence of using peer-to-peer transmission
//to receive a copy likewise does not require acceptance.  However,
//nothing other than this License grants you permission to propagate or
//modify any covered work.  These actions infringe copyright if you do
//not accept this License.  Therefore, by modifying or propagating a
//covered work, you indicate your acceptance of this License to do so.
//
//  10. Automatic Licensing of Downstream Recipients.
//
//  Each time you convey a covered work, the recipient automatically
//receives a license from the original licensors, to run, modify and
//propagate that work, subject to this License.  You are not responsible
//for enforcing compliance by third parties with this License.
//
//  An "entity transaction" is a transaction transferring control of an
//organization, or substantially all assets of one, or subdividing an
//organization, or merging organizations.  If propagation of a covered
//work results from an entity transaction, each party to that
//transaction who receives a copy of the work also receives whatever
//licenses to the work the party's predecessor in interest had or could
//give under the previous paragraph, plus a right to possession of the
//Corresponding Source of the work from the predecessor in interest, if
//the predecessor has it or can get it with reasonable efforts.
//
//  You may not impose any further restrictions on the exercise of the
//rights granted or affirmed under this License.  For example, you may
//not impose a license fee, royalty, or other charge for exercise of
//rights granted under this License, and you may not initiate litigation
//(including a cross-claim or counterclaim in a lawsuit) alleging that
//any patent claim is infringed by making, using, selling, offering for
//sale, or importing the Program or any portion of it.
//
//  11. Patents.
//
//  A "contributor" is a copyright holder who authorizes use under this
//License of the Program or a work on which the Program is based.  The
//work thus licensed is called the contributor's "contributor version".
//
//  A contributor's "essential patent claims" are all patent claims
//owned or controlled by the contributor, whether already acquired or
//hereafter acquired, that would be infringed by some manner, permitted
//by this License, of making, using, or selling its contributor version,
//but do not include claims that would be infringed only as a
//consequence of further modification of the contributor version.  For
//purposes of this definition, "control" includes the right to grant
//patent sublicenses in a manner consistent with the requirements of
//this License.
//
//  Each contributor grants you a non-exclusive, worldwide, royalty-free
//patent license under the contributor's essential patent claims, to
//make, use, sell, offer for sale, import and otherwise run, modify and
//propagate the contents of its contributor version.
//
//  In the following three paragraphs, a "patent license" is any express
//agreement or commitment, however denominated, not to enforce a patent
//(such as an express permission to practice a patent or covenant not to
//sue for patent infringement).  To "grant" such a patent license to a
//party means to make such an agreement or commitment not to enforce a
//patent against the party.
//
//  If you convey a covered work, knowingly relying on a patent license,
//and the Corresponding Source of the work is not available for anyone
//to copy, free of charge and under the terms of this License, through a
//publicly available network server or other readily accessible means,
//then you must either (1) cause the Corresponding Source to be so
//available, or (2) arrange to deprive yourself of the benefit of the
//patent license for this particular work, or (3) arrange, in a manner
//consistent with the requirements of this License, to extend the patent
//license to downstream recipients.  "Knowingly relying" means you have
//actual knowledge that, but for the patent license, your conveying the
//covered work in a country, or your recipient's use of the covered work
//in a country, would infringe one or more identifiable patents in that
//country that you have reason to believe are valid.
//
//  If, pursuant to or in connection with a single transaction or
//arrangement, you convey, or propagate by procuring conveyance of, a
//covered work, and grant a patent license to some of the parties
//receiving the covered work authorizing them to use, propagate, modify
//or convey a specific copy of the covered work, then the patent license
//you grant is automatically extended to all recipients of the covered
//work and works based on it.
//
//  A patent license is "discriminatory" if it does not include within
//the scope of its coverage, prohibits the exercise of, or is
//conditioned on the non-exercise of one or more of the rights that are
//specifically granted under this License.  You may not convey a covered
//work if you are a party to an arrangement with a third party that is
//in the business of distributing software, under which you make payment
//to the third party based on the extent of your activity of conveying
//the work, and under which the third party grants, to any of the
//parties who would receive the covered work from you, a discriminatory
//patent license (a) in connection with copies of the covered work
//conveyed by you (or copies made from those copies), or (b) primarily
//for and in connection with specific products or compilations that
//contain the covered work, unless you entered into that arrangement,
//or that patent license was granted, prior to 28 March 2007.
//
//  Nothing in this License shall be construed as excluding or limiting
//any implied license or other defenses to infringement that may
//otherwise be available to you under applicable patent law.
//
//  12. No Surrender of Others' Freedom.
//
//  If conditions are imposed on you (whether by court order, agreement or
//otherwise) that contradict the conditions of this License, they do not
//excuse you from the conditions of this License.  If you cannot convey a
//covered work so as to satisfy simultaneously your obligations under this
//License and any other pertinent obligations, then as a consequence you may
//not convey it at all.  For example, if you agree to terms that obligate you
//to collect a royalty for further conveying from those to whom you convey
//the Program, the only way you could satisfy both those terms and this
//License would be to refrain entirely from conveying the Program.
//
//  13. Use with the GNU Affero General Public License.
//
//  Notwithstanding any other provision of this License, you have
//permission to link or combine any covered work with a work licensed
//under version 3 of the GNU Affero General Public License into a single
//combined work, and to convey the resulting work.  The terms of this
//License will continue to apply to the part which is the covered work,
//but the special requirements of the GNU Affero General Public License,
//section 13, concerning interaction through a network will apply to the
//combination as such.
//
//  14. Revised Versions of this License.
//
//  The Free Software Foundation may publish revised and/or new versions of
//the GNU General Public License from time to time.  Such new versions will
//be similar in spirit to the present version, but may differ in detail to
//address new problems or concerns.
//
//  Each version is given a distinguishing version number.  If the
//Program specifies that a certain numbered version of the GNU General
//Public License "or any later version" applies to it, you have the
//option of following the terms and conditions either of that numbered
//version or of any later version published by the Free Software
//Foundation.  If the Program does not specify a version number of the
//GNU General Public License, you may choose any version ever published
//by the Free Software Foundation.
//
//  If the Program specifies that a proxy can decide which future
//versions of the GNU General Public License can be used, that proxy's
//public statement of acceptance of a version permanently authorizes you
//to choose that version for the Program.
//
//  Later license versions may give you additional or different
//permissions.  However, no additional obligations are imposed on any
//author or copyright holder as a result of your choosing to follow a
//later version.
//
//  15. Disclaimer of Warranty.
//
//  THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
//APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
//HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
//OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
//THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
//PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
//IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
//ALL NECESSARY SERVICING, REPAIR OR CORRECTION.
//
//  16. Limitation of Liability.
//
//  IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
//WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
//THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
//GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
//USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
//DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
//PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
//EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
//SUCH DAMAGES.
//
//  17. Interpretation of Sections 15 and 16.
//
//  If the disclaimer of warranty and limitation of liability provided
//above cannot be given local legal effect according to their terms,
//reviewing courts shall apply local law that most closely approximates
//an absolute waiver of all civil liability in connection with the
//Program, unless a warranty or assumption of liability accompanies a
//copy of the Program in return for a fee.
//
//                     END OF TERMS AND CONDITIONS
//
//            How to Apply These Terms to Your New Programs
//
//  If you develop a new program, and you want it to be of the greatest
//possible use to the public, the best way to achieve this is to make it
//free software which everyone can redistribute and change under these terms.
//
//  To do so, attach the following notices to the program.  It is safest
//to attach them to the start of each source file to most effectively
//state the exclusion of warranty; and each file should have at least
//the "copyright" line and a pointer to where the full notice is found.
//
//    <one line to give the program's name and a brief idea of what it does.>
//    Copyright (C) <year>  <name of author>
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//Also add information on how to contact you by electronic and paper mail.
//
//  If the program does terminal interaction, make it output a short
//notice like this when it starts in an interactive mode:
//
//    <program>  Copyright (C) <year>  <name of author>
//    This program comes with ABSOLUTELY NO WARRANTY; for details type `show w'.
//    This is free software, and you are welcome to redistribute it
//    under certain conditions; type `show c' for details.
//
//The hypothetical commands `show w' and `show c' should show the appropriate
//parts of the General Public License.  Of course, your program's commands
//might be different; for a GUI interface, you would use an "about box".
//
//  You should also get your employer (if you work as a programmer) or school,
//if any, to sign a "copyright disclaimer" for the program, if necessary.
//For more information on this, and how to apply and follow the GNU GPL, see
//<http://www.gnu.org/licenses/>.
//
//  The GNU General Public License does not permit incorporating your program
//into proprietary programs.  If your program is a subroutine library, you
//may consider it more useful to permit linking proprietary applications with
//the library.  If this is what you want to do, use the GNU Lesser General
//Public License instead of this License.  But first, please read
//<http://www.gnu.org/philosophy/why-not-lgpl.html>.
//-------------------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------
/*
This software module is a careful adaptation of the integer functions
in the GNU MP library.

By "adaptation" I mean the following:

   o Many of the macros and other constructs that affect readability
     have been removed.  This has been done for simplicity.

   o Only the needed functions (large integer and rational number 
     functions) have been included.  This is for use in Dave Ashley's
     set of tools.

   o The number of source modules has been dramatically reduced.

   o Assembly-language has been removed.  Instead, at the lowest
     level, the core is generic 'C' provided by GNU (this is
     selected if processor=none is the processor chosen).  This
     certainly has an effect on speed, but I would rather sacrifice
     quite a bit of speed and keep it simple (in one module).

   o The GNU MP library has been specifically ported for Microsoft
     Visual C++ 6.0.  In other words, its portability has been
     deliberately destroyed.  Do NOT try to use this on another
     platform (or at least certainly do NOT contact me if you have
     problems with this).

   o Certain other stylistic changes have been made.

The GNU MP code is impressive, because it is so quick.  However,
because many of the algorithms were so advanced, I couldn't
understand their basis.  I would rather have a slower software
module that I can debug than a faster one that can't be
decisively debugged if necessary.

This code can be optimized, if necessary, by reincorporating
GNU MP code without changing the interface.  This would be
fairly easy to do.  I would recommend proceeding in the following
order:
   o Replacing the innards of some of the lower-level
     functions with assembly-language (there may be usable
     assembly-language in the GNU code).  The lowest-level
     multiplication function, for example, was noted to be
     VERY inefficient (but correct!).
   o Making algorithmic improvements in the higher-level
     functions by reincorporating GNU code.  These
     algorithmic improvements were removed because
     I couldn't understand them.  For example, I could
     not understand the advanced multiplication and
     division algorithms.

I decline to include the GNU public license statement which is 
included with the GNU MP code, because this software has been
changed enough that it isn't quite the GNU MP code any longer.
However, this software module is still under the GNU public 
license because it is included with Dave Ashley's tool set (The 
Iju Tool Set).  Additionally, although this software module has been
modified enough that it really is a separate effort, I would like to
acknowledge the tremendous job done by the GNU MP team.  Certainly
I would not have undertaken a project of this complexity without
an existing template to work from--i.e. something successful already
in existence.  The really hard work--a working and efficient design--
was done by the GNU MP team, and my modifications, although substantial,
are trivial in comparison to the design work done by the GNU MP team.

It should also be noted that this integer module is not a product
of GNU--I named it GNU_INTS because it is based on the GNU MP
library.  The GNU MP library is faster and more sophisticated.  This
is a pared down module to give some of the speed benefits of binary
arithmetic without too much extra complexity.  Most of the 
sophisticated algorithms used by the GNU MP library have been
removed.

The basic abstraction implemented by this type of integer is that 
any operation can be performed on integers, and NAN flags propagate.
However, once an integer has reached a NAN state, the tests (is_zero, etc.)
are unreliable.  Generally, any client which uses these integers should
check errors before tests, but doesn't need to check them after each
calculation.  I say "before" tests because otherwise an integer may reach
NAN and cause an infinite loop because a test will never become TRUE, etc.
*/

#ifndef GMP_INTS_H_INCLUDED
   #define GMP_INTS_H_INCLUDED

   #ifdef MODULE_GMP_INTS
      #define DECMOD_GMP_INTS
   #else
      #define DECMOD_GMP_INTS extern
   #endif

   typedef unsigned long int GMP_INTS_limb_t;
      //The fundamental data type for representing integers in binary
      //form.  For MSVC++, this is a 32-bit type.  This data type
      //is not normally needed publicly.
   #define GMP_INTS_BITS_PER_LIMB  (32)
      //This is the number of bits that each limb is assumed to hold.
      //There is an explicit function in the interface to check this,
      //in case anything changes with future versions of MSVC++.
      //This software module is NOT meant to be portable, or even
      //particularly quick.  It is meant to be verifiable and
      //testable instead.
   typedef long int  GMP_INTS_limb_signed_t;
      //Signed version of above.
   typedef GMP_INTS_limb_t *  GMP_INTS_limb_ptr;
      //Pointer to a limb (or, more characteristically, to an array
      //of limbs).
   typedef const GMP_INTS_limb_t * GMP_INTS_limb_srcptr;
      //Constant pointer used as the source pointer for copies
      //and other operations.
   typedef long int  GMP_INTS_size_t;
      //Used for representing the size of things to this module
   typedef long int  GMP_INTS_exp_t;
      //Used for representing exponent.  Assumed to be large enough
      //for this.

   //Below is fundamental structure which holds an arbitrary-size
   //integer.
   typedef struct
      {
      int flags;
                       //Bit-flags as to the condition of the
                       //integer.  The bitflags defined so far
                       //are designed to be mutually exclusive,
                       //and are listed below.  At the same time,
                       //we need to define the "official" strings
                       //for these bit flags.
                       #define GMP_INTS_EF_INTOVF_POS        (1)
                       #define GMP_INTS_EF_INTOVF_POS_STRING   \
                               "GMP_INTS_EF_INTOVF_POS"
                          //The integer has become too large
                          //growing in a positive direction.
                          //No further arithmetic will be 
                          //performed on it.
                       #define GMP_INTS_EF_INTOVF_NEG        (2)
                       #define GMP_INTS_EF_INTOVF_NEG_STRING   \
                               "GMP_INTS_EF_INTOVF_NEG"
                          //The integer has become too large
                          //growing in a negative direction.
                          //No further arithmetic will be 
                          //performed on it.
                       #define GMP_INTS_EF_INTOVF_TAINT_POS  (4)
                       #define GMP_INTS_EF_INTOVF_TAINT_POS_STRING   \
                               "GMP_INTS_EF_INTOVF_TAINT_POS"
                          //The integer has been arithmetically
                          //combined with an integer that has
                          //overflowed in the positive direction.
                          //This integer is thus considered tainted
                          //and no further arithmetic will be 
                          //performed on it.
                       #define GMP_INTS_EF_INTOVF_TAINT_NEG  (8)
                       #define GMP_INTS_EF_INTOVF_TAINT_NEG_STRING   \
                               "GMP_INTS_EF_INTOVF_TAINT_NEG"
                          //The integer has been arithmetically
                          //combined with an integer that has
                          //overflowed in the negative direction.
                          //This integer is thus considered tainted
                          //and no further arithmetic will be 
                          //performed on it.
                       //The flags above will prevent arithmetic
                       //on an integer and will taint any integer
                       //that comes into contact with it as the
                       //result of an operation.  These flags are
                       //cleared only by an assignment operation
                       //of any kind.
                       //
                       //The next logical questions are:
                       //   o Why do we want to make a size
                       //     limit on integers, anyway?
                       //   o How big is too big?
                       //
                       //We want a limit because this software module
                       //will be used in programs that should be
                       //reliable.  Allowing arbitrarily large results
                       //would invite users to break the software.
                       //It would also make testing impossible, because
                       //the module could not be tested out to its
                       //limits.  Assigning limits means that testing
                       //can be performed to the limits.  Also, there
                       //are certain constructs in the software that
                       //may break at high limits, such as _alloca().
                       //
                       //The question of how large is too large is a
                       //harder question.  Arbitrarily, let's decide
                       //that we want the equivalent of integers with
                       //100,000 decimal digits, which is about
                       //332,224 bits, or 10,382 limbs.  (By the way
                       //this limit can always be raised later, but
                       //the important point is that it exists and
                       //is tested to).  Any integer with more limbs
                       //than this is considered illegal.
                       #define GMP_INTS_MAXIMUM_LIMBS_PER_INT  (10382)
      int n_allocd;   /* Number of limbs allocated and pointed
                         to by the limbs field.  This gives the
                         number allocated (i.e. physically
                         allocated), but the next field tells how 
                         many are used.  Memory is never shrunk
                         automatically by this software module
                         (per the wise GNU MP design), so this
                         number does not ever automatically 
                         decrease.  */
      int size;       /* abs(size) is the number of limbs the
                         last field points to.  If _mp_size is
                         negative this is a negative number.  This
                         will be zero if the integer is zero. */
      GMP_INTS_limb_t *limbs;  
                      /* Pointer to the limbs.  The first field
                         of the structure (above) tells how many
                         spots are allocated, and the second
                         field (above) tells how many are used.
                         Numbers are stored least-significant
                         limbs first.  */
      } GMP_INTS_mpz_struct;


   /* Wrappers for non-stack memory allocation function to be used
   ** by integer, rational number, and integer and rational number
   ** algorithm functions.  These must be public because they are
   ** used by the other modules that deal with integers and rational
   ** numbers.
   */
   DECMOD_GMP_INTS void *GMP_INTS_malloc( size_t size );
   DECMOD_GMP_INTS void *GMP_INTS_calloc( size_t num, size_t size );
   DECMOD_GMP_INTS void *GMP_INTS_realloc( void *memblock, size_t size );
   DECMOD_GMP_INTS void *GMP_INTS_realloc_w_size( void *memblock, 
                                                  size_t old_size,
                                                  size_t size );
   DECMOD_GMP_INTS void GMP_INTS_free( void *memblock );
   DECMOD_GMP_INTS void GMP_INTS_free_w_size( void *memblock, size_t size );

   /******************************************************************/
   /***  PORTABILITY CHECK FUNCTIONS   *******************************/
   /******************************************************************/
   //Because there is the risk that Microsoft Visual C++ might 
   //change in the future, the following function can be called
   //to see if the assumptions about data sizes are valid.  This
   //function returns TRUE if there is a problem, or FALSE
   //otherwise.
   DECMOD_GMP_INTS int GMP_INTS_data_sizes_are_wrong(void);


   /******************************************************************/
   /***  ERROR STRING IDENTIFICATION AND PROCESSING FUNCTIONS  *******/
   /******************************************************************/
   //These functions are provided because some clients may deal
   //only with symbolic representations (Tcl/Tk, for example).
   //
   //Attempts to identify the passed string as one of the error
   //strings sanctioned by this module.  Returns -1 if not a match
   //or else the following values.
   //  0:  GMP_INTS_EF_INTOVF_POS
   //  1:  GMP_INTS_EF_INTOVF_NEG
   //  2:  GMP_INTS_EF_INTOVF_TAINT_POS
   //  3:  GMP_INTS_EF_INTOVF_TAINT_NEG
   DECMOD_GMP_INTS 
   int GMP_INTS_identify_nan_string(const char *s);

   //Returns the sanctioned string for the nan value indicated.
   //Integers that may be passed are the same as the 
   //indices above.  No parameter except 0-3 is allowed.
   DECMOD_GMP_INTS 
   const char *GMP_INTS_supply_nan_string(int idx);

   /******************************************************************/
   /***  DEBUG PRINTING FUNCTIONS   **********************************/
   /******************************************************************/
   //These functions are for printing out integers and limbs
   //and groups of limbs for unit testing and debugging.
   //
   //Prints out a group of limbs, preceded by a description.
   //n may be 0 and/or the lg pointer supplied may be NULL.
   DECMOD_GMP_INTS
   void GMP_INTS_print_limb_group(FILE *stream,
                                  GMP_INTS_limb_srcptr lg,
                                  GMP_INTS_size_t n,
                                  char *desc);
   //Prints an entire integer for diagnosis.
   DECMOD_GMP_INTS
   void GMP_INTS_mpz_print_int(FILE *stream,
                               const GMP_INTS_mpz_struct *arg,
                               char *desc);

   /****************************************************************/
   /***   MACRO REPLACEMENTS   *************************************/
   /****************************************************************/
   //The functions here are macros that have been replaced and
   //made functions for clarity.  Speed was sacrificed for clarity.
   DECMOD_GMP_INTS
   int GMP_INTS_mpz_get_flags(const GMP_INTS_mpz_struct *arg);
      //Allows a caller to obtain the flags of an integer.  
      //A non-zero value indicates trouble.  To figure out which
      //trouble, compare against the bit constants defined
      //with the "flags" field of the structure type, above.
   DECMOD_GMP_INTS GMP_INTS_size_t GMP_INTS_abs_of_size_t(GMP_INTS_size_t arg);
      //Returns the absolute value of one of the size arguments used
      //to indicate number of limbs.  This function is useful 
      //because integers natively stored with a negative size
      //for negative arguments, so the absolute value gives the
      //number of limbs consumed.
   DECMOD_GMP_INTS int GMP_INTS_mpz_sgn(const GMP_INTS_mpz_struct *arg);
      //Returns -1 if integer is negative, 0 if it is zero,
      //or 1 if it is positive.
   DECMOD_GMP_INTS int GMP_INTS_mpz_is_neg(const GMP_INTS_mpz_struct *arg);
   DECMOD_GMP_INTS int GMP_INTS_mpz_is_zero(const GMP_INTS_mpz_struct *arg);
   DECMOD_GMP_INTS int GMP_INTS_mpz_is_pos(const GMP_INTS_mpz_struct *arg);
      //These functions return 1 if the condition is met, or 0
      //otherwise.
   DECMOD_GMP_INTS int GMP_INTS_mpz_is_odd(const GMP_INTS_mpz_struct *arg);
      //Returns 1 if the integer is odd, or 0 if it is even.
   DECMOD_GMP_INTS int GMP_INTS_mpz_is_even(const GMP_INTS_mpz_struct *arg);
      //Returns 1 if the integer is even, or 0 if it is odd.
   DECMOD_GMP_INTS void GMP_INTS_mpz_negate(GMP_INTS_mpz_struct *arg);
      //Negates the number (changes the sign).
   DECMOD_GMP_INTS
   void GMP_INTS_mpn_normalize(GMP_INTS_limb_ptr  limb_array,
                               GMP_INTS_size_t   *idx);
      //Adjusts an index downward to bypass any most significant
      //zero limbs.  This was a macro in the GNU implementation.
      //This is for when the size of a result is overestimated
      //(the most pessimistic estimates must be made when 
      //allocating memory before an operation).
   DECMOD_GMP_INTS
   void GMP_INTS_mpn_copy_limbs(GMP_INTS_limb_ptr    dest,
                                GMP_INTS_limb_srcptr src,
                                GMP_INTS_size_t      n);
      //Copies limbs from source to destination.  This
      //function replaces a macro that was present in the
      //GNU code.  Again, clarity over speed.
   /****************************************************************/
   /***   LOW-LEVEL ARITHMETIC FUNCTIONS   *************************/
   /****************************************************************/
   //Gives the flags of a result integer as a function of the
   //flags of its two operands.  If the function requires only
   //one operand, just use zero as the second parameter and the
   //result will be correct.  Handles tainting.
   DECMOD_GMP_INTS
   int GMP_INTS_two_op_flags_map(int flags1, int flags2);

   //Adds the single limb s2_limb to the array of limbs s1, 
   //processing carries, and copies the result to the location 
   //res_ptr.  res_ptr is assumed to be [at least] the same size
   //as s1.  From the design, s1 and res_ptr may be the
   //same set of locations.  The
   //result is 1 if there was a carry out of the final limb or
   //0 otherwise.
   DECMOD_GMP_INTS
   GMP_INTS_limb_t GMP_INTS_mpn_add_1 (GMP_INTS_limb_ptr    res_ptr,
                                       GMP_INTS_limb_srcptr s1_ptr,
                                       GMP_INTS_size_t      s1_size,
                                       GMP_INTS_limb_t      s2_limb);

   //Very similar to the addition function.  Counts on the single
   //limb that is being subtracted being not enough to roll over the
   //integer that is being subtracted from.  No harm if this is done
   //(it just obeys 2-s complement rules) but this was probably not
   //the intent.  From the design, s1 and res_ptr may 
   //be the same set of locations.  The
   //result is 1 if there was a borrow out of the final limb or
   //0 otherwise.
   DECMOD_GMP_INTS
   GMP_INTS_limb_t GMP_INTS_mpn_sub_1(GMP_INTS_limb_ptr    res_ptr,
                                      GMP_INTS_limb_srcptr s1_ptr,
                                      GMP_INTS_size_t      s1_size,
                                      GMP_INTS_limb_t      s2_limb);

   //Multiplies a limb by a group of limbs.  The return value is 
   //the unsigned overflow out of the "top", which may be up to
   //one limb large (i.e. it is more than just a carry, it has 
   //a value besides 0 and 1).
   DECMOD_GMP_INTS
   GMP_INTS_limb_t GMP_INTS_mpn_mul_1(GMP_INTS_limb_ptr    res_ptr, 
                                      GMP_INTS_limb_srcptr s1_ptr, 
                                      GMP_INTS_size_t      s1_size, 
                                      GMP_INTS_limb_t      s2_limb);

   //Adds together two groups of limbs of same size.  The result
   //may be the same location as one or both of the two
   //operands.
   DECMOD_GMP_INTS
   GMP_INTS_limb_t GMP_INTS_mpn_add_n(GMP_INTS_limb_ptr    res_ptr,  
                                      GMP_INTS_limb_srcptr s1_ptr, 
                                      GMP_INTS_limb_srcptr s2_ptr, 
                                      GMP_INTS_size_t      size);

   //This function makes the mapping:
   //   res_ptr = res_ptr + s2_limb * s1_ptr
   //From the design, it appears that it is alright if 
   //res_ptr and s1_ptr are the same area.  The return
   //value is the excess that that should exist one limb
   //above the MSL of res_ptr (i.e. it is more than a carry, it
   //may be larger than one).
   DECMOD_GMP_INTS
   GMP_INTS_limb_t GMP_INTS_mpn_addmul_1(GMP_INTS_limb_ptr    res_ptr, 
                                         GMP_INTS_limb_srcptr s1_ptr, 
                                         GMP_INTS_size_t      s1_size, 
                                         GMP_INTS_limb_t      s2_limb);

   //This function adds in general two natural numbers.  Numbers must
   //be arranged so that S2 takes no more limbs than S1, i.e. 
   //LIMBS(s1) >= LIMBS(s2).  Memory must be allocated, and there must
   //be space for the result.  Not clear from the design if memory
   //areas can be coincident, but suspect can.
   DECMOD_GMP_INTS
   GMP_INTS_limb_t GMP_INTS_mpn_add(GMP_INTS_limb_ptr    res_ptr,
                                    GMP_INTS_limb_srcptr s1_ptr,
                                    GMP_INTS_size_t      s1_size,
                                    GMP_INTS_limb_srcptr s2_ptr,
                                    GMP_INTS_size_t      s2_size);

   //Subtracts two same-size operands.  They may be coincident in
   //memory.
   DECMOD_GMP_INTS
   GMP_INTS_limb_t GMP_INTS_mpn_sub_n(GMP_INTS_limb_ptr    res_ptr, 
                                      GMP_INTS_limb_srcptr s1_ptr, 
                                      GMP_INTS_limb_srcptr s2_ptr, 
                                      GMP_INTS_size_t      size);

   //This function subtracts in general two natural numbers.  Numbers must
   //be arranged so that S2 takes no more limbs than S1, i.e. 
   //LIMBS(s1) >= LIMBS(s2).  Memory must be allocated, and there must
   //be space for the result.  Not clear from the design if memory
   //areas can be coincident.  Result is S1-S2.
   DECMOD_GMP_INTS
   GMP_INTS_limb_t GMP_INTS_mpn_sub (GMP_INTS_limb_ptr    res_ptr,
                                     GMP_INTS_limb_srcptr s1_ptr,
                                     GMP_INTS_size_t      s1_size,
                                     GMP_INTS_limb_srcptr s2_ptr,
                                     GMP_INTS_size_t      s2_size);

   //Shifts UP of size USIZE left by CNT.  CNT must be less than the number
   //of bits per limb, which is at this time 32.  "wp" and "up" may be 
   //coincident.  Zero count is not allowed.  Value returned is the bits shifted
   //out the left.
   DECMOD_GMP_INTS
   GMP_INTS_limb_t GMP_INTS_mpn_lshift(GMP_INTS_limb_ptr    wp,
                                       GMP_INTS_limb_srcptr up, 
                                       GMP_INTS_size_t      usize,
                                       unsigned int         cnt);

   //Shifts UP of size USIZE right by CNT.  CNT must be less than the number
   //of bits per limb, which is at this time 32.  "wp" and "up" may be 
   //coincident.  Zero count is not allowed.  Value returned is the bits shifted
   //out the right.
   DECMOD_GMP_INTS
   GMP_INTS_limb_t GMP_INTS_mpn_rshift (GMP_INTS_limb_ptr    wp,
                                        GMP_INTS_limb_srcptr up, 
                                        GMP_INTS_size_t      usize,
                                        unsigned int         cnt);

   //Compares two natural numbers of same size and returns the expected
   //1 iff op1 > op2, -1 if op1 < op2, or 0 if they are equal.  Leading
   //zero limbs are alright and do not affect the result.
   DECMOD_GMP_INTS
   int GMP_INTS_mpn_cmp (GMP_INTS_limb_srcptr op1_ptr, 
                         GMP_INTS_limb_srcptr op2_ptr, 
                         GMP_INTS_size_t      size);
      
   //This is the basic multiplication of two natural numbers, each of
   //which may occuply many limbs.  In the 
   //original GNU MP code, there were several algorithms which could
   //be selected, depending on the size of the operands and other
   //factors.  This was the most basic case--the basic longhand
   //multiplication.  The code has been pared so that this is the
   //only case.
   DECMOD_GMP_INTS
   void GMP_INTS_mpn_mul_basecase (GMP_INTS_limb_ptr prodp,
                                   GMP_INTS_limb_srcptr up, 
                                   GMP_INTS_size_t usize,
                                   GMP_INTS_limb_srcptr vp, 
                                   GMP_INTS_size_t vsize);

   //This is the basic multiplication of two natural numbers.  In the old
   //GNU MP code, one of several algorithms would be selected.  The code
   //has been pared down so that this is just a passthrough to
   //GMP_INTS_mpn_mul_basecase().  Only the simplest multiplication
   //algorithm is used at this point.
   DECMOD_GMP_INTS
   void GMP_INTS_mpn_mul_n (GMP_INTS_limb_ptr    p, 
                            GMP_INTS_limb_srcptr a, 
                            GMP_INTS_limb_srcptr b, 
                            GMP_INTS_size_t      n);
   //Multiplication of two natural numbers.  Value returned is the most
   //significant limb of the array of predicted size un+vn, rather than
   //the spillover (this is unlike most functions).
   DECMOD_GMP_INTS
   GMP_INTS_limb_t GMP_INTS_mpn_mul(GMP_INTS_limb_ptr    prodp,
                                    GMP_INTS_limb_srcptr up, 
                                    GMP_INTS_size_t      un,
                                    GMP_INTS_limb_srcptr vp, 
                                    GMP_INTS_size_t      vn);

   /******************************************************************/
   /***  LIMB SPACE REALLOCATION FUNCTIONS   *************************/
   /******************************************************************/
   DECMOD_GMP_INTS
   void *GMP_INTS_mpz_realloc (GMP_INTS_mpz_struct *m, 
                               GMP_INTS_size_t      new_size);
      //Changes the number of limbs allocated.

   /******************************************************************/
   /***  PUBLIC INITIALIZATION AND MEMORY MANAGEMENT FUNCTIONS   *****/
   /******************************************************************/
   //Allocate space for an integer and sets it to zero.
   //This must be the first call made.
   DECMOD_GMP_INTS
   void GMP_INTS_mpz_init (GMP_INTS_mpz_struct *x);

   //Deallocates space for an integer.  This must be the
   //final call made.
   DECMOD_GMP_INTS
   void GMP_INTS_mpz_clear (GMP_INTS_mpz_struct *x);


   /******************************************************************/
   /***  PUBLIC ASSIGNMENT FUNCTIONS   *******************************/
   /******************************************************************/
   //Copies from one integer to another.  Must both be allocated and \
   //not be the same variable.
   DECMOD_GMP_INTS
   void GMP_INTS_mpz_copy(      GMP_INTS_mpz_struct *dst,
                          const GMP_INTS_mpz_struct *src);

   //Assigns the integer to be the value of an unsigned long.
   DECMOD_GMP_INTS
   void GMP_INTS_mpz_set_ui (GMP_INTS_mpz_struct *dest, 
                             unsigned long int val);

   //Assigns the integer to be the value of a signed long.
   DECMOD_GMP_INTS
   void GMP_INTS_mpz_set_si (GMP_INTS_mpz_struct *dest, 
                             signed long int val);
   //Assigns the integer to be the value of a simple string
   //(simple = no E notation).  Only form handled is
   //   [-]digits.
   //Other forms will not cause catastrophic failure but might
   //not behave as expected.
   DECMOD_GMP_INTS
   void GMP_INTS_mpz_set_simple_char_str(GMP_INTS_mpz_struct *z, 
                                         const char *s);
   //Assigns an arbitary integer to be a number with E notation.
   //*failure set non-zero if can't parse or set.  Number must
   //be pure integer--no lost precison when scientific notation
   //is processed.
   void GMP_INTS_mpz_set_sci_not_num(GMP_INTS_mpz_struct *z,
                                     int *failure, 
                                     const char *s);

   //Attempts to parse an integer using the following three     
   //formats:  a)simple integer, b)simple integer with commas,
   //c)integer in scientific notation.  Returns the value of
   //the integer if parse successful, or 0 and a failure flag
   //otherwise.
   void GMP_INTS_mpz_set_general_int(GMP_INTS_mpz_struct *z,
                                     int *failure, 
                                     const char *s);

   //Attempts to parse an integer as either a simple integer,
   //an integer with commas, or a number in scientific notation
   //into a UINT32.  If the number cannot be parsed into that,
   //the result is zero and failure is true.
   void GMP_INTS_mpz_parse_into_uint32(unsigned *result,
                                       int *failure,
                                       char *s);

   //Swaps a and b.  What this does is a swap of the area
   //pointed to (i.e. the control block).  Pointers held in
   //these blocks are therefore automatically swapped.
   DECMOD_GMP_INTS
   void GMP_INTS_mpz_swap(GMP_INTS_mpz_struct *a, 
                          GMP_INTS_mpz_struct *b);

   /****************************************************************/
   /***   PUBLIC INTEGER ARITHMETIC FUNCTIONS   ********************/
   /****************************************************************/
   //Adds two arbitrary integers to produce an arbitrary result.
   //The result space must be initialized already.
   DECMOD_GMP_INTS
   void GMP_INTS_mpz_add (      GMP_INTS_mpz_struct *w, 
                          const GMP_INTS_mpz_struct *u, 
                          const GMP_INTS_mpz_struct *v);

   //Adds the integer v to u and produces w.  u and w must be
   //be created already.
   void GMP_INTS_mpz_add_ui (      GMP_INTS_mpz_struct *w, 
                             const GMP_INTS_mpz_struct *u, 
                             unsigned long int          v);

   //Subtracts two aribtrary integers to produce an arbitrary
   //result.  The result space must be initialized already.
   DECMOD_GMP_INTS
   void GMP_INTS_mpz_sub (      GMP_INTS_mpz_struct *w, 
                          const GMP_INTS_mpz_struct *u, 
                          const GMP_INTS_mpz_struct *v);

   //Subtracts the integer v from u to produce w.  u and w must 
   //be created already.
   void GMP_INTS_mpz_sub_ui (      GMP_INTS_mpz_struct *w, 
                             const GMP_INTS_mpz_struct *u, 
                             unsigned long int          v);

   //Multiplies two arbitrary integers to produce an arbitrary
   //result.  The result space must be initialized already.
   DECMOD_GMP_INTS
   void GMP_INTS_mpz_mul (      GMP_INTS_mpz_struct *w, 
                          const GMP_INTS_mpz_struct *u, 
                          const GMP_INTS_mpz_struct *v);

   //Multiplies the arbitrary integer by a C-native signed
   //long.
   DECMOD_GMP_INTS
   void GMP_INTS_mpz_mul_si (      GMP_INTS_mpz_struct *w, 
                             const GMP_INTS_mpz_struct *u, 
                                   long int             v);
   //Multiplies the arbitrary integer by a C-native unsigned
   //long.
   DECMOD_GMP_INTS
   void GMP_INTS_mpz_mul_ui (      GMP_INTS_mpz_struct *w, 
                             const GMP_INTS_mpz_struct *u, 
                             unsigned long int          v);

   //Divides integers.
   DECMOD_GMP_INTS
   void GMP_INTS_mpz_tdiv_qr (      GMP_INTS_mpz_struct *quot, 
                                    GMP_INTS_mpz_struct *rem, 
                              const GMP_INTS_mpz_struct *num, 
                              const GMP_INTS_mpz_struct *den);

   //Calculates the factorial.  All values <=1 result in a
   //value of 1.
   DECMOD_GMP_INTS
   void GMP_INTS_mpz_fac_ui(GMP_INTS_mpz_struct *result, 
                            unsigned long int n);

   //Exponentiates a base to an exponent.  0^0=1, N^0=1.  The algorithm
   //used is successive squaring and multiplying.  Not the MOST efficient,
   //but OK.  The result and the base must be different variables.
   DECMOD_GMP_INTS
   void GMP_INTS_mpz_pow_ui(         GMP_INTS_mpz_struct *result,
                            const    GMP_INTS_mpz_struct *base,
                            unsigned                      exponent);

   //Takes the absolute value of the arbitrary integer passed.
   DECMOD_GMP_INTS
   void GMP_INTS_mpz_abs(GMP_INTS_mpz_struct *arg);

   //Calculates the gcd() of the two arguments.  If either argument is zero, the
   //result is automatically 1.  If either argument is negative, its absolute
   //value is used.  The two input pointers may not be the same, because doing
   //that is senseless--it would become a copy operation, because gcd(x,x) = x.
   //However, the result may be the same as either of the two inputs.
   DECMOD_GMP_INTS
   void GMP_INTS_mpz_gcd(GMP_INTS_mpz_struct *result,
                         const GMP_INTS_mpz_struct *arg1,
                         const GMP_INTS_mpz_struct *arg2);


   /******************************************************************/
   /***  PUBLIC CONVERSION AND OUTPUT FUNCTIONS   ********************/
   /******************************************************************/

   //Get an upper bound on the number of characters required for
   //representing an integer, including minus sign and commas,
   //Terminating zero, etc.  This is just a little wasteful, but 
   //will always reserve enough memory.
   DECMOD_GMP_INTS
   int GMP_INTS_mpz_size_in_base_10(const GMP_INTS_mpz_struct *arg);

   //Convert from integer to string.  Commas must be
   //added separately, and there is enough space reserved for
   //them.  This function is only warrantied not to overflow
   //a buffer allocated using the function above as a sizing
   //guide.
   DECMOD_GMP_INTS
   void GMP_INTS_mpz_to_string(char *out,
                               const GMP_INTS_mpz_struct *in);

   //Prints the integer passed to the stream in a long-hand
   //format.
   DECMOD_GMP_INTS
   void GMP_INTS_mpz_long_int_format_to_stream(FILE *s,
                                               const GMP_INTS_mpz_struct *i,
                                               const char *desc);

   //Prints the integer raw to a stream--just digits, no extra chars.
   DECMOD_GMP_INTS
   void GMP_INTS_mpz_arb_int_raw_to_stream(FILE *s,
                                        const GMP_INTS_mpz_struct *i);

   /******************************************************************/
   /***  COMPARISON AND SIZING FUNCTIONS   ***************************/
   /******************************************************************/
   //Returns 1 if arbitrary integer will fit in an unsigned integer,
   //or 0 otherwise.
   DECMOD_GMP_INTS
   int GMP_INTS_mpz_fits_uint_p (const GMP_INTS_mpz_struct *src);

   //Returns 1 if arbitrary integer will fit in a signed integer,
   //or 0 otherwise.
   DECMOD_GMP_INTS
   int GMP_INTS_mpz_fits_sint_p (const GMP_INTS_mpz_struct *src);

   //Retrieves an unsigned version of limb zero from the number,
   //or will return zero if the number is zero.  This is the
   //officially sanctioned way to get the value if it fits in
   //an unsigned.
   DECMOD_GMP_INTS
   unsigned GMP_INTS_mpz_get_limb_zero(const GMP_INTS_mpz_struct *src);

   //Returnes neg value if u<v, 0 if u==v, and 1 if u>v.
   DECMOD_GMP_INTS 
   int GMP_INTS_mpz_cmp (const GMP_INTS_mpz_struct *u, 
                         const GMP_INTS_mpz_struct *v);

   //Compares an arbitrary integer to an unsigned long
   //and returns relative ordering, <0 if u < v, 0 if u==v,
   //and >0 if u > v.
   DECMOD_GMP_INTS
   int GMP_INTS_mpz_cmp_ui (const GMP_INTS_mpz_struct *u, 
                            unsigned long int v_digit);
   //Compares arbitrary integer to a signed long
   //and returns relative ordering, <0 if u<v, 0 if u==v,
   //and >0 if u>v.
   DECMOD_GMP_INTS
   int GMP_INTS_mpz_cmp_si (const GMP_INTS_mpz_struct *u, 
                            signed long int v_digit);

   //Compares the absolute value of two arbitrary integers.
   DECMOD_GMP_INTS
   int GMP_INTS_mpz_cmpabs (const GMP_INTS_mpz_struct *u, 
                            const GMP_INTS_mpz_struct *v);
   //Compares the absolute value of an arbitrary integer and
   //an unsigned long.
   DECMOD_GMP_INTS
   int GMP_INTS_mpz_cmpabs_ui(const GMP_INTS_mpz_struct *u, 
                              unsigned long int v_digit);

   /****************************************************************/
   /***   VERSION CONTROL REPORTING FUNCTIONS   ********************/
   /****************************************************************/
   DECMOD_GMP_INTS const char *GMP_INTS_cvcinfo(void);
   DECMOD_GMP_INTS const char *GMP_INTS_hvcinfo(void);

   /* Preprocessor string to allow the H-file version to be
   ** compiled into the C-file.
   */
   #define GMP_INTS_H_VERSION ("$Header: /cvsroot/esrg/sfesrg/esrgpcpj/shared/c_datd/gmp_ints.h,v 1.23 2001/08/16 19:49:40 dtashley Exp $")

#endif /* GMP_INTS_H_INCLUDED */


//**************************************************************************
// $Log: gmp_ints.h,v $
// Revision 1.23  2001/08/16 19:49:40  dtashley
// Beginning to prepare for v1.05 release.
//
// Revision 1.22  2001/08/15 06:56:05  dtashley
// Substantial progress.  Safety check-in.
//
// Revision 1.21  2001/08/07 10:42:48  dtashley
// Completion of CFRATNUM extensions and DOS command-line utility.
//
// Revision 1.20  2001/07/30 02:51:18  dtashley
// INTGCD extension and command-line utility finished up.
//
// Revision 1.19  2001/07/29 07:18:22  dtashley
// Completion of ARBINT INTFAC extension.
//
// Revision 1.18  2001/07/25 23:40:02  dtashley
// Completion of INTFAC program, many changes to handling of large
// integers.
//
// Revision 1.17  2001/07/21 03:32:44  dtashley
// Addition of several files.
//
// Revision 1.16  2001/07/21 01:39:01  dtashley
// Safety check-in.  Major function to output an integer as rows of digits
// has been completed.  This was the last major function that needed to be
// completed before useful command-line utilities can be constructed.
//
// Revision 1.15  2001/07/19 20:06:03  dtashley
// Division finished.  String formatting functions underway.  Safety check-in.
//
// Revision 1.14  2001/07/18 21:53:09  dtashley
// Division function finished and passes preliminary tests.  Safety check-in.
//
// Revision 1.13  2001/07/17 22:30:14  dtashley
// Safety check-in.  Division function under construction.
//
// Revision 1.12  2001/07/16 17:46:46  dtashley
// Multiplication finished, and only indirectly unit-tested.  More detailed unit
// test must follow, but expect no problems.
//
// Revision 1.11  2001/07/16 00:28:22  dtashley
// Safety check-in.  Addition and subtraction functions finished.
//
// Revision 1.10  2001/07/15 06:40:10  dtashley
// Adaptation of GNU arbitrary-size integer package integrated into IjuScripter
// and IjuConsole.
//
// Revision 1.9  2001/07/15 00:59:52  dtashley
// Safety check-in.  Commit before working on laptop.
//
// Revision 1.8  2001/07/14 07:03:37  dtashley
// Safety check-in.  Modifications and progress.
//
// Revision 1.7  2001/07/14 02:05:02  dtashley
// Safety check-in.  Almost ready for first unit-testing.
//
// Revision 1.6  2001/07/13 21:02:20  dtashley
// Version control reporting changes.
//
// Revision 1.5  2001/07/13 18:17:16  dtashley
// Safety check-in.  Additionally, new files added.
//
// Revision 1.4  2001/07/13 06:54:57  dtashley
// Safety check-in.  Substantial progress and modifications.
//
// Revision 1.3  2001/07/13 00:57:08  dtashley
// Safety check-in.  Substantial progress on port.
//
// Revision 1.2  2001/07/12 06:44:12  dtashley
// Safety check-in.
//
// Revision 1.1  2001/07/12 05:07:02  dtashley
// Initial checkin.
//
//**************************************************************************
// End of GMP_INTS.H. 