// $Header: /cvsroot/esrg/sfesrg/esrgpcpj/shared/c_datd/gmp_ralg.c,v 1.10 2002/01/27 17:58:15 dtashley Exp $

//--------------------------------------------------------------------------------
//Copyright 2001 David T. Ashley
//-------------------------------------------------------------------------------------------------
//This source code and any program in which it is compiled/used is provided under the GNU GENERAL
//PUBLIC LICENSE, Version 3, full license text below.
//-------------------------------------------------------------------------------------------------
//                    GNU GENERAL PUBLIC LICENSE
//                       Version 3, 29 June 2007
//
// Copyright (C) 2007 Free Software Foundation, Inc. <http://fsf.org/>
// Everyone is permitted to copy and distribute verbatim copies
// of this license document, but changing it is not allowed.
//
//                            Preamble
//
//  The GNU General Public License is a free, copyleft license for
//software and other kinds of works.
//
//  The licenses for most software and other practical works are designed
//to take away your freedom to share and change the works.  By contrast,
//the GNU General Public License is intended to guarantee your freedom to
//share and change all versions of a program--to make sure it remains free
//software for all its users.  We, the Free Software Foundation, use the
//GNU General Public License for most of our software; it applies also to
//any other work released this way by its authors.  You can apply it to
//your programs, too.
//
//  When we speak of free software, we are referring to freedom, not
//price.  Our General Public Licenses are designed to make sure that you
//have the freedom to distribute copies of free software (and charge for
//them if you wish), that you receive source code or can get it if you
//want it, that you can change the software or use pieces of it in new
//free programs, and that you know you can do these things.
//
//  To protect your rights, we need to prevent others from denying you
//these rights or asking you to surrender the rights.  Therefore, you have
//certain responsibilities if you distribute copies of the software, or if
//you modify it: responsibilities to respect the freedom of others.
//
//  For example, if you distribute copies of such a program, whether
//gratis or for a fee, you must pass on to the recipients the same
//freedoms that you received.  You must make sure that they, too, receive
//or can get the source code.  And you must show them these terms so they
//know their rights.
//
//  Developers that use the GNU GPL protect your rights with two steps:
//(1) assert copyright on the software, and (2) offer you this License
//giving you legal permission to copy, distribute and/or modify it.
//
//  For the developers' and authors' protection, the GPL clearly explains
//that there is no warranty for this free software.  For both users' and
//authors' sake, the GPL requires that modified versions be marked as
//changed, so that their problems will not be attributed erroneously to
//authors of previous versions.
//
//  Some devices are designed to deny users access to install or run
//modified versions of the software inside them, although the manufacturer
//can do so.  This is fundamentally incompatible with the aim of
//protecting users' freedom to change the software.  The systematic
//pattern of such abuse occurs in the area of products for individuals to
//use, which is precisely where it is most unacceptable.  Therefore, we
//have designed this version of the GPL to prohibit the practice for those
//products.  If such problems arise substantially in other domains, we
//stand ready to extend this provision to those domains in future versions
//of the GPL, as needed to protect the freedom of users.
//
//  Finally, every program is threatened constantly by software patents.
//States should not allow patents to restrict development and use of
//software on general-purpose computers, but in those that do, we wish to
//avoid the special danger that patents applied to a free program could
//make it effectively proprietary.  To prevent this, the GPL assures that
//patents cannot be used to render the program non-free.
//
//  The precise terms and conditions for copying, distribution and
//modification follow.
//
//                       TERMS AND CONDITIONS
//
//  0. Definitions.
//
//  "This License" refers to version 3 of the GNU General Public License.
//
//  "Copyright" also means copyright-like laws that apply to other kinds of
//works, such as semiconductor masks.
//
//  "The Program" refers to any copyrightable work licensed under this
//License.  Each licensee is addressed as "you".  "Licensees" and
//"recipients" may be individuals or organizations.
//
//  To "modify" a work means to copy from or adapt all or part of the work
//in a fashion requiring copyright permission, other than the making of an
//exact copy.  The resulting work is called a "modified version" of the
//earlier work or a work "based on" the earlier work.
//
//  A "covered work" means either the unmodified Program or a work based
//on the Program.
//
//  To "propagate" a work means to do anything with it that, without
//permission, would make you directly or secondarily liable for
//infringement under applicable copyright law, except executing it on a
//computer or modifying a private copy.  Propagation includes copying,
//distribution (with or without modification), making available to the
//public, and in some countries other activities as well.
//
//  To "convey" a work means any kind of propagation that enables other
//parties to make or receive copies.  Mere interaction with a user through
//a computer network, with no transfer of a copy, is not conveying.
//
//  An interactive user interface displays "Appropriate Legal Notices"
//to the extent that it includes a convenient and prominently visible
//feature that (1) displays an appropriate copyright notice, and (2)
//tells the user that there is no warranty for the work (except to the
//extent that warranties are provided), that licensees may convey the
//work under this License, and how to view a copy of this License.  If
//the interface presents a list of user commands or options, such as a
//menu, a prominent item in the list meets this criterion.
//
//  1. Source Code.
//
//  The "source code" for a work means the preferred form of the work
//for making modifications to it.  "Object code" means any non-source
//form of a work.
//
//  A "Standard Interface" means an interface that either is an official
//standard defined by a recognized standards body, or, in the case of
//interfaces specified for a particular programming language, one that
//is widely used among developers working in that language.
//
//  The "System Libraries" of an executable work include anything, other
//than the work as a whole, that (a) is included in the normal form of
//packaging a Major Component, but which is not part of that Major
//Component, and (b) serves only to enable use of the work with that
//Major Component, or to implement a Standard Interface for which an
//implementation is available to the public in source code form.  A
//"Major Component", in this context, means a major essential component
//(kernel, window system, and so on) of the specific operating system
//(if any) on which the executable work runs, or a compiler used to
//produce the work, or an object code interpreter used to run it.
//
//  The "Corresponding Source" for a work in object code form means all
//the source code needed to generate, install, and (for an executable
//work) run the object code and to modify the work, including scripts to
//control those activities.  However, it does not include the work's
//System Libraries, or general-purpose tools or generally available free
//programs which are used unmodified in performing those activities but
//which are not part of the work.  For example, Corresponding Source
//includes interface definition files associated with source files for
//the work, and the source code for shared libraries and dynamically
//linked subprograms that the work is specifically designed to require,
//such as by intimate data communication or control flow between those
//subprograms and other parts of the work.
//
//  The Corresponding Source need not include anything that users
//can regenerate automatically from other parts of the Corresponding
//Source.
//
//  The Corresponding Source for a work in source code form is that
//same work.
//
//  2. Basic Permissions.
//
//  All rights granted under this License are granted for the term of
//copyright on the Program, and are irrevocable provided the stated
//conditions are met.  This License explicitly affirms your unlimited
//permission to run the unmodified Program.  The output from running a
//covered work is covered by this License only if the output, given its
//content, constitutes a covered work.  This License acknowledges your
//rights of fair use or other equivalent, as provided by copyright law.
//
//  You may make, run and propagate covered works that you do not
//convey, without conditions so long as your license otherwise remains
//in force.  You may convey covered works to others for the sole purpose
//of having them make modifications exclusively for you, or provide you
//with facilities for running those works, provided that you comply with
//the terms of this License in conveying all material for which you do
//not control copyright.  Those thus making or running the covered works
//for you must do so exclusively on your behalf, under your direction
//and control, on terms that prohibit them from making any copies of
//your copyrighted material outside their relationship with you.
//
//  Conveying under any other circumstances is permitted solely under
//the conditions stated below.  Sublicensing is not allowed; section 10
//makes it unnecessary.
//
//  3. Protecting Users' Legal Rights From Anti-Circumvention Law.
//
//  No covered work shall be deemed part of an effective technological
//measure under any applicable law fulfilling obligations under article
//11 of the WIPO copyright treaty adopted on 20 December 1996, or
//similar laws prohibiting or restricting circumvention of such
//measures.
//
//  When you convey a covered work, you waive any legal power to forbid
//circumvention of technological measures to the extent such circumvention
//is effected by exercising rights under this License with respect to
//the covered work, and you disclaim any intention to limit operation or
//modification of the work as a means of enforcing, against the work's
//users, your or third parties' legal rights to forbid circumvention of
//technological measures.
//
//  4. Conveying Verbatim Copies.
//
//  You may convey verbatim copies of the Program's source code as you
//receive it, in any medium, provided that you conspicuously and
//appropriately publish on each copy an appropriate copyright notice;
//keep intact all notices stating that this License and any
//non-permissive terms added in accord with section 7 apply to the code;
//keep intact all notices of the absence of any warranty; and give all
//recipients a copy of this License along with the Program.
//
//  You may charge any price or no price for each copy that you convey,
//and you may offer support or warranty protection for a fee.
//
//  5. Conveying Modified Source Versions.
//
//  You may convey a work based on the Program, or the modifications to
//produce it from the Program, in the form of source code under the
//terms of section 4, provided that you also meet all of these conditions:
//
//    a) The work must carry prominent notices stating that you modified
//    it, and giving a relevant date.
//
//    b) The work must carry prominent notices stating that it is
//    released under this License and any conditions added under section
//    7.  This requirement modifies the requirement in section 4 to
//    "keep intact all notices".
//
//    c) You must license the entire work, as a whole, under this
//    License to anyone who comes into possession of a copy.  This
//    License will therefore apply, along with any applicable section 7
//    additional terms, to the whole of the work, and all its parts,
//    regardless of how they are packaged.  This License gives no
//    permission to license the work in any other way, but it does not
//    invalidate such permission if you have separately received it.
//
//    d) If the work has interactive user interfaces, each must display
//    Appropriate Legal Notices; however, if the Program has interactive
//    interfaces that do not display Appropriate Legal Notices, your
//    work need not make them do so.
//
//  A compilation of a covered work with other separate and independent
//works, which are not by their nature extensions of the covered work,
//and which are not combined with it such as to form a larger program,
//in or on a volume of a storage or distribution medium, is called an
//"aggregate" if the compilation and its resulting copyright are not
//used to limit the access or legal rights of the compilation's users
//beyond what the individual works permit.  Inclusion of a covered work
//in an aggregate does not cause this License to apply to the other
//parts of the aggregate.
//
//  6. Conveying Non-Source Forms.
//
//  You may convey a covered work in object code form under the terms
//of sections 4 and 5, provided that you also convey the
//machine-readable Corresponding Source under the terms of this License,
//in one of these ways:
//
//    a) Convey the object code in, or embodied in, a physical product
//    (including a physical distribution medium), accompanied by the
//    Corresponding Source fixed on a durable physical medium
//    customarily used for software interchange.
//
//    b) Convey the object code in, or embodied in, a physical product
//    (including a physical distribution medium), accompanied by a
//    written offer, valid for at least three years and valid for as
//    long as you offer spare parts or customer support for that product
//    model, to give anyone who possesses the object code either (1) a
//    copy of the Corresponding Source for all the software in the
//    product that is covered by this License, on a durable physical
//    medium customarily used for software interchange, for a price no
//    more than your reasonable cost of physically performing this
//    conveying of source, or (2) access to copy the
//    Corresponding Source from a network server at no charge.
//
//    c) Convey individual copies of the object code with a copy of the
//    written offer to provide the Corresponding Source.  This
//    alternative is allowed only occasionally and noncommercially, and
//    only if you received the object code with such an offer, in accord
//    with subsection 6b.
//
//    d) Convey the object code by offering access from a designated
//    place (gratis or for a charge), and offer equivalent access to the
//    Corresponding Source in the same way through the same place at no
//    further charge.  You need not require recipients to copy the
//    Corresponding Source along with the object code.  If the place to
//    copy the object code is a network server, the Corresponding Source
//    may be on a different server (operated by you or a third party)
//    that supports equivalent copying facilities, provided you maintain
//    clear directions next to the object code saying where to find the
//    Corresponding Source.  Regardless of what server hosts the
//    Corresponding Source, you remain obligated to ensure that it is
//    available for as long as needed to satisfy these requirements.
//
//    e) Convey the object code using peer-to-peer transmission, provided
//    you inform other peers where the object code and Corresponding
//    Source of the work are being offered to the general public at no
//    charge under subsection 6d.
//
//  A separable portion of the object code, whose source code is excluded
//from the Corresponding Source as a System Library, need not be
//included in conveying the object code work.
//
//  A "User Product" is either (1) a "consumer product", which means any
//tangible personal property which is normally used for personal, family,
//or household purposes, or (2) anything designed or sold for incorporation
//into a dwelling.  In determining whether a product is a consumer product,
//doubtful cases shall be resolved in favor of coverage.  For a particular
//product received by a particular user, "normally used" refers to a
//typical or common use of that class of product, regardless of the status
//of the particular user or of the way in which the particular user
//actually uses, or expects or is expected to use, the product.  A product
//is a consumer product regardless of whether the product has substantial
//commercial, industrial or non-consumer uses, unless such uses represent
//the only significant mode of use of the product.
//
//  "Installation Information" for a User Product means any methods,
//procedures, authorization keys, or other information required to install
//and execute modified versions of a covered work in that User Product from
//a modified version of its Corresponding Source.  The information must
//suffice to ensure that the continued functioning of the modified object
//code is in no case prevented or interfered with solely because
//modification has been made.
//
//  If you convey an object code work under this section in, or with, or
//specifically for use in, a User Product, and the conveying occurs as
//part of a transaction in which the right of possession and use of the
//User Product is transferred to the recipient in perpetuity or for a
//fixed term (regardless of how the transaction is characterized), the
//Corresponding Source conveyed under this section must be accompanied
//by the Installation Information.  But this requirement does not apply
//if neither you nor any third party retains the ability to install
//modified object code on the User Product (for example, the work has
//been installed in ROM).
//
//  The requirement to provide Installation Information does not include a
//requirement to continue to provide support service, warranty, or updates
//for a work that has been modified or installed by the recipient, or for
//the User Product in which it has been modified or installed.  Access to a
//network may be denied when the modification itself materially and
//adversely affects the operation of the network or violates the rules and
//protocols for communication across the network.
//
//  Corresponding Source conveyed, and Installation Information provided,
//in accord with this section must be in a format that is publicly
//documented (and with an implementation available to the public in
//source code form), and must require no special password or key for
//unpacking, reading or copying.
//
//  7. Additional Terms.
//
//  "Additional permissions" are terms that supplement the terms of this
//License by making exceptions from one or more of its conditions.
//Additional permissions that are applicable to the entire Program shall
//be treated as though they were included in this License, to the extent
//that they are valid under applicable law.  If additional permissions
//apply only to part of the Program, that part may be used separately
//under those permissions, but the entire Program remains governed by
//this License without regard to the additional permissions.
//
//  When you convey a copy of a covered work, you may at your option
//remove any additional permissions from that copy, or from any part of
//it.  (Additional permissions may be written to require their own
//removal in certain cases when you modify the work.)  You may place
//additional permissions on material, added by you to a covered work,
//for which you have or can give appropriate copyright permission.
//
//  Notwithstanding any other provision of this License, for material you
//add to a covered work, you may (if authorized by the copyright holders of
//that material) supplement the terms of this License with terms:
//
//    a) Disclaiming warranty or limiting liability differently from the
//    terms of sections 15 and 16 of this License; or
//
//    b) Requiring preservation of specified reasonable legal notices or
//    author attributions in that material or in the Appropriate Legal
//    Notices displayed by works containing it; or
//
//    c) Prohibiting misrepresentation of the origin of that material, or
//    requiring that modified versions of such material be marked in
//    reasonable ways as different from the original version; or
//
//    d) Limiting the use for publicity purposes of names of licensors or
//    authors of the material; or
//
//    e) Declining to grant rights under trademark law for use of some
//    trade names, trademarks, or service marks; or
//
//    f) Requiring indemnification of licensors and authors of that
//    material by anyone who conveys the material (or modified versions of
//    it) with contractual assumptions of liability to the recipient, for
//    any liability that these contractual assumptions directly impose on
//    those licensors and authors.
//
//  All other non-permissive additional terms are considered "further
//restrictions" within the meaning of section 10.  If the Program as you
//received it, or any part of it, contains a notice stating that it is
//governed by this License along with a term that is a further
//restriction, you may remove that term.  If a license document contains
//a further restriction but permits relicensing or conveying under this
//License, you may add to a covered work material governed by the terms
//of that license document, provided that the further restriction does
//not survive such relicensing or conveying.
//
//  If you add terms to a covered work in accord with this section, you
//must place, in the relevant source files, a statement of the
//additional terms that apply to those files, or a notice indicating
//where to find the applicable terms.
//
//  Additional terms, permissive or non-permissive, may be stated in the
//form of a separately written license, or stated as exceptions;
//the above requirements apply either way.
//
//  8. Termination.
//
//  You may not propagate or modify a covered work except as expressly
//provided under this License.  Any attempt otherwise to propagate or
//modify it is void, and will automatically terminate your rights under
//this License (including any patent licenses granted under the third
//paragraph of section 11).
//
//  However, if you cease all violation of this License, then your
//license from a particular copyright holder is reinstated (a)
//provisionally, unless and until the copyright holder explicitly and
//finally terminates your license, and (b) permanently, if the copyright
//holder fails to notify you of the violation by some reasonable means
//prior to 60 days after the cessation.
//
//  Moreover, your license from a particular copyright holder is
//reinstated permanently if the copyright holder notifies you of the
//violation by some reasonable means, this is the first time you have
//received notice of violation of this License (for any work) from that
//copyright holder, and you cure the violation prior to 30 days after
//your receipt of the notice.
//
//  Termination of your rights under this section does not terminate the
//licenses of parties who have received copies or rights from you under
//this License.  If your rights have been terminated and not permanently
//reinstated, you do not qualify to receive new licenses for the same
//material under section 10.
//
//  9. Acceptance Not Required for Having Copies.
//
//  You are not required to accept this License in order to receive or
//run a copy of the Program.  Ancillary propagation of a covered work
//occurring solely as a consequence of using peer-to-peer transmission
//to receive a copy likewise does not require acceptance.  However,
//nothing other than this License grants you permission to propagate or
//modify any covered work.  These actions infringe copyright if you do
//not accept this License.  Therefore, by modifying or propagating a
//covered work, you indicate your acceptance of this License to do so.
//
//  10. Automatic Licensing of Downstream Recipients.
//
//  Each time you convey a covered work, the recipient automatically
//receives a license from the original licensors, to run, modify and
//propagate that work, subject to this License.  You are not responsible
//for enforcing compliance by third parties with this License.
//
//  An "entity transaction" is a transaction transferring control of an
//organization, or substantially all assets of one, or subdividing an
//organization, or merging organizations.  If propagation of a covered
//work results from an entity transaction, each party to that
//transaction who receives a copy of the work also receives whatever
//licenses to the work the party's predecessor in interest had or could
//give under the previous paragraph, plus a right to possession of the
//Corresponding Source of the work from the predecessor in interest, if
//the predecessor has it or can get it with reasonable efforts.
//
//  You may not impose any further restrictions on the exercise of the
//rights granted or affirmed under this License.  For example, you may
//not impose a license fee, royalty, or other charge for exercise of
//rights granted under this License, and you may not initiate litigation
//(including a cross-claim or counterclaim in a lawsuit) alleging that
//any patent claim is infringed by making, using, selling, offering for
//sale, or importing the Program or any portion of it.
//
//  11. Patents.
//
//  A "contributor" is a copyright holder who authorizes use under this
//License of the Program or a work on which the Program is based.  The
//work thus licensed is called the contributor's "contributor version".
//
//  A contributor's "essential patent claims" are all patent claims
//owned or controlled by the contributor, whether already acquired or
//hereafter acquired, that would be infringed by some manner, permitted
//by this License, of making, using, or selling its contributor version,
//but do not include claims that would be infringed only as a
//consequence of further modification of the contributor version.  For
//purposes of this definition, "control" includes the right to grant
//patent sublicenses in a manner consistent with the requirements of
//this License.
//
//  Each contributor grants you a non-exclusive, worldwide, royalty-free
//patent license under the contributor's essential patent claims, to
//make, use, sell, offer for sale, import and otherwise run, modify and
//propagate the contents of its contributor version.
//
//  In the following three paragraphs, a "patent license" is any express
//agreement or commitment, however denominated, not to enforce a patent
//(such as an express permission to practice a patent or covenant not to
//sue for patent infringement).  To "grant" such a patent license to a
//party means to make such an agreement or commitment not to enforce a
//patent against the party.
//
//  If you convey a covered work, knowingly relying on a patent license,
//and the Corresponding Source of the work is not available for anyone
//to copy, free of charge and under the terms of this License, through a
//publicly available network server or other readily accessible means,
//then you must either (1) cause the Corresponding Source to be so
//available, or (2) arrange to deprive yourself of the benefit of the
//patent license for this particular work, or (3) arrange, in a manner
//consistent with the requirements of this License, to extend the patent
//license to downstream recipients.  "Knowingly relying" means you have
//actual knowledge that, but for the patent license, your conveying the
//covered work in a country, or your recipient's use of the covered work
//in a country, would infringe one or more identifiable patents in that
//country that you have reason to believe are valid.
//
//  If, pursuant to or in connection with a single transaction or
//arrangement, you convey, or propagate by procuring conveyance of, a
//covered work, and grant a patent license to some of the parties
//receiving the covered work authorizing them to use, propagate, modify
//or convey a specific copy of the covered work, then the patent license
//you grant is automatically extended to all recipients of the covered
//work and works based on it.
//
//  A patent license is "discriminatory" if it does not include within
//the scope of its coverage, prohibits the exercise of, or is
//conditioned on the non-exercise of one or more of the rights that are
//specifically granted under this License.  You may not convey a covered
//work if you are a party to an arrangement with a third party that is
//in the business of distributing software, under which you make payment
//to the third party based on the extent of your activity of conveying
//the work, and under which the third party grants, to any of the
//parties who would receive the covered work from you, a discriminatory
//patent license (a) in connection with copies of the covered work
//conveyed by you (or copies made from those copies), or (b) primarily
//for and in connection with specific products or compilations that
//contain the covered work, unless you entered into that arrangement,
//or that patent license was granted, prior to 28 March 2007.
//
//  Nothing in this License shall be construed as excluding or limiting
//any implied license or other defenses to infringement that may
//otherwise be available to you under applicable patent law.
//
//  12. No Surrender of Others' Freedom.
//
//  If conditions are imposed on you (whether by court order, agreement or
//otherwise) that contradict the conditions of this License, they do not
//excuse you from the conditions of this License.  If you cannot convey a
//covered work so as to satisfy simultaneously your obligations under this
//License and any other pertinent obligations, then as a consequence you may
//not convey it at all.  For example, if you agree to terms that obligate you
//to collect a royalty for further conveying from those to whom you convey
//the Program, the only way you could satisfy both those terms and this
//License would be to refrain entirely from conveying the Program.
//
//  13. Use with the GNU Affero General Public License.
//
//  Notwithstanding any other provision of this License, you have
//permission to link or combine any covered work with a work licensed
//under version 3 of the GNU Affero General Public License into a single
//combined work, and to convey the resulting work.  The terms of this
//License will continue to apply to the part which is the covered work,
//but the special requirements of the GNU Affero General Public License,
//section 13, concerning interaction through a network will apply to the
//combination as such.
//
//  14. Revised Versions of this License.
//
//  The Free Software Foundation may publish revised and/or new versions of
//the GNU General Public License from time to time.  Such new versions will
//be similar in spirit to the present version, but may differ in detail to
//address new problems or concerns.
//
//  Each version is given a distinguishing version number.  If the
//Program specifies that a certain numbered version of the GNU General
//Public License "or any later version" applies to it, you have the
//option of following the terms and conditions either of that numbered
//version or of any later version published by the Free Software
//Foundation.  If the Program does not specify a version number of the
//GNU General Public License, you may choose any version ever published
//by the Free Software Foundation.
//
//  If the Program specifies that a proxy can decide which future
//versions of the GNU General Public License can be used, that proxy's
//public statement of acceptance of a version permanently authorizes you
//to choose that version for the Program.
//
//  Later license versions may give you additional or different
//permissions.  However, no additional obligations are imposed on any
//author or copyright holder as a result of your choosing to follow a
//later version.
//
//  15. Disclaimer of Warranty.
//
//  THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
//APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
//HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
//OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
//THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
//PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
//IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
//ALL NECESSARY SERVICING, REPAIR OR CORRECTION.
//
//  16. Limitation of Liability.
//
//  IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
//WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
//THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
//GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
//USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
//DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
//PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
//EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
//SUCH DAMAGES.
//
//  17. Interpretation of Sections 15 and 16.
//
//  If the disclaimer of warranty and limitation of liability provided
//above cannot be given local legal effect according to their terms,
//reviewing courts shall apply local law that most closely approximates
//an absolute waiver of all civil liability in connection with the
//Program, unless a warranty or assumption of liability accompanies a
//copy of the Program in return for a fee.
//
//                     END OF TERMS AND CONDITIONS
//
//            How to Apply These Terms to Your New Programs
//
//  If you develop a new program, and you want it to be of the greatest
//possible use to the public, the best way to achieve this is to make it
//free software which everyone can redistribute and change under these terms.
//
//  To do so, attach the following notices to the program.  It is safest
//to attach them to the start of each source file to most effectively
//state the exclusion of warranty; and each file should have at least
//the "copyright" line and a pointer to where the full notice is found.
//
//    <one line to give the program's name and a brief idea of what it does.>
//    Copyright (C) <year>  <name of author>
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//Also add information on how to contact you by electronic and paper mail.
//
//  If the program does terminal interaction, make it output a short
//notice like this when it starts in an interactive mode:
//
//    <program>  Copyright (C) <year>  <name of author>
//    This program comes with ABSOLUTELY NO WARRANTY; for details type `show w'.
//    This is free software, and you are welcome to redistribute it
//    under certain conditions; type `show c' for details.
//
//The hypothetical commands `show w' and `show c' should show the appropriate
//parts of the General Public License.  Of course, your program's commands
//might be different; for a GUI interface, you would use an "about box".
//
//  You should also get your employer (if you work as a programmer) or school,
//if any, to sign a "copyright disclaimer" for the program, if necessary.
//For more information on this, and how to apply and follow the GNU GPL, see
//<http://www.gnu.org/licenses/>.
//
//  The GNU General Public License does not permit incorporating your program
//into proprietary programs.  If your program is a subroutine library, you
//may consider it more useful to permit linking proprietary applications with
//the library.  If this is what you want to do, use the GNU Lesser General
//Public License instead of this License.  But first, please read
//<http://www.gnu.org/philosophy/why-not-lgpl.html>.
//-------------------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------
#define MODULE_GMP_RALG

#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <time.h>


#include "fcmiof.h"
#include "gmp_ints.h"
#include "gmp_rats.h"
#include "gmp_ralg.h"
#include "intfunc.h"

#if defined(APP_TYPE_SIMPLE_DOS_CONSOLE)
   #include "ccmalloc.h"
#elif defined(APP_TYPE_IJUSCRIPTER_IJUCONSOLE)
   #include "tclalloc.h"
#else
   /* Do nothing. */
#endif


/******************************************************************/
/***  INITIALIZATION AND DESTRUCTION FUNCTIONS  *******************/
/******************************************************************/
//08/16/01: Visual inspection OK.
void GMP_RALG_cfdecomp_init(
            GMP_RALG_cf_app_struct *decomp,
            int                    *failure,
            GMP_INTS_mpz_struct    *num,
            GMP_INTS_mpz_struct    *den)
   {
   int loop_counter, i;
   GMP_INTS_mpz_struct arb_temp1, arb_temp2;

   //Eyeball the input parameters.  The rest of the eyeballing
   //will occur as functions are called to manipulate the
   //numerator and denominator passed in.
   assert(decomp  != NULL);
   assert(failure != NULL);
   assert(num     != NULL);
   assert(den     != NULL);

   //Allocate space for temporary integers.
   GMP_INTS_mpz_init(&arb_temp1);
   GMP_INTS_mpz_init(&arb_temp2);

   //Begin believing no failure.
   *failure = 0;

   //Initialize the copy of the numerator and denominator and
   //copy these into the structure.
   GMP_INTS_mpz_init(&(decomp->num));
   GMP_INTS_mpz_copy(&(decomp->num), num);
   GMP_INTS_mpz_init(&(decomp->den));
   GMP_INTS_mpz_copy(&(decomp->den), den);

   //Allocate the space for the first increment of the
   //growable areas.  We need to use different memory
   //allocation functions depending on whether we're 
   //in a Tcl build or a DOS command-line utility
   //build.
   //Space for partial quotients.
   decomp->a =
      #if defined(APP_TYPE_SIMPLE_DOS_CONSOLE)
         CCMALLOC_malloc(sizeof(GMP_INTS_mpz_struct) * GMP_RALG_CF_ALLOC_INCREMENT);
      #elif defined(APP_TYPE_IJUSCRIPTER_IJUCONSOLE)
         (GMP_INTS_mpz_struct *)
         TclpAlloc(sizeof(GMP_INTS_mpz_struct) * GMP_RALG_CF_ALLOC_INCREMENT);
      #else
         malloc(sizeof(GMP_INTS_mpz_struct) * GMP_RALG_CF_ALLOC_INCREMENT);
      #endif 

   //Dividends.
   decomp->dividend =
      #if defined(APP_TYPE_SIMPLE_DOS_CONSOLE)
         CCMALLOC_malloc(sizeof(GMP_INTS_mpz_struct) * GMP_RALG_CF_ALLOC_INCREMENT);
      #elif defined(APP_TYPE_IJUSCRIPTER_IJUCONSOLE)
         (GMP_INTS_mpz_struct *)
         TclpAlloc(sizeof(GMP_INTS_mpz_struct) * GMP_RALG_CF_ALLOC_INCREMENT);
      #else
         malloc(sizeof(GMP_INTS_mpz_struct) * GMP_RALG_CF_ALLOC_INCREMENT);
      #endif 

   //Divisors.
   decomp->divisor =
      #if defined(APP_TYPE_SIMPLE_DOS_CONSOLE)
         CCMALLOC_malloc(sizeof(GMP_INTS_mpz_struct) * GMP_RALG_CF_ALLOC_INCREMENT);
      #elif defined(APP_TYPE_IJUSCRIPTER_IJUCONSOLE)
         (GMP_INTS_mpz_struct *)
         TclpAlloc(sizeof(GMP_INTS_mpz_struct) * GMP_RALG_CF_ALLOC_INCREMENT);
      #else
         malloc(sizeof(GMP_INTS_mpz_struct) * GMP_RALG_CF_ALLOC_INCREMENT);
      #endif 

   //Remainders.
   decomp->remainder =
      #if defined(APP_TYPE_SIMPLE_DOS_CONSOLE)
         CCMALLOC_malloc(sizeof(GMP_INTS_mpz_struct) * GMP_RALG_CF_ALLOC_INCREMENT);
      #elif defined(APP_TYPE_IJUSCRIPTER_IJUCONSOLE)
         (GMP_INTS_mpz_struct *)
         TclpAlloc(sizeof(GMP_INTS_mpz_struct) * GMP_RALG_CF_ALLOC_INCREMENT);
      #else
         malloc(sizeof(GMP_INTS_mpz_struct) * GMP_RALG_CF_ALLOC_INCREMENT);
      #endif 

   //Convergent numerators.
   decomp->p =
      #if defined(APP_TYPE_SIMPLE_DOS_CONSOLE)
         CCMALLOC_malloc(sizeof(GMP_INTS_mpz_struct) * GMP_RALG_CF_ALLOC_INCREMENT);
      #elif defined(APP_TYPE_IJUSCRIPTER_IJUCONSOLE)
         (GMP_INTS_mpz_struct *)
         TclpAlloc(sizeof(GMP_INTS_mpz_struct) * GMP_RALG_CF_ALLOC_INCREMENT);
      #else
         malloc(sizeof(GMP_INTS_mpz_struct) * GMP_RALG_CF_ALLOC_INCREMENT);
      #endif 

   //Convergent denominators.
   decomp->q =
      #if defined(APP_TYPE_SIMPLE_DOS_CONSOLE)
         CCMALLOC_malloc(sizeof(GMP_INTS_mpz_struct) * GMP_RALG_CF_ALLOC_INCREMENT);
      #elif defined(APP_TYPE_IJUSCRIPTER_IJUCONSOLE)
         (GMP_INTS_mpz_struct *)
         TclpAlloc(sizeof(GMP_INTS_mpz_struct) * GMP_RALG_CF_ALLOC_INCREMENT);
      #else
         malloc(sizeof(GMP_INTS_mpz_struct) * GMP_RALG_CF_ALLOC_INCREMENT);
      #endif 

   //Now the number of allocated slots is what we just allocated.
   decomp->nallocd = GMP_RALG_CF_ALLOC_INCREMENT;

   //The number of slots actually used is zero, to start with.
   decomp->n = 0;

   //There are a number of conditions that will lead to an error
   //where we can't successfully form the continued fraction 
   //decomposition.  These errors are:
   //   a)Either component is NAN.
   //   b)Zero denominator.
   //   c)Either component negative.
   //In these cases, we'll pretend we got 0/1 for the number
   //and set things accordingly, and we'll set the failure
   //flag for the caller.
   //
   if (GMP_INTS_mpz_get_flags(num) 
       ||
       GMP_INTS_mpz_get_flags(den)
       ||
       GMP_INTS_mpz_is_zero(den)
       ||
       GMP_INTS_mpz_is_neg(num)
       ||
       GMP_INTS_mpz_is_neg(den))
      {
      *failure  = 1;
      decomp->n = 1;

      GMP_INTS_mpz_set_ui(&(decomp->num),     0);
      GMP_INTS_mpz_set_ui(&(decomp->den),     1);

      GMP_INTS_mpz_init(decomp->dividend);
      GMP_INTS_mpz_set_ui(decomp->dividend,   0);

      GMP_INTS_mpz_init(decomp->divisor);
      GMP_INTS_mpz_set_ui(decomp->divisor,    1);

      GMP_INTS_mpz_init(decomp->a);
      GMP_INTS_mpz_set_ui(decomp->a,          0);

      GMP_INTS_mpz_init(decomp->remainder);
      GMP_INTS_mpz_set_ui(decomp->remainder,  0);

      GMP_INTS_mpz_init(decomp->p);
      GMP_INTS_mpz_set_ui(decomp->p,          0);

      GMP_INTS_mpz_init(decomp->q);
      GMP_INTS_mpz_set_ui(decomp->q,          1);

      goto return_point;
      }

   //If we're here there are not any errors that we 
   //are willing to detect.  We should be clear
   //for a continued fraction decomposition.
   loop_counter = 0;
   do
      {
      //Increment the count of "rows", because we're
      //about to add one.
      decomp->n++;

      //If we have used up all the space available
      //for integers, we have to allocate more.
      if (decomp->n > decomp->nallocd)
         {
         //We now have more allocated space.
         decomp->nallocd += GMP_RALG_CF_ALLOC_INCREMENT;

         //Be absolutely sure we have not made a greivous
         //error.
         assert(decomp->n <= decomp->nallocd);

         //Space for dividends.
         decomp->dividend =
            #if defined(APP_TYPE_SIMPLE_DOS_CONSOLE)
            CCMALLOC_realloc(
               decomp->dividend,
               sizeof(GMP_INTS_mpz_struct) * decomp->nallocd);
            #elif defined(APP_TYPE_IJUSCRIPTER_IJUCONSOLE)
            (GMP_INTS_mpz_struct *)
            TclpRealloc((char *)decomp->dividend,
                        sizeof(GMP_INTS_mpz_struct) * decomp->nallocd);
            #else
            realloc(decomp->dividend, sizeof(GMP_INTS_mpz_struct) * decomp->nallocd);
            #endif 

         //Space for divisors.
         decomp->divisor =
            #if defined(APP_TYPE_SIMPLE_DOS_CONSOLE)
            CCMALLOC_realloc(
               decomp->divisor,
               sizeof(GMP_INTS_mpz_struct) * decomp->nallocd);
            #elif defined(APP_TYPE_IJUSCRIPTER_IJUCONSOLE)
            (GMP_INTS_mpz_struct *)
            TclpRealloc((char *)decomp->divisor,
                        sizeof(GMP_INTS_mpz_struct) * decomp->nallocd);
            #else
            realloc(decomp->divisor, sizeof(GMP_INTS_mpz_struct) * decomp->nallocd);
            #endif 

         //Space for partial quotients.
         decomp->a =
            #if defined(APP_TYPE_SIMPLE_DOS_CONSOLE)
            CCMALLOC_realloc(
               decomp->a,
               sizeof(GMP_INTS_mpz_struct) * decomp->nallocd);
            #elif defined(APP_TYPE_IJUSCRIPTER_IJUCONSOLE)
            (GMP_INTS_mpz_struct *)
            TclpRealloc((char *)decomp->a,
                        sizeof(GMP_INTS_mpz_struct) * decomp->nallocd);
            #else
            realloc(decomp->a, sizeof(GMP_INTS_mpz_struct) * decomp->nallocd);
            #endif 

         //Space for remainders.
         decomp->remainder =
            #if defined(APP_TYPE_SIMPLE_DOS_CONSOLE)
            CCMALLOC_realloc(
               decomp->remainder,
               sizeof(GMP_INTS_mpz_struct) * decomp->nallocd);
            #elif defined(APP_TYPE_IJUSCRIPTER_IJUCONSOLE)
            (GMP_INTS_mpz_struct *)
            TclpRealloc((char *)decomp->remainder,
                        sizeof(GMP_INTS_mpz_struct) * decomp->nallocd);
            #else
            realloc(decomp->remainder, sizeof(GMP_INTS_mpz_struct) * decomp->nallocd);
            #endif 

         //Space for partial quotient numerators.
         decomp->p =
            #if defined(APP_TYPE_SIMPLE_DOS_CONSOLE)
            CCMALLOC_realloc(
               decomp->p,
               sizeof(GMP_INTS_mpz_struct) * decomp->nallocd);
            #elif defined(APP_TYPE_IJUSCRIPTER_IJUCONSOLE)
            (GMP_INTS_mpz_struct *)
            TclpRealloc((char *)decomp->p,
                        sizeof(GMP_INTS_mpz_struct) * decomp->nallocd);
            #else
            realloc(decomp->p, sizeof(GMP_INTS_mpz_struct) * decomp->nallocd);
            #endif 

         //Space for partial quotient denominators.
         decomp->q =
            #if defined(APP_TYPE_SIMPLE_DOS_CONSOLE)
            CCMALLOC_realloc(
               decomp->q,
               sizeof(GMP_INTS_mpz_struct) * decomp->nallocd);
            #elif defined(APP_TYPE_IJUSCRIPTER_IJUCONSOLE)
            (GMP_INTS_mpz_struct *)
            TclpRealloc((char *)decomp->q,
                        sizeof(GMP_INTS_mpz_struct) * decomp->nallocd);
            #else
            realloc(decomp->q, sizeof(GMP_INTS_mpz_struct) * decomp->nallocd);
            #endif 
         }

      //At this point, we have enough space to do the next round of operations.
      //Set up an index variable.
      i = decomp->n - 1;

      //Initialize all of the integers at this round.
      GMP_INTS_mpz_init(decomp->dividend    + i);
      GMP_INTS_mpz_init(decomp->divisor     + i);
      GMP_INTS_mpz_init(decomp->a           + i);
      GMP_INTS_mpz_init(decomp->remainder   + i);
      GMP_INTS_mpz_init(decomp->p           + i);
      GMP_INTS_mpz_init(decomp->q           + i);

      //Perform the next round of continued fraction decomposition.  This
      //is standard stuff.
      if (i==0)
         {
         //In the 0th round, we essentially perform initial assignments.
         GMP_INTS_mpz_copy(decomp->dividend, &(decomp->num));
         GMP_INTS_mpz_copy(decomp->divisor,  &(decomp->den));

         //Make the division to get quotient and remainder.
         GMP_INTS_mpz_tdiv_qr(decomp->a, decomp->remainder, decomp->dividend, decomp->divisor);

         //The convergents in the first round are always the quotient over 1.
         GMP_INTS_mpz_copy(decomp->p, decomp->a); 
         GMP_INTS_mpz_set_ui(decomp->q, 1);
         }
      else if (i==1)
         {
         //In the 1st round, the only point of caution is that we have to
         //consider p(k-2)/q(k-2) carefully.
         GMP_INTS_mpz_copy(decomp->dividend + 1, decomp->divisor   + 0);
         GMP_INTS_mpz_copy(decomp->divisor  + 1, decomp->remainder + 0);

         //Make the division to get quotient and remainder.
         GMP_INTS_mpz_tdiv_qr(decomp->a          + 1, 
                              decomp->remainder  + 1, 
                              decomp->dividend   + 1, 
                              decomp->divisor    + 1);

         //Need to compute the numerator of the convergent.  This will be:
         //  a(1) p(0) + p(-1) = a(1)p(0) + 1.
         GMP_INTS_mpz_mul(decomp->p + 1, decomp->a + 1, decomp->p + 0);
         GMP_INTS_mpz_set_ui(&arb_temp1, 1);
         GMP_INTS_mpz_add(decomp->p + 1, decomp->p + 1, &arb_temp1);

         //Need to compute the denominator of the convergent.  This will
         //be a(1)q(0) + q(-1) = a(1) q(0) = a(1).
         GMP_INTS_mpz_copy(decomp->q + 1, decomp->a + 1);
         }
      else
         {
         //In the general case, it is a simple formula.
         //Rotate in the dividend and divisor from the previous round.
         GMP_INTS_mpz_copy(decomp->dividend + i, decomp->divisor   + i - 1);
         GMP_INTS_mpz_copy(decomp->divisor  + i, decomp->remainder + i - 1);

         //Make the division to get quotient and remainder.
         GMP_INTS_mpz_tdiv_qr(decomp->a          + i, 
                              decomp->remainder  + i, 
                              decomp->dividend   + i, 
                              decomp->divisor    + i);

         //Need to compute the numerator of the convergent.  This will be:
         //  p(i) = a(i)p(i-1) + p(i-2)
         GMP_INTS_mpz_mul(decomp->p + i, decomp->a + i, decomp->p + i - 1);
         GMP_INTS_mpz_add(decomp->p + i, decomp->p + i, decomp->p + i - 2);

         //Need to compute the numerator of the convergent.  This will be:
         //  q(i) = q(i)q(i-1) + q(i-2)
         GMP_INTS_mpz_mul(decomp->q + i, decomp->a + i, decomp->q + i - 1);
         GMP_INTS_mpz_add(decomp->q + i, decomp->q + i, decomp->q + i - 2);
         }

      loop_counter++;
      } while(!GMP_INTS_mpz_is_zero(decomp->remainder + decomp->n - 1) && loop_counter < 100000);

   //In debug builds, be sure we did not terminate based on the loop counter.
   assert(loop_counter != 100000);

   return_point:

   //Deallocate space for temporary integers.
   GMP_INTS_mpz_clear(&arb_temp1);
   GMP_INTS_mpz_clear(&arb_temp2);
   }


//08/16/01: Visual inspection OK.
void GMP_RALG_cfdecomp_destroy(
            GMP_RALG_cf_app_struct *decomp
            )
   {
   int i;

   //Eyeball the input parameters.  Other eyeballing
   //will be done as integers are destroyed.
   assert(decomp != NULL);

   //First, destroy the things that are bound directly
   //to the record.
   GMP_INTS_mpz_clear(&(decomp->num));
   GMP_INTS_mpz_clear(&(decomp->den));

   //Now, destroy every integer which is allocated.
   for (i=0; i < decomp->n; i++)
      {
      GMP_INTS_mpz_clear(decomp->dividend     + i); 
      GMP_INTS_mpz_clear(decomp->divisor      + i); 
      GMP_INTS_mpz_clear(decomp->a            + i); 
      GMP_INTS_mpz_clear(decomp->remainder    + i); 
      GMP_INTS_mpz_clear(decomp->p            + i); 
      GMP_INTS_mpz_clear(decomp->q            + i); 
      }

   //Now, destroy the arrays of integers.
   #if defined(APP_TYPE_SIMPLE_DOS_CONSOLE)
      CCMALLOC_free(decomp->dividend);
   #elif defined(APP_TYPE_IJUSCRIPTER_IJUCONSOLE)
      TclpFree((char *)decomp->dividend);
   #else
      free(decomp->dividend);
   #endif 
   #if defined(APP_TYPE_SIMPLE_DOS_CONSOLE)
      CCMALLOC_free(decomp->divisor);
   #elif defined(APP_TYPE_IJUSCRIPTER_IJUCONSOLE)
      TclpFree((char *)decomp->divisor);
   #else
      free(decomp->divisor);
   #endif 
   #if defined(APP_TYPE_SIMPLE_DOS_CONSOLE)
      CCMALLOC_free(decomp->a);
   #elif defined(APP_TYPE_IJUSCRIPTER_IJUCONSOLE)
      TclpFree((char *)decomp->a);
   #else
      free(decomp->a);
   #endif 
   #if defined(APP_TYPE_SIMPLE_DOS_CONSOLE)
      CCMALLOC_free(decomp->remainder);
   #elif defined(APP_TYPE_IJUSCRIPTER_IJUCONSOLE)
      TclpFree((char *)decomp->remainder);
   #else
      free(decomp->remainder);
   #endif 
   #if defined(APP_TYPE_SIMPLE_DOS_CONSOLE)
      CCMALLOC_free(decomp->p);
   #elif defined(APP_TYPE_IJUSCRIPTER_IJUCONSOLE)
      TclpFree((char *)decomp->p);
   #else
      free(decomp->p);
   #endif 
   #if defined(APP_TYPE_SIMPLE_DOS_CONSOLE)
      CCMALLOC_free(decomp->q);
   #elif defined(APP_TYPE_IJUSCRIPTER_IJUCONSOLE)
      TclpFree((char *)decomp->q);
   #else
      free(decomp->q);
   #endif 
   }


/******************************************************************/
/***  FORMATTED OUTPUT FUNCTIONS  *********************************/
/******************************************************************/
//08/16/01: Visual inspection OK.
void GMP_RALG_cfdecomp_emit(
            FILE                        *s,
            char                        *banner,
            GMP_RALG_cf_app_struct      *decomp,
            int                          nf,
            int                          dap,
            const GMP_INTS_mpz_struct   *dap_denominator)
   {
   int i;

   GMP_INTS_mpz_struct arb_temp, arb_quotient, arb_remainder;

   //Eyeball the input parameters.  The banner is allowed to
   //be null, so can't check that.
   assert(s != NULL);
   assert(decomp != NULL);

   //Allocate our temporary integers.
   GMP_INTS_mpz_init(&arb_temp);
   GMP_INTS_mpz_init(&arb_quotient);
   GMP_INTS_mpz_init(&arb_remainder);

   //If banner requested and noformat option not used.
   if (banner && !nf)
      {
      FCMIOF_stream_bannerheading(s, banner, 1);
      }

   //Dump the input numerator.
   if (!nf)
      {
      GMP_INTS_mpz_long_int_format_to_stream(s,
                                             &(decomp->num),
                                             "Input Numerator");
      }
   else
      {
      GMP_INTS_mpz_arb_int_raw_to_stream(s, &(decomp->num));
      fprintf(s, "\n");
      }

   //Separator if not in unformatted mode.
   if (!nf)
      FCMIOF_stream_hline(s);

   //Dump the input denominator.
   if (!nf)
      {
      GMP_INTS_mpz_long_int_format_to_stream(s,
                                             &(decomp->den),
                                             "Input Denominator");
      }
   else
      {
      GMP_INTS_mpz_arb_int_raw_to_stream(s, &(decomp->den));
      fprintf(s, "\n");
      }

   //Separator if not in unformatted mode.
   if (!nf)
      FCMIOF_stream_hline(s);

   for (i=0; i<decomp->n; i++)
      {
      char strbuf[100];
         //Buffer to prepare description.

      //Print out the dividend at each round.
      if (!nf)
         {
         sprintf(strbuf, "dividend(%d)", i);
         GMP_INTS_mpz_long_int_format_to_stream(s,
                                                decomp->dividend + i,
                                                strbuf);
         }
      else
         {
         GMP_INTS_mpz_arb_int_raw_to_stream(s, decomp->dividend+i);
         fprintf(s, "\n");
         }

      //Separator if not in unformatted mode.
      if (!nf)
         FCMIOF_stream_hline(s);

      //Print out the divisor at each round.
      if (!nf)
         {
         sprintf(strbuf, "divisor(%d)", i);
         GMP_INTS_mpz_long_int_format_to_stream(s,
                                                decomp->divisor + i,
                                                strbuf);
         }
      else
         {
         GMP_INTS_mpz_arb_int_raw_to_stream(s, decomp->divisor+i);
         fprintf(s, "\n");
         }

      //Separator if not in unformatted mode.
      if (!nf)
         FCMIOF_stream_hline(s);

      //Print out partial quotient at each round.
      if (!nf)
         {
         sprintf(strbuf, "a(%d)", i);
         GMP_INTS_mpz_long_int_format_to_stream(s,
                                                decomp->a + i,
                                                strbuf);
         }
      else
         {
         GMP_INTS_mpz_arb_int_raw_to_stream(s, decomp->a+i);
         fprintf(s, "\n");
         }

      //Separator if not in unformatted mode.
      if (!nf)
         FCMIOF_stream_hline(s);

      //It doesn't make any sense to print out the 
      //remainder, because this becomes the divisor
      //for the next round.  It is just wasted output
      //lines.

      //Print out the convergent numerator at
      //each round.
      if (!nf)
         {
         sprintf(strbuf, "p(%d)", i);
         GMP_INTS_mpz_long_int_format_to_stream(s,
                                                decomp->p + i,
                                                strbuf);
         }
      else
         {
         GMP_INTS_mpz_arb_int_raw_to_stream(s, decomp->p+i);
         fprintf(s, "\n");
         }

      //Separator if not in unformatted mode.
      if (!nf)
         FCMIOF_stream_hline(s);

      //Print out the convergent denominator at
      //each round.
      if (!nf)
         {
         sprintf(strbuf, "q(%d)", i);
         GMP_INTS_mpz_long_int_format_to_stream(s,
                                                decomp->q + i,
                                                strbuf);
         }
      else
         {
         GMP_INTS_mpz_arb_int_raw_to_stream(s, decomp->q+i);
         fprintf(s, "\n");
         }

      //Separator if not in unformatted mode.
      if (!nf)
         FCMIOF_stream_hline(s);

      if (dap)
         {
         //Calculate the DAP numerator 
         GMP_INTS_mpz_mul(&arb_temp, dap_denominator, decomp->p + i);
         GMP_INTS_mpz_tdiv_qr(&arb_quotient, &arb_remainder,
                              &arb_temp, decomp->q + i);

         //Print DAP numerator.
         if (!nf)
            {
            sprintf(strbuf, "dap_h(%d)", i);
            GMP_INTS_mpz_long_int_format_to_stream(s,
                                                   &arb_quotient,
                                                   strbuf);
            }
         else
            {
            GMP_INTS_mpz_arb_int_raw_to_stream(s, &arb_quotient);
            fprintf(s, "\n");
            }

         //Separator if not in unformatted mode.
         if (!nf)
            FCMIOF_stream_hline(s);

         //Print DAP denominator.
         if (!nf)
            {
            sprintf(strbuf, "dap_k(%d)", i);
            GMP_INTS_mpz_long_int_format_to_stream(s,
                                                   dap_denominator,
                                                   strbuf);
            }
         else
            {
            GMP_INTS_mpz_arb_int_raw_to_stream(s, dap_denominator);
            fprintf(s, "\n");
            }

         //Separator if not in unformatted mode.
         if (!nf)
            FCMIOF_stream_hline(s);
         }
      }

   //Deallocate our temporary integers.
   GMP_INTS_mpz_clear(&arb_temp);
   GMP_INTS_mpz_clear(&arb_quotient);
   GMP_INTS_mpz_clear(&arb_remainder);
   }


/******************************************************************/
/***  FAREY SERIES PREDECESSOR AND SUCCESSOR FUNCTIONS  ***********/
/******************************************************************/
//08/16/01: Visual inspection OK.
void GMP_RALG_farey_predecessor(
                    GMP_RATS_mpq_struct *result,
              const GMP_RATS_mpq_struct *plus_two,
              const GMP_RATS_mpq_struct *plus_one,
              const GMP_INTS_mpz_struct *N)
   {
   GMP_RATS_mpq_struct result_copy;
      //Used to hold return value in case the result
      //is the same as either of the input arguments.
   GMP_INTS_mpz_struct temp1, temp2, floor_func;
      //Temporary integers.

   assert(result   != NULL);
   assert(plus_two != NULL);
   assert(plus_one != NULL);
   assert(N        != NULL);

   //Initialize the variables used.
   GMP_RATS_mpq_init(&result_copy);
   GMP_INTS_mpz_init(&temp1);
   GMP_INTS_mpz_init(&temp2);
   GMP_INTS_mpz_init(&floor_func);

   //Numerator of the term in the floor function.
   GMP_INTS_mpz_add(&temp1, &(plus_two->den), N);

   //Term in the floor function.  This is used to 
   //calculate both numerator and denominator, so we save it.
   GMP_INTS_mpz_tdiv_qr(&floor_func, &temp2, &temp1, &(plus_one->den));

   //Product of result of floor function and numerator--now
   //forming the numerator of the output.
   GMP_INTS_mpz_mul(&temp2, &floor_func, &(plus_one->num));

   //Final result assigned to numerator.
   GMP_INTS_mpz_sub(&(result_copy.num), &temp2, &(plus_two->num));

   //Product of result of floor function and denominator--now
   //forming the denominator of the output.
   GMP_INTS_mpz_mul(&temp2, &floor_func, &(plus_one->den));

   //Final result assigned to denominator.
   GMP_INTS_mpz_sub(&(result_copy.den), &temp2, &(plus_two->den));

   //Copy the result to the object owned by the caller.
   GMP_RATS_mpq_copy(result, &result_copy);

   //Deallocate dynamic memory.
   GMP_RATS_mpq_clear(&result_copy);
   GMP_INTS_mpz_clear(&temp1);
   GMP_INTS_mpz_clear(&temp2);
   GMP_INTS_mpz_clear(&floor_func);
   }


//08/16/01: Visual inspection OK.
void GMP_RALG_farey_successor(
                    GMP_RATS_mpq_struct *result,
              const GMP_RATS_mpq_struct *minus_two,
              const GMP_RATS_mpq_struct *minus_one,
              const GMP_INTS_mpz_struct *N)
   {
   GMP_RATS_mpq_struct result_copy;
      //Used to hold return value in case the result
      //is the same as either of the input arguments.
   GMP_INTS_mpz_struct temp1, temp2, floor_func;
      //Temporary integers.

   assert(result   != NULL);
   assert(minus_two != NULL);
   assert(minus_one != NULL);
   assert(N        != NULL);

   //Initialize the variables used.
   GMP_RATS_mpq_init(&result_copy);
   GMP_INTS_mpz_init(&temp1);
   GMP_INTS_mpz_init(&temp2);
   GMP_INTS_mpz_init(&floor_func);

   //Numerator of the term in the floor function.
   GMP_INTS_mpz_add(&temp1, &(minus_two->den), N);

   //Term in the floor function.  This is used to 
   //calculate both numerator and denominator, so we save it.
   GMP_INTS_mpz_tdiv_qr(&floor_func, &temp2, &temp1, &(minus_one->den));

   //Product of result of floor function and numerator--now
   //forming the numerator of the output.
   GMP_INTS_mpz_mul(&temp2, &floor_func, &(minus_one->num));

   //Final result assigned to numerator.
   GMP_INTS_mpz_sub(&(result_copy.num), &temp2, &(minus_two->num));

   //Product of result of floor function and denominator--now
   //forming the denominator of the output.
   GMP_INTS_mpz_mul(&temp2, &floor_func, &(minus_one->den));

   //Final result assigned to denominator.
   GMP_INTS_mpz_sub(&(result_copy.den), &temp2, &(minus_two->den));

   //Copy the result to the object owned by the caller.
   GMP_RATS_mpq_copy(result, &result_copy);

   //Deallocate dynamic memory.
   GMP_RATS_mpq_clear(&result_copy);
   GMP_INTS_mpz_clear(&temp1);
   GMP_INTS_mpz_clear(&temp2);
   GMP_INTS_mpz_clear(&floor_func);
   }


//08/16/01: Visual inspection OK.
void GMP_RALG_enclosing_farey_neighbors(
              const GMP_RATS_mpq_struct    *rn_in,
              const GMP_INTS_mpz_struct    *N,
              const GMP_RALG_cf_app_struct *cf_rep,
              int                          *equality,
                    GMP_RATS_mpq_struct    *left,
                    GMP_RATS_mpq_struct    *right)
   {
   GMP_RATS_mpq_struct rn_abs;
      //Absolute value of rational number supplied.
   GMP_RATS_mpq_struct previous_convergent;
      //Convergent before the one that has the denominator
      //not exceeding the order of the series.  Need to fudge
      //a little bit because don't have -1-th order convergents
      //tabulated.
   GMP_RATS_mpq_struct other_neighbor;
      //The other neighbor besides the highest-order convergent
      //without denominator too large.
   GMP_INTS_mpz_struct temp1, temp2, temp3, temp4;
      //Temporary integers.
   int ho_conv;
      //Index of highest-ordered convergent that does not have
      //denominator too large.

   //Eyeball the parameters.
   assert(rn_in     != NULL);
   assert(N         != NULL);
   assert(cf_rep    != NULL);
   assert(equality  != NULL);
   assert(left      != NULL);
   assert(right     != NULL);

   //Allocate dynamic variables.
   GMP_RATS_mpq_init(&rn_abs);
   GMP_RATS_mpq_init(&previous_convergent);
   GMP_RATS_mpq_init(&other_neighbor);
   GMP_INTS_mpz_init(&temp1);
   GMP_INTS_mpz_init(&temp2);
   GMP_INTS_mpz_init(&temp3);
   GMP_INTS_mpz_init(&temp4);

   //Zero is a troublesome case, because it requires us to
   //cross signs.  Split this case out explicitly.
   if (GMP_INTS_mpz_is_zero(&(rn_in->num)))
      {
      *equality = 1;  //0/1 a member of Farey series of any order.
      GMP_INTS_mpz_set_si(&(left->num), -1);
      GMP_INTS_mpz_copy(&(left->den), N);
      GMP_INTS_mpz_set_si(&(right->num), 1);
      GMP_INTS_mpz_copy(&(right->den), N);
      }
   else
      {
      //Make a copy of the rational number in.  As a condition of
      //using this function, it must be normalized, but it still
      //may be negative.  Our strategy is to treat the number as
      //positive, find the neighbors, then if it was negative
      //complement and re-order the neighbors.  In other words,
      //find neighbors to a negative number by symmetry, not
      //by forming the CF representation of a negative number.
      //Also, we can't touch the input parameter.
      GMP_RATS_mpq_copy(&rn_abs, rn_in);
      GMP_INTS_mpz_abs(&(rn_abs.num));

      //Find the index of the highest-ordered convergent
      //with a denominator not exceeding the denominator of
      //the rational number supplied.  The zero'th order
      //convergent has a denominator of 1, so that one
      //at least is safe.

      //Assign either the "left" or right
      //neighbor to be the highest-ordered
      //convergent with a denominator not exceeding the
      //denominator of the rational number input.  I say 
      //"either" because the properties of convergents let
      //us know based on the oddness or evenness of the order
      //which side it is on.
      ho_conv = 0;
      while (((ho_conv + 1) < cf_rep->n) && (GMP_INTS_mpz_cmp(cf_rep->q + ho_conv + 1, N) <= 0))
         {
         #if 0
         //Some questions about this loop--debugging output.
         printf("ho_conv : %d\n", ho_conv);
         GMP_INTS_mpz_long_int_format_to_stream(stdout,
                                                cf_rep->q + ho_conv + 1,
                                                "decomp_den");
         GMP_INTS_mpz_long_int_format_to_stream(stdout,
                                                &(rn_abs.den),
                                                "rn_in_den");
         printf("Compare result: %d\n\n", GMP_INTS_mpz_cmp(cf_rep->q + ho_conv + 1, &(rn_abs.den)));
         #endif

         ho_conv++;
         }

      if (INTFUNC_is_even(ho_conv))
         {
         GMP_INTS_mpz_copy(&(left->num), cf_rep->p + ho_conv);
         GMP_INTS_mpz_copy(&(left->den), cf_rep->q + ho_conv);
         }
      else
         {
         GMP_INTS_mpz_copy(&(right->num), cf_rep->p + ho_conv);
         GMP_INTS_mpz_copy(&(right->den), cf_rep->q + ho_conv);
         }

      //Now, we need to calculate the other neighbor based
      //on the standard formula.  This is a little tricky
      //because we don't have the -1-th order convergents
      //tabulated so we have to fudge a little bit.
      if (ho_conv == 0)
         {
         GMP_RATS_mpq_set_si(&previous_convergent, 1, 0);
         }
      else
         {
         GMP_INTS_mpz_copy(&(previous_convergent.num), cf_rep->p + ho_conv - 1);
         GMP_INTS_mpz_copy(&(previous_convergent.den), cf_rep->q + ho_conv - 1);
         }

      //Calculate the other neighbor according to the standard
      //formula.
      GMP_INTS_mpz_sub(&temp1, N, &(previous_convergent.den));
      GMP_INTS_mpz_tdiv_qr(&temp2, &temp3, &temp1, cf_rep->q + ho_conv);
      //temp2 now contains term from floor() function in formula.
      GMP_INTS_mpz_mul(&temp1, &temp2, cf_rep->p + ho_conv);
      GMP_INTS_mpz_add(&(other_neighbor.num), &temp1, &(previous_convergent.num));
      GMP_INTS_mpz_mul(&temp1, &temp2, cf_rep->q + ho_conv);
      GMP_INTS_mpz_add(&(other_neighbor.den), &temp1, &(previous_convergent.den));

      //Copy the other neighbor into the right slot.
      if (INTFUNC_is_even(ho_conv))
         {
         GMP_RATS_mpq_copy(right, &other_neighbor);
         }
      else
         {
         GMP_RATS_mpq_copy(left, &other_neighbor);
         }

      //Set the equality flag.  We have equality if and only
      //if the denominator of the rational number is less than
      //or equal to N.
      if (GMP_INTS_mpz_cmp(&(rn_abs.den), N) <= 0)
         {
         *equality = 1;
         }
      else
         {
         *equality = 0;
         }

      //In the event of equality, we don't really have the true
      //neighbors.  If the final convergent is even-ordered,
      //the left needs to be replaced.  If the final convergent
      //is odd-ordered, the right needs to be replaced.
      if (*equality)
         {
         if (INTFUNC_is_even(ho_conv))
            {
            //Left needs to be replaced.
            GMP_RALG_farey_predecessor(
                    left,
                    right,
                    &rn_abs,
                    N);
            }
         else
            {
            //Right needs to be replaced.
            GMP_RALG_farey_successor(
                    right,
                    left,
                    &rn_abs,
                    N);
            }
         }

      //OK, we should be all done.  The final catch is that if
      //the number supplied was negative, we need to invert
      //and re-order the neighbors.
      if (GMP_INTS_mpz_is_neg(&(rn_in->num)))
         {
         GMP_RATS_mpq_swap(left, right);
         GMP_INTS_mpz_negate(&(left->num));
         GMP_INTS_mpz_negate(&(right->num));
         }
      } //End if (rn==0) else clause

   //Deallocate dynamic variables.
   GMP_RATS_mpq_clear(&rn_abs);
   GMP_RATS_mpq_clear(&previous_convergent);
   GMP_RATS_mpq_clear(&other_neighbor);
   GMP_INTS_mpz_clear(&temp1);
   GMP_INTS_mpz_clear(&temp2);
   GMP_INTS_mpz_clear(&temp3);
   GMP_INTS_mpz_clear(&temp4);
   }



//08/16/01: Visual inspection OK.  Did not fully inspect the
//iterative part of this function.  Unit testing will be
//careful, expect that to catch any anomalies.
void GMP_RALG_consecutive_fab_terms(
                                    const GMP_RATS_mpq_struct *rn_in,
                                    const GMP_INTS_mpz_struct *kmax,
                                    const GMP_INTS_mpz_struct *hmax,
                                    int n_left_in,
                                    int n_right_in,
                                    GMP_RALG_fab_neighbor_collection_struct *result
                                    )
   {
   int   error_flag, equality_flag;
   char *estring_kmax_neg = "KMAX is zero, negative, or NAN.";
   char *estring_hmax_neg = "HMAX is negative or NAN.";
   char *estring_general  = "Unspecified general error in GMP_RALG_consecutive_fab_terms().";

   GMP_RATS_mpq_struct q_temp1, q_temp2, q_temp3, q_temp4, 
                       left_neighbor, right_neighbor, 
                       left_neighbor_abs, right_neighbor_abs,
                       hmax_over_one_neg, corner_point_neg,
                       abs_norm_recip_rn;

   //Eyeball input parameters.
   assert(rn_in      != NULL);
   assert(kmax       != NULL);
   assert(n_left_in  >= 0);
   assert(n_left_in  <= 0x00FFFFFF);
   assert(n_right_in >= 0);
   assert(n_right_in <= 0x00FFFFFF);
   assert(result     != NULL);

   //Allocate all of the dynamic memory we'll need for this function.  It will be
   //released at the end.
   GMP_RATS_mpq_init(&q_temp1);
   GMP_RATS_mpq_init(&q_temp2);
   GMP_RATS_mpq_init(&q_temp3);
   GMP_RATS_mpq_init(&q_temp4);
   GMP_RATS_mpq_init(&left_neighbor);
   GMP_RATS_mpq_init(&right_neighbor);
   GMP_RATS_mpq_init(&left_neighbor_abs);
   GMP_RATS_mpq_init(&right_neighbor_abs);
   GMP_RATS_mpq_init(&hmax_over_one_neg);
   GMP_RATS_mpq_init(&corner_point_neg);
   GMP_RATS_mpq_init(&abs_norm_recip_rn);

   //Zero out the result block.  This is the easiest way to give many variables
   //default values of 0, FALSE, and NULL.
   memset(result, 0, sizeof(GMP_RALG_fab_neighbor_collection_struct));

   //Allocate all integer and rational number structures in the result block.
   GMP_RATS_mpq_init(&(result->rn_in));
   GMP_INTS_mpz_init(&(result->kmax_in));
   GMP_INTS_mpz_init(&(result->hmax_in));
   GMP_RATS_mpq_init(&(result->hmax_over_one));
   GMP_RATS_mpq_init(&(result->corner_point));
   GMP_RATS_mpq_init(&(result->corner_point_minus_one));
   GMP_RATS_mpq_init(&(result->corner_point_plus_one));
   GMP_RATS_mpq_init(&(result->norm_rn));
   GMP_RATS_mpq_init(&(result->abs_norm_rn));

   //Fill in the rational number, exactly as passed.
   GMP_RATS_mpq_copy(&(result->rn_in), rn_in);

   //Fill in the number of left and right neighbors that the caller wants.
   //However, let's of course say nothing less than zero and nothing more
   //than 10000 neighbors on either side.
   result->n_left_in  = INTFUNC_min(INTFUNC_max(0, n_left_in), 10000);
   result->n_right_in = INTFUNC_min(INTFUNC_max(0, n_right_in), 10000);

   //Fill in the value of KMAX, exactly as passed.  If it is not at least
   //the value of 1 or if error flags, croak.
   GMP_INTS_mpz_copy(&(result->kmax_in), kmax);
   if (GMP_INTS_mpz_get_flags(kmax) || GMP_INTS_mpz_is_zero(kmax) || GMP_INTS_mpz_is_neg(kmax))
      {
      result->error = 
         #if defined(APP_TYPE_SIMPLE_DOS_CONSOLE)
            CCMALLOC_malloc(sizeof(char) * (strlen(estring_kmax_neg) + 1));
         #elif defined(APP_TYPE_IJUSCRIPTER_IJUCONSOLE)
            TclpAlloc(sizeof(char) * (strlen(estring_kmax_neg) + 1));
         #else
            malloc(sizeof(char) * (strlen(estring_kmax_neg) + 1));
         #endif 
      strcpy(result->error, estring_kmax_neg);
      goto return_point;
      }

   //Decide whether the caller intends to use HMAX.  Neg is error, but zero
   //is a signal that don't intend to use.
   if (hmax)
      {
      GMP_INTS_mpz_copy(&(result->hmax_in), hmax);
      if (GMP_INTS_mpz_get_flags(hmax) || GMP_INTS_mpz_is_neg(hmax))
         {
         result->error = 
            #if defined(APP_TYPE_SIMPLE_DOS_CONSOLE)
               CCMALLOC_malloc(sizeof(char) * (strlen(estring_hmax_neg) + 1));
            #elif defined(APP_TYPE_IJUSCRIPTER_IJUCONSOLE)
               TclpAlloc(sizeof(char) * (strlen(estring_hmax_neg) + 1));
            #else
               malloc(sizeof(char) * (strlen(estring_hmax_neg) + 1));
            #endif 
         strcpy(result->error, estring_hmax_neg);
         goto return_point;
         }
      else if (GMP_INTS_mpz_is_pos(hmax))
         {
         result->hmax_supplied = 1;
         }
      }

   //If HMAX has been supplied, assign and normalize the
   //corner point.
   if (result->hmax_supplied)
      {
      GMP_INTS_mpz_copy(&(result->corner_point.num), &(result->hmax_in));
      GMP_INTS_mpz_copy(&(result->corner_point.den), &(result->kmax_in));
      GMP_RATS_mpq_normalize(&(result->corner_point));
      }

   //If HMAX has been supplied, we want to get the continued
   //fraction representation of both the corner point and its
   //reciprocal.  This is because we're going to need to
   //find its adjacent points so we can easily crawl
   //around a rectangular region of the integer lattice.
   if (result->hmax_supplied)
      {
      //CF representation of corner point.
      GMP_RALG_cfdecomp_init(&(result->corner_point_cf_rep),
                             &error_flag,
                             &(result->corner_point.num),
                             &(result->corner_point.den));
      result->corner_point_cf_rep_formed = 1;

      //If there was an error forming the CF representation
      //of the corner point, bail out.
      if (error_flag)
         {
         result->error = 
            #if defined(APP_TYPE_SIMPLE_DOS_CONSOLE)
               CCMALLOC_malloc(sizeof(char) * (strlen(estring_general) + 1));
            #elif defined(APP_TYPE_IJUSCRIPTER_IJUCONSOLE)
               TclpAlloc(sizeof(char) * (strlen(estring_general) + 1));
            #else
               malloc(sizeof(char) * (strlen(estring_general) + 1));
            #endif 
         strcpy(result->error, estring_general);
         goto return_point;
         }

      //CF representation of reciprocal of corner point.
      GMP_RALG_cfdecomp_init(&(result->corner_point_recip_cf_rep),
                             &error_flag,
                             &(result->corner_point.den),
                             &(result->corner_point.num));
      result->corner_point_recip_cf_rep_formed = 1;

      //If there was an error forming the CF representation
      //of the reciprocal of the corner point, bail out.
      if (error_flag)
         {
         result->error = 
            #if defined(APP_TYPE_SIMPLE_DOS_CONSOLE)
               CCMALLOC_malloc(sizeof(char) * (strlen(estring_general) + 1));
            #elif defined(APP_TYPE_IJUSCRIPTER_IJUCONSOLE)
               TclpAlloc(sizeof(char) * (strlen(estring_general) + 1));
            #else
               malloc(sizeof(char) * (strlen(estring_general) + 1));
            #endif 
         strcpy(result->error, estring_general);
         goto return_point;
         }
      }

   //Normalize the rational number supplied.
   GMP_RATS_mpq_copy(&(result->norm_rn), rn_in);
   GMP_RATS_mpq_normalize(&(result->norm_rn));

   //Form the absolute value of the normalized
   //version, and set the neg flag.
   GMP_RATS_mpq_copy(&(result->abs_norm_rn),&(result->norm_rn));
   if (GMP_INTS_mpz_is_neg(&(result->abs_norm_rn.num)))
      {
      GMP_INTS_mpz_negate(&(result->abs_norm_rn.num));
      result->rn_is_neg = 1;
      }

   //Form the continued fraction representation of the
   //absolute value of the rational number supplied.
   //This is always required, because we cannot get any
   //neighbors without it.
   GMP_RALG_cfdecomp_init(&(result->rn_abs_cf_rep),
                          &error_flag,
                          &(result->abs_norm_rn.num),
                          &(result->abs_norm_rn.den));
   result->rn_abs_cf_rep_formed = 1;

   //If there was an error forming the CF representation
   //of the absolute value of rational number supplied, bail out.
   if (error_flag)
      {
      result->error = 
         #if defined(APP_TYPE_SIMPLE_DOS_CONSOLE)
            CCMALLOC_malloc(sizeof(char) * (strlen(estring_general) + 1));
         #elif defined(APP_TYPE_IJUSCRIPTER_IJUCONSOLE)
            TclpAlloc(sizeof(char) * (strlen(estring_general) + 1));
         #else
            malloc(sizeof(char) * (strlen(estring_general) + 1));
         #endif 
      strcpy(result->error, estring_general);
      goto return_point;
      }

   //Set the equality flag.  The rational number supplied is
   //in the series of interest if and only if, in its normalized
   //form, K <= KMAX, and if HMAX was supplied, H <= HMAX.
   if (GMP_INTS_mpz_cmp(&(result->abs_norm_rn.den), kmax) <= 0)
      {
      if (result->hmax_supplied)
         {
         if (GMP_INTS_mpz_cmp(&(result->abs_norm_rn.num), hmax) <= 0)
            {
            result->equality = 1;
            }
         else
            {
            result->equality = 0;
            }
         }
      else
         {
         result->equality = 1;
         }
      }
   else
      {
      result->equality = 0;
      }

   //The final cause of error is if the rational number
   //supplied is outside the interval [-HMAX/1, HMAX/1].
   //In such cases, simply refuse to calculate
   //any approximations.  This error can only occur
   //if HMAX is specified.  If only KMAX is specified,
   //this error cannot occur.
   if (result->hmax_supplied)
      {
      //Form the rational number HMAX/1.  We will use it for
      //a comparison.
      GMP_INTS_mpz_copy(&(result->hmax_over_one.num), hmax);
      GMP_INTS_mpz_set_ui(&(result->hmax_over_one.den), 1);

      //If the comparison shows that the absolute value of
      //the rational number in is larger than HMAX over 1,
      //then declare an error.
      if (GMP_RATS_mpq_cmp(&(result->abs_norm_rn),&(result->hmax_over_one),NULL) > 0)
         {
         result->error = 
            #if defined(APP_TYPE_SIMPLE_DOS_CONSOLE)
               CCMALLOC_malloc(sizeof(char) * (strlen(estring_general) + 1));
            #elif defined(APP_TYPE_IJUSCRIPTER_IJUCONSOLE)
               TclpAlloc(sizeof(char) * (strlen(estring_general) + 1));
            #else
               malloc(sizeof(char) * (strlen(estring_general) + 1));
            #endif 
         strcpy(result->error, estring_general);
         goto return_point;
         }
      }

   //If we're here, we're very much clean.  The only thing
   //that could go wrong is an overflow.

   //Allocate space for the left and right arrays of 
   //neighbors.
   if (result->n_left_in)
      {
      result->lefts = 
         #if defined(APP_TYPE_SIMPLE_DOS_CONSOLE)
            CCMALLOC_malloc(sizeof(GMP_RALG_fab_neighbor_struct) * result->n_left_in);
         #elif defined(APP_TYPE_IJUSCRIPTER_IJUCONSOLE)
            (GMP_RALG_fab_neighbor_struct *)
            TclpAlloc(sizeof(GMP_RALG_fab_neighbor_struct) * result->n_left_in);
         #else
            malloc(sizeof(GMP_RALG_fab_neighbor_struct) * result->n_left_in);
         #endif 
      }

   if (result->n_right_in)
      {
      result->rights = 
         #if defined(APP_TYPE_SIMPLE_DOS_CONSOLE)
            CCMALLOC_malloc(sizeof(GMP_RALG_fab_neighbor_struct) * result->n_right_in);
         #elif defined(APP_TYPE_IJUSCRIPTER_IJUCONSOLE)
            (GMP_RALG_fab_neighbor_struct *)
            TclpAlloc(sizeof(GMP_RALG_fab_neighbor_struct) * result->n_right_in);
         #else
            malloc(sizeof(GMP_RALG_fab_neighbor_struct) * result->n_right_in);
         #endif 
      }

   //If the rational number supplied is above the corner
   //point, we want to form the continued fraction representation
   //of its reciprocal.
   if (result->hmax_supplied)
      {
      if (GMP_RATS_mpq_cmp(&(result->abs_norm_rn),&(result->corner_point),NULL) > 0)
         {
         GMP_RALG_cfdecomp_init(&(result->rn_abs_recip_cf_rep),
                                &error_flag,
                                &(result->abs_norm_rn.den),
                                &(result->abs_norm_rn.num));
         result->rn_abs_recip_cf_rep_formed = 1;
         }
      }

   //If HMAX has been supplied, we want to calculate the points just below and above
   //the corner point.  The reason we want to do this is because we need to gracefully
   //"round the corner" in either direction.
   //
   //Calculate the point just to the left of the corner point.
   if (result->hmax_supplied)
      {
      GMP_RALG_enclosing_farey_neighbors(
                                      &(result->corner_point),
                                      &(result->kmax_in),
                                      &(result->corner_point_cf_rep),
                                      &equality_flag,
                                      &(result->corner_point_minus_one),
                                      &(q_temp1)
                                      );
      }

   //Calculate the point just to the right of the corner point.  This is
   //where HMAX is the dominant constraint.  We need to find the left
   //Farey neighbor to the reciprocal of the corner point in the Farey
   //series of order HMAX, then take its reciprocal.  There is the possibility
   //if KMAX=1 that this point will have a denominator of zero, but we
   //will accept that as a number here, since we should never hit it,
   //as it will be beyond HMAX/1.
   if (result->hmax_supplied)
      {
      GMP_RATS_mpq_copy(&q_temp1, &(result->corner_point));
      GMP_INTS_mpz_swap(&(q_temp1.num), &(q_temp1.den));
      GMP_RALG_enclosing_farey_neighbors(
                                      &q_temp1,
                                      &(result->hmax_in),
                                      &(result->corner_point_recip_cf_rep),
                                      &equality_flag,
                                      &(result->corner_point_plus_one),
                                      &(q_temp2)
                                      );
      GMP_INTS_mpz_swap(&(result->corner_point_plus_one.num), &(result->corner_point_plus_one.den));
      }

   //Calculate the complement of HMAX/1.  Nothing that we generate can go beyond
   //this to the south.
   if (result->hmax_supplied)
      {
      GMP_RATS_mpq_copy(&(hmax_over_one_neg), &(result->hmax_over_one));
      GMP_INTS_mpz_negate(&(hmax_over_one_neg.num));
      }

   //Also calculate the complement of HMAX/KMAX.
   if (result->hmax_supplied)
      {
      GMP_RATS_mpq_copy(&(corner_point_neg), &(result->corner_point));
      GMP_INTS_mpz_negate(&(corner_point_neg.num));
      }

   //Form the reciprocal of the absolute value of the normalized value of
   //the rational number in.
   GMP_RATS_mpq_copy(&abs_norm_recip_rn, &(result->abs_norm_rn));
   GMP_RATS_mpq_swap_components(&abs_norm_recip_rn);

   //OK, now we get down to brass tacks.  Iterate first to get the
   //left neighbors.  The ordinary complexity of this is also compounded
   //by the fact that we must handle negative numbers as well--everything
   //from -HMAX/1 to HMAX/1.
   //
   //PSEUDO-CODE:
   //  a)If the rational number to approximate is <= -HMAX/1 or there are no
   //    left neighbors requested, terminate with no neighbors on the left.
   //
   //  b)Find the right neighbor of the rational number supplied.
   //
   //  c)Find the left neighbor of the rational number supplied.
   //
   //  d)While (queued_count < count)
   //
   //     e)Queue the left neighbor, queued_count++
   //
   //     f)If (queued_count >= count), break.
   //
   //     g)If (left_neighbor <= -HMAX/1), break
   //
   //     h)Advance the frame one to the left.
   //
   //**************************************************************************
   //  a)If the rational number to approximate is <= -HMAX/1 or there are no
   //    left neighbors requested, terminate with no neighbors on the left.
   //**************************************************************************
   if ((result->hmax_supplied && GMP_RATS_mpq_cmp(&(result->norm_rn), &hmax_over_one_neg, NULL) <= 0)
      || (n_left_in <= 0))
      goto done_with_left_neighbors;

   //**************************************************************************
   //  b)Find the right neighbor of the rational number supplied.
   //**************************************************************************
   //  c)Find the left neighbor of the rational number supplied.
   //**************************************************************************
   if (!result->hmax_supplied)
      {
      //In this case, the notion of corner point is meaningless, because 
      //there is no constraint on H.  We can just go on our merry way.  Get
      //the two neighbors.
      GMP_RALG_enclosing_farey_neighbors(
                                      &(result->norm_rn),
                                      &(result->kmax_in),
                                      &(result->rn_abs_cf_rep),
                                      &equality_flag,
                                      &left_neighbor,
                                      &right_neighbor
                                      );
      //The enclosing Farey neighbor function is prohibited from identifying the
      //rational number itself as a Farey term.  If the number is in the Farey
      //series, we must replace the right neighbor, otherwise we cannot apply
      //the standard recursive formulas.
      if (equality_flag)
         {
         GMP_RATS_mpq_copy(&right_neighbor, &(result->norm_rn));
         }
      }
   else if (GMP_RATS_mpq_cmp(&(result->norm_rn), &corner_point_neg, NULL) < 0)
      {
      //The rational number specified is negative and below the negative corner point.
      //This means that HMAX is the dominant constraint.  We need to find the
      //neighbors in the Farey series of order HMAX, then reorder and invert, etc.
      GMP_RALG_enclosing_farey_neighbors(
                                      &abs_norm_recip_rn,
                                      &(result->hmax_in),
                                      &(result->rn_abs_recip_cf_rep),
                                      &equality_flag,
                                      &left_neighbor,
                                      &right_neighbor
                                      );

      //Again, if the number specified was already in the series of interest,
      //we need to swap in the right neighbor.
      if (equality_flag)
         {
         GMP_RATS_mpq_copy(&right_neighbor, &abs_norm_recip_rn);
         }

      //Take the reciprocal of both neighbors, which will put them out of order,
      //then negate them, which will put them back in order.
      GMP_RATS_mpq_swap_components(&left_neighbor);
      GMP_INTS_mpz_negate(&(left_neighbor.num));
      GMP_RATS_mpq_swap_components(&right_neighbor);
      GMP_INTS_mpz_negate(&(right_neighbor.num));
      }
   else if (GMP_RATS_mpq_cmp(&(result->norm_rn), &corner_point_neg, NULL) == 0)
      {
      //The rational number specified is the negative corner point.  In this case
      //Because we can never return the corner point itself as a left neighbor,
      //we need to set the left value to be the negative of one past, and the right
      //to be the negative of the corner point.
      GMP_RATS_mpq_copy(&left_neighbor, &(result->corner_point_plus_one));
      GMP_INTS_mpz_negate(&(left_neighbor.num));
      GMP_RATS_mpq_copy(&right_neighbor, &(result->corner_point));
      GMP_INTS_mpz_negate(&(right_neighbor.num));
      }
   else if ((GMP_RATS_mpq_cmp(&(result->norm_rn), &corner_point_neg, NULL) > 0) 
            && 
            (GMP_RATS_mpq_cmp(&(result->norm_rn), &(result->corner_point), NULL) < 0))
      {
      //The rational number specified is in the area dominated by the KMAX constraint
      //between -HMAX/KMAX and HMAX/KMAX.  The ordinary Farey neighbor function will
      //handle this correctly.  Again, we need to adjust the output if the number
      //is already formable, because the Farey neighbor function is prohibited from
      //returning the number itself as a neighbor.
      GMP_RALG_enclosing_farey_neighbors(
                                      &(result->norm_rn),
                                      &(result->kmax_in),
                                      &(result->rn_abs_cf_rep),
                                      &equality_flag,
                                      &left_neighbor,
                                      &right_neighbor
                                      );
      //The enclosing Farey neighbor function is prohibited from identifying the
      //rational number itself as a Farey term.  If the number is in the Farey
      //series, we must replace the right neighbor, otherwise we cannot apply
      //the standard recursive formulas.
      if (equality_flag)
         {
         GMP_RATS_mpq_copy(&right_neighbor, &(result->norm_rn));
         }
      }
   else if (GMP_RATS_mpq_cmp(&(result->norm_rn), &(result->corner_point), NULL) == 0)
      {
      //The rational number specified is the corner point.  In this case
      //because we can never return the corner point itself as a left neighbor,
      //we need to set the left value to be one before, and the right
      //to be the corner point.
      GMP_RATS_mpq_copy(&left_neighbor, &(result->corner_point_minus_one));
      GMP_RATS_mpq_copy(&right_neighbor, &(result->corner_point));
      }
   else
      {
      //The only possibility left is that the number is positive and above the
      //corner point where HMAX is the dominant constraint.
      GMP_RALG_enclosing_farey_neighbors(
                                      &abs_norm_recip_rn,
                                      &(result->hmax_in),
                                      &(result->rn_abs_recip_cf_rep),
                                      &equality_flag,
                                      &left_neighbor,
                                      &right_neighbor
                                      );

      //Again, if the number specified was already in the series of interest,
      //we need to swap in the neighbor.  This time, however, it must be the
      //left neighbor because taking the reciprocals will reverse the order.
      if (equality_flag)
         {
         GMP_RATS_mpq_copy(&left_neighbor, &abs_norm_recip_rn);
         }

      //Take the reciprocal of both neighbors, which will put them out of order,
      //then swap them, which will put them back in order.
      GMP_RATS_mpq_swap_components(&left_neighbor);
      GMP_RATS_mpq_swap_components(&right_neighbor);
      GMP_RATS_mpq_swap(&left_neighbor, &right_neighbor);
      }

   #if 0
   //Print out the left neighbor and right neighbor determined, for debugging.
   GMP_INTS_mpz_long_int_format_to_stream(stdout,
                                          &(left_neighbor.num),
                                          "left_neigh_num");
   GMP_INTS_mpz_long_int_format_to_stream(stdout,
                                          &(left_neighbor.den),
                                          "left_neigh_den");
   GMP_INTS_mpz_long_int_format_to_stream(stdout,
                                          &(right_neighbor.num),
                                          "right_neigh_num");
   GMP_INTS_mpz_long_int_format_to_stream(stdout,
                                          &(right_neighbor.den),
                                          "right_neigh_den");
   #endif


   //**************************************************************************
   //  d)While (queued_count < count)
   //**************************************************************************
   while (result->n_left_out < result->n_left_in)
      {
   //**************************************************************************
   //     e)Queue the left neighbor, queued_count++
   //**************************************************************************
      (result->lefts + result->n_left_out)->index = -(result->n_left_out + 1);
      GMP_RATS_mpq_init(&((result->lefts + result->n_left_out)->neighbor));
      GMP_RATS_mpq_copy(&((result->lefts + result->n_left_out)->neighbor), &left_neighbor);
      (result->n_left_out)++;

   //**************************************************************************
   //     f)If (queued_count >= count), break.
   //**************************************************************************
   //By the way, this step is to save unnecessary contortions once we've met
   //the quota.
      if (result->n_left_out >= result->n_left_in)
         break;

   //**************************************************************************
   //     g)If (left_neighbor <= -HMAX/1), break
   //**************************************************************************
   //This breaks us when we've queued the most negative number we can--can't go
   //further.  This only applies for cases where KMAX is also specified.
      if (result->hmax_supplied 
          && 
          GMP_RATS_mpq_cmp(&left_neighbor, &hmax_over_one_neg, NULL) <= 0)
         break;

   //**************************************************************************
   //     h)Advance the frame one to the left.
   //**************************************************************************
   //Advancing one frame to the left is a complicated affair, requiring several
   //subcases.  We break it up into regions which are best visualized using
   //a graph of the integer lattice with dots for each rational number.
      if (!(result->hmax_supplied))
         {
         //This is the case where we're are looking only in the 
         //Farey series.
         if (GMP_INTS_mpz_is_pos(&left_neighbor.num))
            {
            //In this case, the left neighbor and right neighbor
            //are both positive, and we can apply the standard
            //formulas.
            GMP_RALG_farey_predecessor(&q_temp1,
                                       &right_neighbor,
                                       &left_neighbor,
                                       &(result->kmax_in));
            GMP_RATS_mpq_copy(&right_neighbor, &left_neighbor);
            GMP_RATS_mpq_copy(&left_neighbor,  &q_temp1);
            }
         else if (GMP_INTS_mpz_is_zero(&left_neighbor.num))
            {
            //In this case, we are making the transition from positive
            //to negative.
            GMP_INTS_mpz_set_si(&(left_neighbor.num), -1); 
            GMP_INTS_mpz_copy(&(left_neighbor.den), &(result->kmax_in));
            GMP_RATS_mpq_set_si(&right_neighbor, 0, 1);
            }
         else
            {
            //Here, we are purely negative and decreasing.  Need to negate
            //the numbers, find successor, then negate.
            GMP_RATS_mpq_copy(&q_temp1, &left_neighbor);
            GMP_RATS_mpq_copy(&q_temp2, &right_neighbor);
            GMP_INTS_mpz_abs(&(q_temp1.num));
            GMP_INTS_mpz_abs(&(q_temp2.num));
            GMP_RALG_farey_successor(&q_temp3,
                                     &q_temp2,
                                     &q_temp1,
                                     &(result->kmax_in));
            GMP_RATS_mpq_copy(&right_neighbor, &left_neighbor);
            GMP_RATS_mpq_copy(&left_neighbor, &q_temp3);
            GMP_INTS_mpz_negate(&(left_neighbor.num));
            }
         }
      else if (GMP_RATS_mpq_cmp(&left_neighbor, &(result->corner_point), NULL) > 0)
         {
         //We are above the top corner point.  In this case HMAX is the dominant
         //constraint, and we find our food by taking reciprocals and applying
         //the standard relationships in the Farey series of order HMAX.
         GMP_RATS_mpq_copy(&q_temp1, &left_neighbor);
         GMP_RATS_mpq_copy(&q_temp2, &right_neighbor);
         GMP_RATS_mpq_swap_components(&q_temp1);
         GMP_RATS_mpq_swap_components(&q_temp2);
         GMP_RALG_farey_successor(&q_temp3,
                                  &q_temp2,
                                  &q_temp1,
                                  &(result->hmax_in));
         GMP_RATS_mpq_swap_components(&q_temp3);
         GMP_RATS_mpq_copy(&right_neighbor, &left_neighbor);
         GMP_RATS_mpq_copy(&left_neighbor, &q_temp3);
         }
      else if (GMP_RATS_mpq_cmp(&left_neighbor, &(result->corner_point), NULL) == 0)
         {
         //We are precisely at the corner point.  This is where we round the corner.
         GMP_RATS_mpq_copy(&right_neighbor, &left_neighbor);
         GMP_RATS_mpq_copy(&left_neighbor, &(result->corner_point_minus_one));
         }
      else if (GMP_INTS_mpz_is_pos(&left_neighbor.num))
         {
         //In this region we are going straight down the Farey series.
         GMP_RALG_farey_predecessor(&q_temp1,
                                    &right_neighbor,
                                    &left_neighbor,
                                    &(result->kmax_in));
         GMP_RATS_mpq_copy(&right_neighbor, &left_neighbor);
         GMP_RATS_mpq_copy(&left_neighbor,  &q_temp1);
         }
      else if (GMP_INTS_mpz_is_zero(&left_neighbor.num))
         {
         //In this case, we are making the transition from positive
         //to negative.
         GMP_INTS_mpz_set_si(&(left_neighbor.num), -1); 
         GMP_INTS_mpz_copy(&(left_neighbor.den), &(result->kmax_in));
         GMP_RATS_mpq_set_si(&right_neighbor, 0, 1);
         }
      else if (GMP_RATS_mpq_cmp(&left_neighbor, &corner_point_neg, NULL) > 0)
         {
         //Here, we are purely negative and decreasing.  Need to negate
         //the numbers, find successor, then negate.
         GMP_RATS_mpq_copy(&q_temp1, &left_neighbor);
         GMP_RATS_mpq_copy(&q_temp2, &right_neighbor);
         GMP_INTS_mpz_abs(&(q_temp1.num));
         GMP_INTS_mpz_abs(&(q_temp2.num));
         GMP_RALG_farey_successor(&q_temp3,
                                  &q_temp2,
                                  &q_temp1,
                                  &(result->kmax_in));
         GMP_RATS_mpq_copy(&right_neighbor, &left_neighbor);
         GMP_RATS_mpq_copy(&left_neighbor, &q_temp3);
         GMP_INTS_mpz_negate(&(left_neighbor.num));
         }
      else if (GMP_RATS_mpq_cmp(&left_neighbor, &corner_point_neg, NULL) == 0)
         {
         //This is where we are rounding the negative corner.
         GMP_RATS_mpq_copy(&right_neighbor, &left_neighbor);
         GMP_RATS_mpq_copy(&left_neighbor, &(result->corner_point_plus_one));
         GMP_INTS_mpz_negate(&(left_neighbor.num));
         }
      else
         {
         //Here we're going in the negative direction along the "bottom" of the
         //rectangle.
         GMP_RATS_mpq_copy(&q_temp1, &left_neighbor);
         GMP_RATS_mpq_copy(&q_temp2, &right_neighbor);
         GMP_INTS_mpz_abs(&(q_temp1.num));
         GMP_INTS_mpz_abs(&(q_temp2.num));
         GMP_RATS_mpq_swap_components(&q_temp1);
         GMP_RATS_mpq_swap_components(&q_temp2);
         GMP_RALG_farey_predecessor(&q_temp3,
                                    &q_temp2,
                                    &q_temp1,
                                    &(result->hmax_in));
         GMP_RATS_mpq_swap_components(&q_temp3);
         GMP_INTS_mpz_negate(&(q_temp3.num));
         GMP_RATS_mpq_copy(&right_neighbor, &left_neighbor);
         GMP_RATS_mpq_copy(&left_neighbor, &q_temp3);
         }
      }


   done_with_left_neighbors: ;

   //**************************************************************************
   //  a)If the rational number to approximate is >= HMAX/1 or there are no
   //    right neighbors requested, terminate with no neighbors on the right.
   //**************************************************************************
   if ((result->hmax_supplied && GMP_RATS_mpq_cmp(&(result->norm_rn), &(result->hmax_over_one), NULL) >= 0)
      || (n_right_in <= 0))
      goto done_with_right_neighbors;

   //**************************************************************************
   //  b)Find the right neighbor of the rational number supplied.
   //**************************************************************************
   //  c)Find the left neighbor of the rational number supplied.
   //**************************************************************************
   if (!result->hmax_supplied)
      {
      //In this case, the notion of corner point is meaningless, because 
      //there is no constraint on H.  We can just go on our merry way.  Get
      //the two neighbors.
      GMP_RALG_enclosing_farey_neighbors(
                                      &(result->norm_rn),
                                      &(result->kmax_in),
                                      &(result->rn_abs_cf_rep),
                                      &equality_flag,
                                      &left_neighbor,
                                      &right_neighbor
                                      );
      //The enclosing Farey neighbor function is prohibited from identifying the
      //rational number itself as a Farey term.  If the number is in the Farey
      //series, we must replace the left neighbor, otherwise we cannot apply
      //the standard recursive formulas.
      if (equality_flag)
         {
         GMP_RATS_mpq_copy(&left_neighbor, &(result->norm_rn));
         }
      }
   else if (GMP_RATS_mpq_cmp(&(result->norm_rn), &corner_point_neg, NULL) < 0)
      {
      //The rational number specified is negative and below the negative corner point.
      //This means that HMAX is the dominant constraint.  We need to find the
      //neighbors in the Farey series of order HMAX, then reorder and invert, etc.
      GMP_RALG_enclosing_farey_neighbors(
                                      &abs_norm_recip_rn,
                                      &(result->hmax_in),
                                      &(result->rn_abs_recip_cf_rep),
                                      &equality_flag,
                                      &left_neighbor,
                                      &right_neighbor
                                      );

      //Again, if the number specified was already in the series of interest,
      //we need to swap in the left neighbor.
      if (equality_flag)
         {
         GMP_RATS_mpq_copy(&left_neighbor, &abs_norm_recip_rn);
         }

      //Take the reciprocal of both neighbors, which will put them out of order,
      //then negate them, which will put them back in order.
      GMP_RATS_mpq_swap_components(&left_neighbor);
      GMP_INTS_mpz_negate(&(left_neighbor.num));
      GMP_RATS_mpq_swap_components(&right_neighbor);
      GMP_INTS_mpz_negate(&(right_neighbor.num));
      }
   else if (GMP_RATS_mpq_cmp(&(result->norm_rn), &corner_point_neg, NULL) == 0)
      {
      //The rational number specified is the negative corner point.  In this case
      //Because we can never return the corner point itself as a left neighbor,
      //we need to set the right value to be the negative of one before, and the left
      //to be the negative of the corner point.
      GMP_RATS_mpq_copy(&left_neighbor, &(result->corner_point));
      GMP_INTS_mpz_negate(&(left_neighbor.num));
      GMP_RATS_mpq_copy(&right_neighbor, &(result->corner_point_minus_one));
      GMP_INTS_mpz_negate(&(right_neighbor.num));
      }
   else if ((GMP_RATS_mpq_cmp(&(result->norm_rn), &corner_point_neg, NULL) > 0) 
            && 
            (GMP_RATS_mpq_cmp(&(result->norm_rn), &(result->corner_point), NULL) < 0))
      {
      //The rational number specified is in the area dominated by the KMAX constraint
      //between -HMAX/KMAX and HMAX/KMAX.  The ordinary Farey neighbor function will
      //handle this correctly.  Again, we need to adjust the output if the number
      //is already formable, because the Farey neighbor function is prohibited from
      //returning the number itself as a neighbor.
      GMP_RALG_enclosing_farey_neighbors(
                                      &(result->norm_rn),
                                      &(result->kmax_in),
                                      &(result->rn_abs_cf_rep),
                                      &equality_flag,
                                      &left_neighbor,
                                      &right_neighbor
                                      );
      //The enclosing Farey neighbor function is prohibited from identifying the
      //rational number itself as a Farey term.  If the number is in the Farey
      //series, we must replace the left neighbor, otherwise we cannot apply
      //the standard recursive formulas.
      if (equality_flag)
         {
         GMP_RATS_mpq_copy(&left_neighbor, &(result->norm_rn));
         }
      }
   else if (GMP_RATS_mpq_cmp(&(result->norm_rn), &(result->corner_point), NULL) == 0)
      {
      //The rational number specified is the positive corner point.  In this case.
      //because we can never return the corner point itself as a right neighbor,
      //we need to set the right value to be one after, and the left
      //to be the corner point.
      GMP_RATS_mpq_copy(&left_neighbor, &(result->corner_point));
      GMP_RATS_mpq_copy(&right_neighbor, &(result->corner_point_plus_one));
      }
   else
      {
      //The only possibility left is that the number is positive and at or above the
      //corner point where HMAX is the dominant constraint.
      GMP_RALG_enclosing_farey_neighbors(
                                      &abs_norm_recip_rn,
                                      &(result->hmax_in),
                                      &(result->rn_abs_recip_cf_rep),
                                      &equality_flag,
                                      &left_neighbor,
                                      &right_neighbor
                                      );

      //Again, if the number specified was already in the series of interest,
      //we need to swap in the neighbor.  This time, however, it must be the
      //right neighbor because taking the reciprocals will reverse the order.
      if (equality_flag)
         {
         GMP_RATS_mpq_copy(&right_neighbor, &abs_norm_recip_rn);
         }

      //Take the reciprocal of both neighbors, which will put them out of order,
      //then swap them, which will put them back in order.
      GMP_RATS_mpq_swap_components(&left_neighbor);
      GMP_RATS_mpq_swap_components(&right_neighbor);
      GMP_RATS_mpq_swap(&left_neighbor, &right_neighbor);
      }

   #if 0
   //Print out the left neighbor and right neighbor determined, for debugging.
   GMP_INTS_mpz_long_int_format_to_stream(stdout,
                                          &(left_neighbor.num),
                                          "left_neigh_num");
   GMP_INTS_mpz_long_int_format_to_stream(stdout,
                                          &(left_neighbor.den),
                                          "left_neigh_den");
   GMP_INTS_mpz_long_int_format_to_stream(stdout,
                                          &(right_neighbor.num),
                                          "right_neigh_num");
   GMP_INTS_mpz_long_int_format_to_stream(stdout,
                                          &(right_neighbor.den),
                                          "right_neigh_den");
   #endif


   //**************************************************************************
   //  d)While (queued_count < count)
   //**************************************************************************
   while (result->n_right_out < result->n_right_in)
      {
   //**************************************************************************
   //     e)Queue the right neighbor, queued_count++
   //**************************************************************************
      (result->rights + result->n_right_out)->index = (result->n_right_out + 1);
      GMP_RATS_mpq_init(&((result->rights + result->n_right_out)->neighbor));
      GMP_RATS_mpq_copy(&((result->rights + result->n_right_out)->neighbor), &right_neighbor);
      (result->n_right_out)++;

   //**************************************************************************
   //     f)If (queued_count >= count), break.
   //**************************************************************************
   //By the way, this step is to save unnecessary contortions once we've met
   //the quota.
      if (result->n_right_out >= result->n_right_in)
         break;

   //**************************************************************************
   //     g)If (right_neighbor >= HMAX/1), break
   //**************************************************************************
   //This breaks us when we've queued the most positive number we can--can't go
   //further.  This only applies for cases where KMAX is also specified.
      if (result->hmax_supplied 
          && 
          GMP_RATS_mpq_cmp(&right_neighbor, &(result->hmax_over_one), NULL) >= 0)
         break;

   //**************************************************************************
   //     h)Advance the frame one to the right.
   //**************************************************************************
   //Advancing one frame to the right is a complicated affair, requiring several
   //subcases.  We break it up into regions which are best visualized using
   //a graph of the integer lattice with dots for each rational number.
      if (!(result->hmax_supplied))
         {
         //This is the case where we're are looking only in the 
         //Farey series.
         if (GMP_INTS_mpz_is_neg(&right_neighbor.num))
            {
            //Neg and increasing towards zero.  Can handle by symmetry.
            GMP_RATS_mpq_copy(&q_temp1, &left_neighbor);
            GMP_RATS_mpq_copy(&q_temp2, &right_neighbor);
            GMP_INTS_mpz_abs(&(q_temp1.num));
            GMP_INTS_mpz_abs(&(q_temp2.num));
            GMP_RALG_farey_predecessor(&q_temp3,
                                     &q_temp1,
                                     &q_temp2,
                                     &(result->kmax_in));
            GMP_RATS_mpq_copy(&left_neighbor,  &right_neighbor);
            GMP_RATS_mpq_copy(&right_neighbor, &q_temp3);
            GMP_INTS_mpz_negate(&(right_neighbor.num));
            }
         else if (GMP_INTS_mpz_is_zero(&right_neighbor.num))
            {
            //Right term just hit zero and are increasing.  
            //Left will become 0/1, right becomes 1/KMAX.
            GMP_RATS_mpq_set_si(&left_neighbor, 0, 1);
            GMP_INTS_mpz_set_si(&(right_neighbor.num), 1); 
            GMP_INTS_mpz_copy(&(right_neighbor.den), &(result->kmax_in));
            }
         else
            {
            //Are above zero and increasing.  Can use standard Farey
            //successor formula.
            GMP_RALG_farey_successor(&q_temp1,
                                     &left_neighbor,
                                     &right_neighbor,
                                     &(result->kmax_in));
            GMP_RATS_mpq_copy(&left_neighbor,   &right_neighbor);
            GMP_RATS_mpq_copy(&right_neighbor,  &q_temp1);
            }
         }
      else if (GMP_RATS_mpq_cmp(&right_neighbor, &corner_point_neg, NULL) < 0)
         {
         //We are below the negative corner point and moving towards
         //zero, with HMAX the dominant constraint.  We can proceed by
         //symmetry, producing a Farey successor and negating and inverting.
         GMP_RATS_mpq_copy(&q_temp1, &left_neighbor);
         GMP_RATS_mpq_copy(&q_temp2, &right_neighbor);
         GMP_INTS_mpz_abs(&(q_temp1.num));
         GMP_INTS_mpz_abs(&(q_temp2.num));
         GMP_RATS_mpq_swap_components(&q_temp1);
         GMP_RATS_mpq_swap_components(&q_temp2);
         GMP_RALG_farey_successor(&q_temp3,
                                  &q_temp1,
                                  &q_temp2,
                                  &(result->hmax_in));
         GMP_RATS_mpq_swap_components(&q_temp3);
         GMP_INTS_mpz_negate(&(q_temp3.num));
         GMP_RATS_mpq_copy(&left_neighbor, &right_neighbor);
         GMP_RATS_mpq_copy(&right_neighbor, &q_temp3);
         }
      else if (GMP_RATS_mpq_cmp(&right_neighbor, &corner_point_neg, NULL) == 0)
         {
         //We are at the bottom negative corner point and need to "round the corner".
         GMP_RATS_mpq_copy(&left_neighbor, &right_neighbor);
         GMP_RATS_mpq_copy(&right_neighbor, &(result->corner_point_minus_one));
         GMP_INTS_mpz_negate(&(right_neighbor.num));
         }
      else if (GMP_INTS_mpz_is_neg(&right_neighbor.num))
         {
         //In this region we are going straight up the Farey series approaching
         //zero.  Need to negate to use standard relationships.
         GMP_RATS_mpq_copy(&q_temp1, &left_neighbor);
         GMP_RATS_mpq_copy(&q_temp2, &right_neighbor);
         GMP_INTS_mpz_abs(&(q_temp1.num));
         GMP_INTS_mpz_abs(&(q_temp2.num));
         GMP_RALG_farey_predecessor(&q_temp3,
                                    &q_temp1,
                                    &q_temp2,
                                    &(result->kmax_in));
         GMP_RATS_mpq_copy(&left_neighbor, &right_neighbor);
         GMP_RATS_mpq_copy(&right_neighbor,  &q_temp3);
         GMP_INTS_mpz_negate(&(right_neighbor.num));
         }
      else if (GMP_INTS_mpz_is_zero(&right_neighbor.num))
         {
         //Zero crossover.
         GMP_RATS_mpq_set_si(&left_neighbor, 0, 1);
         GMP_INTS_mpz_set_si(&(right_neighbor.num), 1); 
         GMP_INTS_mpz_copy(&(right_neighbor.den), &(result->kmax_in));
         }
      else if (GMP_RATS_mpq_cmp(&right_neighbor, &(result->corner_point), NULL) < 0)
         {
         //Below corner point.  Standard relationship applies.
         GMP_RALG_farey_successor(&q_temp1,
                                  &left_neighbor,
                                  &right_neighbor,
                                  &(result->kmax_in));
         GMP_RATS_mpq_copy(&left_neighbor,   &right_neighbor);
         GMP_RATS_mpq_copy(&right_neighbor,  &q_temp1);
         }
      else if (GMP_RATS_mpq_cmp(&right_neighbor, &(result->corner_point), NULL) == 0)
         {
         //At the positive corner point.
         GMP_RATS_mpq_copy(&left_neighbor, &right_neighbor);
         GMP_RATS_mpq_copy(&right_neighbor, &(result->corner_point_plus_one));
         }
      else
         {
         //Above the positive corner point and heading for HMAX/1.
         GMP_RATS_mpq_copy(&q_temp1, &left_neighbor);
         GMP_RATS_mpq_copy(&q_temp2, &right_neighbor);
         GMP_RATS_mpq_swap_components(&q_temp1);
         GMP_RATS_mpq_swap_components(&q_temp2);
         GMP_RALG_farey_predecessor(&q_temp3,
                                    &q_temp1,
                                    &q_temp2,
                                    &(result->hmax_in));
         GMP_RATS_mpq_swap_components(&q_temp3);
         GMP_RATS_mpq_copy(&left_neighbor, &right_neighbor);
         GMP_RATS_mpq_copy(&right_neighbor, &q_temp3);
         }
      }

   done_with_right_neighbors: ;

   //This is a single return point so we catch all the dynamic memory 
   //deallocation.
   return_point:
      GMP_RATS_mpq_clear(&q_temp1);
      GMP_RATS_mpq_clear(&q_temp2);
      GMP_RATS_mpq_clear(&q_temp3);
      GMP_RATS_mpq_clear(&q_temp4);
      GMP_RATS_mpq_clear(&left_neighbor);
      GMP_RATS_mpq_clear(&right_neighbor);
      GMP_RATS_mpq_clear(&left_neighbor_abs);
      GMP_RATS_mpq_clear(&right_neighbor_abs);
      GMP_RATS_mpq_clear(&hmax_over_one_neg);
      GMP_RATS_mpq_clear(&corner_point_neg);
      GMP_RATS_mpq_clear(&abs_norm_recip_rn);
   }


//08/16/01: Visual inspection OK.
void GMP_RALG_consecutive_fab_terms_result_free(
                                               GMP_RALG_fab_neighbor_collection_struct *arg
                                               )
   {
   int i;

   //Eyeball the input.
   assert(arg != NULL);

   //Deallocate all rational numbers and integers that MUST be allocated, i.e. they are
   //never conditional.
   GMP_RATS_mpq_clear(&(arg->rn_in));
   GMP_INTS_mpz_clear(&(arg->kmax_in));
   GMP_INTS_mpz_clear(&(arg->hmax_in));
   GMP_RATS_mpq_clear(&(arg->hmax_over_one));
   GMP_RATS_mpq_clear(&(arg->corner_point));
   GMP_RATS_mpq_clear(&(arg->corner_point_minus_one));
   GMP_RATS_mpq_clear(&(arg->corner_point_plus_one));
   GMP_RATS_mpq_clear(&(arg->norm_rn));
   GMP_RATS_mpq_clear(&(arg->abs_norm_rn));

   //Destroy any continued fraction representations that were
   //formed.
   if (arg->rn_abs_cf_rep_formed)
      {
      GMP_RALG_cfdecomp_destroy(&(arg->rn_abs_cf_rep));
      }
   if (arg->rn_abs_recip_cf_rep_formed)
      {
      GMP_RALG_cfdecomp_destroy(&(arg->rn_abs_recip_cf_rep));
      }
   if(arg->corner_point_cf_rep_formed)
      {
      GMP_RALG_cfdecomp_destroy(&(arg->corner_point_cf_rep));
      }
   if(arg->corner_point_recip_cf_rep_formed)
      {
      GMP_RALG_cfdecomp_destroy(&(arg->corner_point_recip_cf_rep));
      }

   //Walk through the lefts, freeing up any allocated rational numbers.
   for (i=0; i < arg->n_left_out; i++)
      {
      GMP_RATS_mpq_clear(&(arg->lefts[i].neighbor));
      }

   //Walk through the rights, freeing up any allocated rational numbers.
   for (i=0; i < arg->n_right_out; i++)
      {
      GMP_RATS_mpq_clear(&(arg->rights[i].neighbor));
      }
   
   //Deallocate any area assigned for lefts.
   if (arg->lefts)
      {
      #if defined(APP_TYPE_SIMPLE_DOS_CONSOLE)
         CCMALLOC_free(arg->lefts);
      #elif defined(APP_TYPE_IJUSCRIPTER_IJUCONSOLE)
         TclpFree((char *)arg->lefts);
      #else
         free(arg->lefts);
      #endif 

      arg->lefts = NULL;
      }

   //Deallocate any area assigned for rights.
   if (arg->rights)
      {
      #if defined(APP_TYPE_SIMPLE_DOS_CONSOLE)
         CCMALLOC_free(arg->rights);
      #elif defined(APP_TYPE_IJUSCRIPTER_IJUCONSOLE)
         TclpFree((char *)arg->rights);
      #else
         free(arg->rights);
      #endif 

      arg->rights = NULL;
      }
   }


//08/16/01: Visual inspection OK.
void GMP_RALG_consecutive_fab_terms_result_dump(
                                               FILE *s,
                                               GMP_RALG_fab_neighbor_collection_struct *arg
                                               )
   {
   int i;
   char buf[250];

   //Eyeball the input parameters.
   assert(s   != NULL);
   assert(arg != NULL);

   //Announce intent.
   FCMIOF_stream_bannerheading(s,
                               "Emitting Neighbor Decomp For Analysis", 
                               1);

   //Dump the fields, in order.
   GMP_INTS_mpz_long_int_format_to_stream(s,
                                          &(arg->rn_in.num),
                                          "rn_in_num");
   GMP_INTS_mpz_long_int_format_to_stream(s,
                                          &(arg->rn_in.den),
                                          "rn_in_den");
   GMP_INTS_mpz_long_int_format_to_stream(s,
                                          &(arg->kmax_in),
                                          "kmax_in");
   fprintf(s, "       hmax_supplied:                          %12d\n", arg->hmax_supplied);
   GMP_INTS_mpz_long_int_format_to_stream(s,
                                          &(arg->hmax_in),
                                          "hmax_in");
   if (arg->error)
      {
      fprintf(s, "               error: %s\n", arg->error);
      }
   else
      {
      fprintf(s, "               error:                                     NULL\n");
      }

   GMP_INTS_mpz_long_int_format_to_stream(s,
                                          &(arg->corner_point.num),
                                          "corner_point_num");
   GMP_INTS_mpz_long_int_format_to_stream(s,
                                          &(arg->corner_point.den),
                                          "corner_point_den");

   GMP_INTS_mpz_long_int_format_to_stream(s,
                                          &(arg->corner_point_minus_one.num),
                                          "cor_pt_minus_one_num");
   GMP_INTS_mpz_long_int_format_to_stream(s,
                                          &(arg->corner_point_minus_one.den),
                                          "cor_pt_minus_one_den");

   GMP_INTS_mpz_long_int_format_to_stream(s,
                                          &(arg->corner_point_plus_one.num),
                                          "cor_pt_plus_one_num");
   GMP_INTS_mpz_long_int_format_to_stream(s,
                                          &(arg->corner_point_plus_one.den),
                                          "cor_pt_plus_one_den");

   GMP_INTS_mpz_long_int_format_to_stream(s,
                                          &(arg->hmax_over_one.num),
                                          "hmax/1_num");
   GMP_INTS_mpz_long_int_format_to_stream(s,
                                          &(arg->hmax_over_one.den),
                                          "hmax/1_den");

   GMP_INTS_mpz_long_int_format_to_stream(s,
                                          &(arg->norm_rn.num),
                                          "norm_rn_num");
   GMP_INTS_mpz_long_int_format_to_stream(s,
                                          &(arg->norm_rn.den),
                                          "norm_rn_den");

   GMP_INTS_mpz_long_int_format_to_stream(s,
                                          &(arg->abs_norm_rn.num),
                                          "abs_norm_rn_num");
   GMP_INTS_mpz_long_int_format_to_stream(s,
                                          &(arg->abs_norm_rn.den),
                                          "abs_norm_rn_den");

   fprintf(s, "           rn_is_neg:                          %12d\n", arg->rn_is_neg);

   fprintf(s, "           n_left_in:                          %12d\n", arg->n_left_in);

   fprintf(s, "          n_right_in:                          %12d\n", arg->n_right_in);

   fprintf(s, "rn_abs_cf_rep_formed:                          %12d\n", arg->rn_abs_cf_rep_formed);

   if (arg->rn_abs_cf_rep_formed)
      {
      GMP_RALG_cfdecomp_emit(s, "Abs Of RN In CF Decomp", &(arg->rn_abs_cf_rep), 0, 0, NULL);
      }

   fprintf(s, "rn_abs_recip_cf_rep_formed:                    %12d\n", arg->rn_abs_recip_cf_rep_formed);

   if (arg->rn_abs_recip_cf_rep_formed)
      {
      GMP_RALG_cfdecomp_emit(s, "Abs Of Recip Of RN In CF Decomp", &(arg->rn_abs_recip_cf_rep), 0, 0, NULL);
      }

   fprintf(s, "corner_point_cf_rep_formed:                    %12d\n", arg->corner_point_cf_rep_formed);

   if (arg->corner_point_cf_rep_formed)
      {
      GMP_RALG_cfdecomp_emit(s, "Corner Point CF Decomp", &(arg->corner_point_cf_rep), 0, 0, NULL);
      }

   fprintf(s, "cor_pt_recip_cf_rep_formed:                    %12d\n", arg->corner_point_recip_cf_rep_formed);

   if (arg->corner_point_recip_cf_rep_formed)
      {
      GMP_RALG_cfdecomp_emit(s, "Corner Point Recip CF Decomp", &(arg->corner_point_recip_cf_rep), 0, 0, NULL);
      }

   fprintf(s, "                  equality:                    %12d\n", arg->equality);

   fprintf(s, "                n_left_out:                    %12d\n", arg->n_left_out);

   fprintf(s, "               n_right_out:                    %12d\n", arg->n_right_out);

   for (i=0; i < arg->n_left_out; i++)
      {
      sprintf(buf, "Contents Of Left Neighbor Array [%d]", i);
      FCMIOF_stream_bannerheading(s,
                                  buf, 
                                  0);
      fprintf(s, "                     index:                    %12d\n", arg->lefts[i].index);
      GMP_INTS_mpz_long_int_format_to_stream(s,
                                             &(arg->lefts[i].neighbor.num),
                                             "neighbor_num");
      GMP_INTS_mpz_long_int_format_to_stream(s,
                                             &(arg->lefts[i].neighbor.den),
                                             "neighbor_den");
      }

   for (i=0; i < arg->n_right_out; i++)
      {
      sprintf(buf, "Contents Of Right Neighbor Array [%d]", i);
      FCMIOF_stream_bannerheading(s,
                                  buf, 
                                  0);
      fprintf(s, "                     index:                    %12d\n", arg->rights[i].index);
      GMP_INTS_mpz_long_int_format_to_stream(s,
                                             &(arg->rights[i].neighbor.num),
                                             "neighbor_num");
      GMP_INTS_mpz_long_int_format_to_stream(s,
                                             &(arg->rights[i].neighbor.den),
                                             "neighbor_den");
      }

   FCMIOF_stream_hline(s);
   }


/******************************************************************/
/***  VERSION CONTROL REPORTING FUNCTIONS  ************************/
/******************************************************************/

//08/16/01: Visual inspection OK.
const char *GMP_RALG_cvcinfo(void)
   {
   return("$Header: /cvsroot/esrg/sfesrg/esrgpcpj/shared/c_datd/gmp_ralg.c,v 1.10 2002/01/27 17:58:15 dtashley Exp $");
   }


//08/16/01: Visual inspection OK.
const char *GMP_RALG_hvcinfo(void)
   {
   return(GMP_RALG_H_VERSION);
   }


//**************************************************************************
// $Log: gmp_ralg.c,v $
// Revision 1.10  2002/01/27 17:58:15  dtashley
// CRC32, other programs modified to work under new directory structure.
//
// Revision 1.9  2001/08/18 18:33:13  dtashley
// Preparing for release of v1.05.
//
// Revision 1.8  2001/08/16 19:49:40  dtashley
// Beginning to prepare for v1.05 release.
//
// Revision 1.7  2001/08/15 06:56:05  dtashley
// Substantial progress.  Safety check-in.
//
// Revision 1.6  2001/08/12 10:20:58  dtashley
// Safety check-in.  Substantial progress.
//
// Revision 1.5  2001/08/07 10:42:48  dtashley
// Completion of CFRATNUM extensions and DOS command-line utility.
//
// Revision 1.4  2001/07/13 21:02:20  dtashley
// Version control reporting changes.
//
// Revision 1.3  2001/07/13 20:44:42  dtashley
// Changes, CVS keyword expansion test.
//
// Revision 1.2  2001/07/13 00:57:08  dtashley
// Safety check-in.  Substantial progress on port.
//
// Revision 1.1  2001/07/12 05:42:06  dtashley
// Initial checkin.
//
//**************************************************************************
// End of GMP_RALG.C.